#ifdef COUPLED
#include "params.h"
#else
#include "om3params.h"
#endif

/* OM3 was written by John Anderson of the Space Science
  and Engineering Center, University of Wisconsin, Madison */

/* this version of OM3 modified for use within FOAM
   by Robert Jacob.   */

/* See README.FOAM for usage info */

/*version for parallel machines*/

/* 
   om3_v1.4.cx -> om3.c
   128 x 128 point model on pole to pole geometry, currently with reduced
   metric for latitudes > 70 deg.

   Hydrostatic level model with 16 levels taken as the odd levels
   from the GFDL 33 level set.

   Double time split (LF advection/diffustion, F-B gravity, Euler external)
   with reduced surface gravity.
*/

/*
   Define file names for restart init and save files if needed.
   This should become part of an integrated model control procedure
   ASAP.
*/
#define RestartRun
#define RestartSave
#define RestartRunName "om3.restart"
#define RestartSaveName "om3.save"



#include <stdio.h>
#include <fcntl.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "mpi.h"

#define Np (Ncols*Nrows)
#define Lx 4.0e7  /* Global domain from 0W - 360W */
#define Ly 2.0e7  /* 90S - 90N */
#define Nx 128
#define Ny 128
#define Nz 24
#define Mx ((Nx/Ncols)+2)
#define My ((Ny/Nrows)+2)
#define Mz Nz
#define MISSING 1.0e35

/* include the constants for the MOM polynomial
 approximation to the UNESCO eqn of state */
#include "dncoef24.h"

#define Delx (Lx / Nx)
#define Dely (Ly / Ny)
#define Diff_Mks 8.0e3
#define Diff_Ratio 0.5
#define G 0.50
#define N_Split 3
#define Sub_Split 8

#define Delt (6.0*3600.0)

int num_nodes, my_address;
int row,col,north,south,east,west;
int ofile;
int hold_tme,milliclock();
extern int hold_tme;
extern char time_inv[256];

float t_forcing[My][Mx], s_forcing[My][Mx];
float taux_for[My][Mx], tauy_for[My][Mx];
float avtaux[My][Mx], avtauy[My][Mx];
float avheat[My][Mx], avsalt[My][Mx];
float convec_freq[My][Mx];
float convec_freq2[My][Mx];
/* these fields are used for the short, 2d history*/

float uvts_levels[Mz],w_levels[Mz]; /* read from geometry file */
float h[Mz][My][Mx], uvh[Mz][My][Mx]; /* level thickneses, 0 if no water */
float hw[Nz][Ny][Nx];

float strx[My][Mx],stry[My][Mx]; /* wind stress */ 
float annual_salt[Mz][My][Mx],annual_temp[Mz][My][Mx]; /* mean salt and temp for spin up */

float u[Mz][My][Mx],ul[Mz][My][Mx],ux[Mz][My][Mx],dtu[Mz][My][Mx]; /* State Vars */
float v[Mz][My][Mx],vl[Mz][My][Mx],vx[Mz][My][Mx],dtv[Mz][My][Mx];
float t[Mz][My][Mx],tl[Mz][My][Mx],tx[Mz][My][Mx],dtt[Mz][My][Mx];
float s[Mz][My][Mx],sl[Mz][My][Mx],sx[Mz][My][Mx],dts[Mz][My][Mx];


/* old 3-D diagnostic fields */
/*float diag1[Mz][My][Mx],diag2[Mz][My][Mx],diag3[Mz][My][Mx],
      diag4[Mz][My][Mx],diag5[Mz][My][Mx];*/

/* new 3-D diagnostic fields for T and S */
float diag6[2][Mz][My][Mx],diag7[2][Mz][My][Mx];
/*float diag0[2][Mz][My][Mx],diag8[2][Mz][My][Mx];*/

/* these fields are used for the long, 3d history*/

/* these fields are used for the long, 3d history*/
float uav[Mz][My][Mx],vav[Mz][My][Mx],tav[Mz][My][Mx];
float wav[Mz][My][Mx],sav[Mz][My][Mx],pav[Mz][My][Mx];
float ptop[My][Mx],ptopl[My][Mx],ptopx[My][Mx]; /* surface pressure */
/* end of state */

float mix_u[Mz][My][Mx],mix_v[Mz][My][Mx],mix_t[Mz][My][Mx],mix_s[Mz][My][Mx];
float risq[Mz][My][Mx], nsq[Mz][My][Mx], kuv[Mz][My][Mx], kts[Mz][My][Mx];
/* Richarson number squared, k's */

float w[Mz][My][Mx],p[Mz][My][Mx],dzt[Mz][My][Mx],dzs[Mz][My][Mx]; /* derived state variables */
float dzu[Mz][My][Mx],dzv[Mz][My][Mx]; /* derived state variables */
float ubaro[My][Mx],vbaro[My][Mx],dtubaro[My][Mx],dtvbaro[My][Mx]; /* barotropic split variables */
float ubstar[My][Mx],vbstar[My][Mx],ptstar[My][Mx]; /* Euler step temps */

float f[My],taux[My][Mx],tauy[My][Mx]; /* Derived geometries */
float geo1[My], geo2[My], geo3[My];
float salt_for[My][Mx],temp_for[My][Mx];
float depth[My][Mx], uvdepth[My][Mx], rdepth[My][Mx], ruvdepth[My][Mx];
float maxdepth;


int displace[Nrows*Ncols],counts[Nrows*Ncols];
int rows[Nrows*Ncols],cols[Nrows*Ncols];
int mydisplace;

/* sea ice formation variables */
#define izmxice 3
float qice[Mz][My][Mx];
float qflx[My][Mx];
float Cp,Lathf,Tf,potice;
float rhoref, Sref;
float CtoK; /* celcius to kelvin */

float to_coupler[2][My-2][Mx-2], from_coupler[4][My-2][Mx-2];

float base_t[Mz],base_s[Mz]; /* base stratification */

float work[My][Mx];
char buff[256];
char resname[256];
char hisname[256];
char srbuff[Mx*My*8*10];
FILE *of,*ef;
FILE *checkfile;
FILE *hfile;

MPI_Datatype mpi_edge_type;

extern MPI_Comm split_comm, coupler_ocean;
extern char prefix[100], storage[100];
extern int rnlng,hsfrq,rsfrq,finished;

#ifdef OMPICS
unsigned char picbuf[My-2][Mx-2];
#endif

void init_state(); /* initialization routines */
void read_geometry();
void read_stress();
void read_annual_temp();
void read_annual_salt();

#ifdef ANNUAL_SST
void read_sst_data();
float sstint();
float sssint();
float tauxint();
float tauyint();
float heatint();
float saltint();
float sstm[12][My-2][Mx-2];
float sssm[12][My-2][Mx-2];
float sssom3[My-2][Mx-2];
float tauxm[12][My-2][Mx-2];
float tauym[12][My-2][Mx-2];
float saltm[12][My-2][Mx-2];
float heatm[12][My-2][Mx-2];
#define Start_step 0
#define Restore (1.0/(30.0 * 24.0 * 3600.0))
#endif


void time_step(); /* model run time code */
float density(float s, float t, int iz);
void filt4s(float q[Mz][My][Mx]);
void filt4m(float q[Mz][My][Mx]);
void filt4p(float q[My][Mx]);
void wrap_q(float a[My][Mx]);
void wrap_q3(float a[My][Mx], float b[My][Mx], float c[My][Mx]);
void wrap_qz(float a[Mz][My][Mx]);

#ifndef OCEAN_ALONE
void ocean_to_coupler(int nfields,float data[][My-2][Mx-2],int tag);
void ocean_from_coupler(int nfields,float data[][My-2][Mx-2],int tag);
#endif

#ifdef OMPICS
void write_picture(float a[My][Mx],float vmax); /* movie and history file output
routines */
void fwrite_picture(float a[My][Mx],float vmax);
#endif
void write_history(float a[Mz][My][Mx]);
void write_history_slice(float a[Mz][My][Mx], int iz);
void write_2dhistory_slice(float a[My][Mx], int iz);

float f_max(float a[My][Mx]); /* non-timestep utility routines */
float f_min(float a[My][Mx]); 
void node_timer_clear(int n);
void node_timer_start(int n);
void node_timer_stop(int n);
float node_timer_busy(int n);
void load_restart(char* resname);
void read_restart_slice(float a[My][Mx], FILE *tfile);
void save_restart();
void write_restart_slice(float a[My][Mx], FILE *tfile);
int iblock,N_Blocks,istep,N_Steps,N_File;
int history_rate,restart_rate,history_count,restart_count;
int overall_proc;

float max_u, max_v, max_t, max_p, max_q, max_s, min_s;
float max_an_t,max_an_s, max_sx, max_sy;
float max_tf, max_sf, max_tx, max_ty;

void ocean_init(overall_proc)
{
   int ix,iy,iz,i;
   int relfrq;
   int suffix; /* Will either be day or month since beginning of restart*/

   int ellen,elcnt,stride;

   
   N_File=1;

   N_Steps=((int)((hsfrq*24.0*3600.0/Delt)+0.499));
   N_Blocks=rnlng/hsfrq;


   fflush(stdout);

#ifndef COLOCATED_OCEAN
/* In the colocated case, only one buffer can be attached, and the one
used in the atmos model is kept because it is necessarily larger */
   MPI_Buffer_attach(srbuff, Mx*My*8*10); 
#endif

   ellen = sizeof(float); /* make mti type for e-w edge */
   elcnt = My;
   stride = Mx * ellen;
   MPI_Type_vector(elcnt, ellen, stride, MPI_CHAR, &mpi_edge_type);
   MPI_Type_commit(&mpi_edge_type);


   /* Get self address and number of nodes.*/

   fflush(stdout);

   MPI_Comm_rank(split_comm, &my_address);
   MPI_Comm_size(split_comm, &num_nodes);

#ifdef UPSHOT
   overall_proc=PP_NPROC_LT*PP_NPROC_LN+my_address+1;
#endif

   fflush(stdout);
   /* open stdout file for root node, flush on other nodes */


   if(my_address <= 0 )
   {
     sprintf(buff,"%som3.out.%d_%d.%d",prefix,Nrows,Ncols,my_address);
     if((of = freopen(buff, "w", stdout)) == NULL)  exit(89);
   }
   else
   {
     if((of = freopen("/dev/null", "w", stdout)) == NULL) exit(89);
   }
#ifndef COLOCATED_OCEAN
   sprintf(buff,"%som3.err.%d_%d.%d",prefix,Nrows,Ncols,my_address);
   if((ef = freopen(buff, "w", stderr)) == NULL)  exit(89); 
#endif


   /* open picture file */

#ifdef OMPICS 
   if(my_address == 0) {
# ifdef SIMPLE_OUTPUT
       suffix = finished + rsfrq;
       sprintf(buff,"history/ocean/picture/om3_%d_%d.%7.7d.cmpics",Nrows,Ncols,suffix);
# else
       sprintf(buff,"%som3_%d_%d.cmpics",prefix,Nrows,Ncols);
# endif
       ofile = global_open(buff, O_CREAT | O_WRONLY, 0666);
   }
#endif


   /* validate machine geometry */

   if(Np != num_nodes) {
     printf("Node number mismatch - needed %d got %d ...exiting\n",Np,num_nodes);
     exit(89);
   }
   else
     printf("Started on %d nodes.\n",num_nodes);
   fflush(of);


   /* Calculate node geometery and neighbor node numbers */

   if(my_address == 0)
      printf("Ncols,Nrows %d %d Mx,My: %d %d\n",Ncols,Nrows,Mx,My);

   row = my_address / Ncols;
   col = my_address % Ncols;


   mydisplace=my_address*(Mx-2)*(My-2);
   for(i=0;i<(Nrows*Ncols);i++) counts[i]= (Mx-2)*(My-2);
   MPI_Gather(&mydisplace,1,MPI_INT,displace,1,MPI_INT,0,split_comm);
   MPI_Gather(&row,1,MPI_INT,rows,1,MPI_INT,0,split_comm);
   MPI_Gather(&col,1,MPI_INT,cols,1,MPI_INT,0,split_comm);



#ifdef ANNUAL_SST
   read_sst_data();
#endif

   if(col == 0) west = my_address + (Ncols-1); /* 2-D torus */
   else west = my_address - 1;

   if(col == Ncols-1) east = my_address - (Ncols-1);
   else east = my_address + 1;

   if(row == 0) south = my_address + (Nrows-1)*Ncols;
   else south = my_address - Ncols;

   if(row == Nrows-1)  north = my_address - (Nrows-1)*Ncols;
   else north = my_address + Ncols;

   /* Initialization processing */

   node_timer_clear(0);
   node_timer_start(0);

   read_geometry();
   read_stress();
   read_annual_temp();
   read_annual_salt();

   node_timer_stop(0);

   if(my_address == 0){
      printf("Forcing and geometry read time=%f [secs]\n",node_timer_busy(0));
      printf("Max depth = %f \n",maxdepth);}

#ifdef OMPICS
   write_picture(strx,0.1);
   write_picture(stry,0.1);
   write_picture(&(annual_temp[Nz-1][0][0]),10.0);
   write_picture(&(annual_salt[Nz-1][0][0]),2.0);
#endif

   max_sx = f_max(strx);
   max_sy = f_max(stry);
   max_an_t = f_max(&(annual_temp[Nz-1][0]));
   max_an_s = f_max(&(annual_salt[Nz-1][0]));

   if(my_address == 0)
   {
	 printf("Max sx: %f, sy: %f, ann. temp, %f, ann. salt %f\n",
            max_sx,max_sy,max_an_t,max_an_s);
   }
   fflush(of);

   node_timer_clear(0);
   node_timer_start(0);

   init_state();

   node_timer_stop(0);

   if(my_address == 0)
      printf("State init time=%f [secs] ,\n",node_timer_busy(0));
   fflush(of);

#ifdef RestartRun
# ifdef OCEAN_ALONE
   sprintf(resname,"restart/restart.ocean.%7.7d", finished);
# else
   sprintf(resname,"restart/ocean/restart.ocean.%7.7d", finished);
# endif
   load_restart(resname);
   if(my_address == 0)
   {
     printf("Model initialized from restart file: %s\n",resname);
   }
   fflush(of);
#endif

   max_u = f_max(&(u[Nz-1][0]));
   max_v = f_max(&(v[Nz-1][0]));
   max_t = f_max(&(t[Nz-1][0]));
   max_p = f_max(&(ptop[0]));
   max_q = f_max(&(qflx[0]));

   if(my_address == 0)
   {
      printf("Max u: %f, v: %f, t: %f, p: %f qflx: %e\n",max_u,max_v,max_t,max_p, max_q);
      printf("Delt = %f seconds, Block time = %f days.\n",Delt,Delt*N_Steps/(3600.0*24.0));
   }
   fflush(of);

   istep = 0;
   restart_rate = rsfrq * (int)(3600.*24./Delt);
   history_rate = hsfrq * (int)(3600.*24./Delt);
   restart_count = 0;
   history_count = 0;

} /* This is the end of ocean_init */

#ifndef COLOCATED_OCEAN
void ocean_step()
#else
# ifdef RS6000
void ocean_step()
# else
#  ifdef T3E
void OCEAN_STEP()
#  else
void ocean_step_()
#  endif
# endif
#endif
{
   int hiswrite = 0;
   int reswrite = 0;
   int suffix;
   int ix,iy,iz;

   istep = istep + 1;

/* This section tests whether it is time for history and restart writes */
  if ((istep % restart_rate) == 0)
   {
     restart_count++;
     reswrite = restart_count;
   }

   if ((istep % history_rate) == 0)
   {
     history_count++;
     hiswrite = history_count;
   }


   node_timer_clear(0);
   node_timer_start(0);
/*   for(iblock = 0; iblock < N_Blocks; iblock++)
   {
      for(istep = 0; istep < N_Steps; istep++)
      { */
#ifdef PROFILE
       mark_event_C(3,1);
#endif
#ifndef OCEAN_ALONE
#ifdef UPSHOT
        MPE_Log_event(6,overall_proc,"EndOcn");
        MPE_Log_event(9,overall_proc,"StrtRecv");
#endif
        ocean_from_coupler(4,from_coupler,200);

#ifdef UPSHOT
        MPE_Log_event(10,overall_proc,"EndRecv");
        MPE_Log_event(5,overall_proc,"StrtOcn");
#endif
#endif

/* Compute forcing parameters here. */
        for(iy = 1; iy < My-1; iy++)
        {
           for(ix = 1; ix < Mx-1; ix++)
           {
                  if(h[Nz-1][iy][ix] > 1.0)
                  {
#ifdef OCEAN_ALONE
                     temp_for[iy][ix] = 0.;
                     salt_for[iy][ix] = 0.;
#else
                     temp_for[iy][ix] = from_coupler[0][iy-1][ix-1]/(rhoref * h[Nz-1][iy][ix]* Cp);
                     salt_for[iy][ix] = -s[Nz-1][iy][ix]*from_coupler[1][iy-1][ix-1]/(rhoref * h[Nz-1][iy][ix]);
                     taux[iy][ix] = from_coupler[2][iy-1][ix-1]/(rhoref * h[Nz-1][iy][ix]);
                     tauy[iy][ix] = from_coupler[3][iy-1][ix-1]/(rhoref * h[Nz-1][iy][ix]);
#endif
                  }
                  else
                  {
                     taux[iy][ix] = 0.0;
                     tauy[iy][ix] = 0.0;
                  }
           }
        }

        wrap_q(taux);
        wrap_q(tauy);
        wrap_q(temp_for);
        wrap_q(salt_for);

       max_tf = f_max(&(temp_for[0]));
       max_sf = f_max(&(salt_for[0]));
       max_tx = f_max(&(taux[0]));
       max_ty = f_max(&(tauy[0]));

       if(my_address == 0)
       {
             printf("istep %d Max tempf: %e, saltf: %e, taux: %e, tauy: %e\n",istep,max_tf,max_sf,max_tx,max_ty);
       }
       fflush(of);

#ifdef OMPICS_EX
         write_picture(&(temp_for[0][0]),fabs(max_tf));
         write_picture(&(salt_for[0][0]),fabs(max_sf));
         write_picture(&(taux[0][0]),fabs(max_tx));
         write_picture(&(tauy[0][0]),fabs(max_ty));
#endif

#ifndef OCEAN_ALONE
/* Return ocean state back to coupler */
        for(iy = 1; iy < My-1; iy++)
        {
           for(ix = 1; ix < Mx-1; ix++)
           {
              to_coupler[0][iy-1][ix-1] = t[Nz-1][iy][ix]+ CtoK;/* SST */
              to_coupler[1][iy-1][ix-1] = qflx[iy][ix]; /* melt potential */
           }
        }

#ifdef UPSHOT
        MPE_Log_event(6,overall_proc,"EndOcn");
        MPE_Log_event(7,overall_proc,"StrtSend");
#endif
      
        ocean_to_coupler(2,to_coupler,300);

#ifdef UPSHOT
        MPE_Log_event(8,overall_proc,"EndSend");
        MPE_Log_event(5,overall_proc,"StrtOcn");
#endif
#endif

/* zero qice and qflx */
   for (iz =  0; iz < Mz; iz++)
   for(iy = 0; iy < My; iy++)
    for(ix = 0; ix < Mx; ix++)
     qice[iz][iy][ix]=0.;

   for(iy = 0; iy < My; iy++)
    for(ix = 0; ix < Mx; ix++)
     qflx[iy][ix]=0.;

	    time_step();
	    fflush(of);
	    fflush(ef);

       max_t = f_max(&(t[Nz-1][0]));
       max_s = f_max(&(s[Nz-1][0]));
       min_s = f_min(&(s[Nz-1][0]));
       max_u = f_max(&(u[Nz-1][0]));
       max_v = f_max(&(v[Nz-1][0]));
       if(my_address == 0)
             printf("istep %d Max t: %e, s: %e, u: %e, v: %e Min s: %e\n",istep,max_t,max_s,max_u,max_v,min_s);

#ifdef PROFILE
       mark_event_C(3,0);
#endif
   for (iz =  0; iz < Mz; iz++)
   {
     for (iy = 0; iy < My; iy++)
     {
      for (ix = 0; ix < Mx; ix++)
      {
      tav[iz][iy][ix] += t[iz][iy][ix];
      sav[iz][iy][ix] += s[iz][iy][ix];
      uav[iz][iy][ix] += u[iz][iy][ix];
      vav[iz][iy][ix] += v[iz][iy][ix];
      wav[iz][iy][ix] += w[iz][iy][ix];
      pav[iz][iy][ix] += p[iz][iy][ix];
      }
      }
      }
     for (iy = 0; iy < My; iy++)
     {
      for (ix = 0; ix < Mx; ix++)
      {
      }
      }

   /*   }  This is the end of the NStep loop */

  if(hiswrite > 0) {
      node_timer_stop(0);
      if(my_address == 0)
	    printf("Timestep calculation time=%f [secs],\n",node_timer_busy(0));
      node_timer_clear(0);
      fflush(of);
      fflush(ef);

      max_u = f_max(&(u[Nz-1][0]));
      max_v = f_max(&(v[Nz-1][0]));
      max_t = f_max(&(t[Nz-1][0]));
      max_p = f_max(&(ptop[0]));

      if(my_address == 0)
      {
	    printf("Max u: %f, v: %f, t: %f, p: %f\n",max_u,max_v,max_t,max_p);
      }
      fflush(of);

#ifdef OMPICS
      max_tf = f_max(&(t_forcing[0][0]));
      max_sf = f_max(&(s_forcing[0][0]));
      max_tx = f_max(&(taux_for[0][0]));
      max_ty = f_max(&(tauy_for[0][0]));

           write_picture(&(t_forcing[0][0]),fabs(max_tf)/3.0);
           write_picture(&(s_forcing[0][0]),fabs(max_sf)/3.0);
           write_picture(&(taux_for[0][0]),fabs(max_tx));
           write_picture(&(tauy_for[0][0]),fabs(max_ty));
#endif



#ifdef PROFILE
       mark_event_C(8,1);
#endif


     suffix = finished + hsfrq * hiswrite;
/* Write surface history file */
     if(my_address == 0) {

#ifdef SIMPLE_OUTPUT
        sprintf(hisname,"history.osurf.%7.7d", suffix);
# ifdef OCEAN_ALONE
        sprintf(buff,"history/%s",hisname);
# else
        sprintf(buff,"history/ocean/%s",hisname);
# endif
#else
        sprintf(buff,"%shistory.ocean.%7.7d",prefix,suffix);
#endif

        hfile = fopen(buff,"wb");

/* there's always one time slice per file */
        fprintf(hfile,"%d\n",N_File);
        printf("Opening Surface History File: %s\n",buff);
     }

        for (iy = 1; iy < My-1; iy++)
 	{
   	 for (ix = 1; ix < Mx-1; ix++)
   	  {
	  avtaux[iy][ix]=avtaux[iy][ix]/(float)N_Steps;
	  avtauy[iy][ix]=avtauy[iy][ix]/(float)N_Steps;
	  avheat[iy][ix]=avheat[iy][ix]/(float)N_Steps;
	  avsalt[iy][ix]=avsalt[iy][ix]/(float)N_Steps;

	  diag6[0][Nz-1][iy][ix]=diag6[0][Nz-1][iy][ix]/((float)N_Steps);
	  diag6[0][Nz-2][iy][ix]=diag6[0][Nz-2][iy][ix]/((float)N_Steps);
	  diag6[1][Nz-1][iy][ix]=diag6[1][Nz-1][iy][ix]/((float)N_Steps);
	  diag6[1][Nz-2][iy][ix]=diag6[1][Nz-2][iy][ix]/((float)N_Steps);

	  diag7[0][Nz-1][iy][ix]=diag7[0][Nz-1][iy][ix]/((float)N_Steps);
	  diag7[0][Nz-2][iy][ix]=diag7[0][Nz-2][iy][ix]/((float)N_Steps);
	  diag7[1][Nz-1][iy][ix]=diag7[1][Nz-1][iy][ix]/((float)N_Steps);
	  diag7[1][Nz-2][iy][ix]=diag7[1][Nz-2][iy][ix]/((float)N_Steps);

	  }
        }
        node_timer_clear(0);
        node_timer_start(0);

        write_2dhistory_slice(avtaux,Nz-1);
        write_2dhistory_slice(avtauy,Nz-1);
        write_2dhistory_slice(avheat,Nz-1);
        write_2dhistory_slice(avsalt,Nz-1);
        write_2dhistory_slice(convec_freq,Nz-1);
        write_2dhistory_slice(convec_freq2,Nz-1);
        write_2dhistory_slice(&(diag6[0][Nz-1][0]),Nz-1);
        write_2dhistory_slice(&(diag6[0][Nz-2][0]),Nz-2);
        write_2dhistory_slice(&(diag6[1][Nz-1][0]),Nz-1);
        write_2dhistory_slice(&(diag6[1][Nz-2][0]),Nz-2);
        write_2dhistory_slice(&(diag7[0][Nz-1][0]),Nz-1);
        write_2dhistory_slice(&(diag7[0][Nz-2][0]),Nz-2);
        write_2dhistory_slice(&(diag7[1][Nz-1][0]),Nz-1);
        write_2dhistory_slice(&(diag7[1][Nz-2][0]),Nz-2);

        node_timer_stop(0);
        for (iy = 0; iy < My; iy++)
 	{
   	 for (ix = 0; ix < Mx; ix++)
   	  {
	  avtaux[iy][ix]= 0.;
	  avtauy[iy][ix]=0.;
	  avheat[iy][ix]= 0.;
	  avsalt[iy][ix]=0.;
	  convec_freq[iy][ix]=0.;
	  convec_freq2[iy][ix]=0.;
	  }
        }
      for (iz =  0; iz < Mz; iz++)
        for (iy = 0; iy < My; iy++)
         for (ix = 0; ix < Mx; ix++){
          diag6[0][iz][iy][ix]=0.;
          diag6[1][iz][iy][ix]=0.;
          diag7[0][iz][iy][ix]=0.;
          diag7[1][iz][iy][ix]=0.;
          /*diag0[0][iz][iy][ix]=0.;
          diag0[1][iz][iy][ix]=0.;
          diag8[0][iz][iy][ix]=0.;
          diag8[1][iz][iy][ix]=0.;*/
         }

      if (my_address==0)
      {
        fclose(hfile);
      }

/* write full history */
        if(my_address == 0) {

#ifdef SIMPLE_OUTPUT
        sprintf(hisname,"history.ocean.%7.7d", suffix);
# ifdef OCEAN_ALONE
        sprintf(buff,"history/%s",hisname);
# else
        sprintf(buff,"history/ocean/%s",hisname);
# endif
#else
        sprintf(buff,"%shistory.ocean.%7.7d",prefix,suffix);
#endif

        hfile = fopen(buff,"wb");
        fprintf(hfile,"%d\n",N_File);
        printf("Opening Full History File: %s\n",buff);
        }
        for (iz =  0; iz < Mz; iz++)
   	{
     	for (iy = 0; iy < My; iy++)
     	{
      	for (ix = 0; ix < Mx; ix++)
      	{
      	tav[iz][iy][ix] = tav[iz][iy][ix]/(N_Steps);
      	sav[iz][iy][ix] = sav[iz][iy][ix]/(N_Steps);
      	uav[iz][iy][ix] = uav[iz][iy][ix]/(N_Steps);
      	vav[iz][iy][ix] = vav[iz][iy][ix]/(N_Steps);
      	wav[iz][iy][ix] = wav[iz][iy][ix]/(N_Steps);
      	pav[iz][iy][ix] = pav[iz][iy][ix]/(N_Steps);
/*      	pav[iz][iy][ix] = pav[iz][iy][ix]/(N_Steps*relfrq);*/
      	}
      	}
      	}
        node_timer_start(0);
        write_history(uav);
        write_history(vav);
        write_history(wav);
        write_history(tav);
        write_history(sav);
        write_history(pav);
/*
        write_history(diag1);
        write_history(diag2);
        write_history(diag3);
        write_history(diag4);
        write_history(diag5);
*/
        node_timer_stop(0);
        for (iz =  0; iz < Mz; iz++)
   	{
     	for (iy = 0; iy < My; iy++)
     	{
      	for (ix = 0; ix < Mx; ix++)
      	{
      	tav[iz][iy][ix] = 0.;
      	sav[iz][iy][ix] = 0.;
      	uav[iz][iy][ix] = 0.;
      	vav[iz][iy][ix] = 0.;
      	wav[iz][iy][ix] = 0.;
      	pav[iz][iy][ix] = 0.;
      	}
      	}
      	}
     /*}*/
      
        if(my_address == 0)
	    printf("history write time=%f [secs]\n",node_timer_busy(0));


      if (my_address==0)
      {
        fclose(hfile);
      }

#ifdef PROFILE
        mark_event_C(8,0);
#endif
  } /* End of if history write */

      if(reswrite > 0)
      {
#ifdef RestartSave
#ifdef PROFILE
      mark_event_C(9,1);
#endif
        suffix = finished + rsfrq * reswrite;
        node_timer_clear(0);
        node_timer_start(0);

        save_restart(suffix);

        node_timer_stop(0);
        if(my_address == 0){
	    printf("Restart write time=%f [secs]\n",node_timer_busy(0));
            fflush(of);}

#ifdef OCEAN_ALONE
        sprintf(buff,"%s/Restart.%7.7d",prefix,suffix);
        checkfile = fopen(buff,"w");
        fclose(checkfile);
#endif

#ifdef OMPICS
   if(my_address==0)
   {
      close(ofile);
      if (iblock < (N_Blocks - 1)) 
      {
#ifdef SIMPLE_OUTPUT
         suffix = suffix + rsfrq;
         sprintf(buff,"history/ocean/picture/om3_%d_%d.%7.7d.cmpics",Nrows,Ncols,suffix);
#else
         sprintf(buff,"%som3_%d_%d.cmpics",prefix,Nrows,Ncols);
#endif
         ofile = global_open(buff, O_CREAT | O_WRONLY, 0666);

      }
   }
#endif

#ifdef PROFILE
      mark_event_C(9,0);
#endif
#endif
      } /* End of if restart write */

 /*  }  End of the N_Blocks loop */

} /* This is the end of ocean_step */

void ocean_end()
{

fflush(of);
fflush(ef);

}



float f_max(a)
float a[My][Mx];
{
   int ix,iy,maxix,maxiy;
   float max,gmax,gmin,maxamp;

   maxamp = -1.0;

   for(iy = 1; iy < My-1; iy++)
   {
      for(ix = 1; ix < Mx-1; ix++)
      {
	     if((fabs(a[iy][ix]) > maxamp) &&(h[Nz-1][iy][ix] > 1.0))
         {
             max = a[iy][ix];
             maxamp = fabs(a[iy][ix]);
             maxix = ix;
             maxiy = iy;
         }
      }
   }
/*
   if(my_address == 0) printf("Max ix,iy: %d %d %f\n",maxix,maxiy,max);
*/

   MPI_Allreduce(&max, &gmax, 1, MPI_FLOAT, MPI_MAX, split_comm);
   MPI_Allreduce(&max, &gmin, 1, MPI_FLOAT, MPI_MIN, split_comm);



   if(gmax > -gmin) return gmax;
   else return gmin;
}

float f_min(a)
float a[My][Mx];
{
   int ix,iy,minix,miniy;
   float min,gmin,maxamp;

   maxamp = 1000.;;

   for(iy = 1; iy < My-1; iy++)
   {
      for(ix = 1; ix < Mx-1; ix++)
      {
	     if((a[iy][ix] < maxamp) &&(h[Nz-1][iy][ix] > 1.0))
         {
             min = a[iy][ix];
	     maxamp=min;
             minix = ix;
             miniy = iy;
         }
      }
   }
/*
   if(my_address == 0) printf("Max ix,iy: %d %d %f\n",maxix,maxiy,max);
*/

   MPI_Allreduce(&min, &gmin, 1, MPI_FLOAT, MPI_MIN, split_comm);

   return gmin;
}

float f_total(a)
float a[My][Mx];
{
   int ix,iy;
   float total,gtotal;

   total = 0.0;

   for(iy = 1; iy < My-1; iy++)
   {
      for(ix = 1; ix < Mx-1; ix++)
      {
         {
            if(h[Nz-1][iy][ix] > 1.0)
              total = total + a[iy][ix];
         }
      }
   }
   MPI_Allreduce(&total,&gtotal,1,MPI_FLOAT, MPI_SUM, split_comm);


   return gtotal;
}

#ifdef OMPICS

/*
 * We write a picture one node at a time in cm5 order.
 * In this (non piofs) version each write is staged throught the
 * root node.
 */
void write_picture(float a[My][Mx],float vmax)
{
   int ix,iy;
   int len,pixel_val;
   float val,scl;
   char *bptr,buf[200];

   fflush(of);

   /* write header with Nx, Ny, Ncols, Nrows */

   sprintf(buf,"%d %d %d %d\n",Nx,Ny,Ncols,Nrows);
   len = strlen(buf);
   if(my_address == 0) write(ofile,buf,len);

   scl = 40.0/(vmax);

   for(iy = 1; iy < My-1; iy++)
   {
     for(ix = 1; ix < Mx-1; ix++) /* fill buffer in parallel */
     {
	   val = scl * a[iy][ix];
	   pixel_val = ((int) fabs(val+0.5)) % 40;

	   if(val > 0)
         pixel_val = 41 + pixel_val; /* red colors in std map */
	   else
         pixel_val = 81 + pixel_val; /* blue colrs in std map */

	   if(h[Nz-1][iy][ix] > 1.0)
         picbuf[iy-1][ix-1] = pixel_val;
       else
         picbuf[iy-1][ix-1] = 0; /* zero out dry cells */
    }
  }
  global_write();
}

void fwrite_picture(float a[My][Mx],float vmax)
{
   int ix,iy;
   int len,pixel_val;
   float val,scl;
   char *bptr,buf[200];

   /* write header with Nx, Ny, Ncols, Nrows */
   sprintf(buf,"%d %d %d %d\n",Nx,Ny,Ncols,Nrows);
   len = strlen(buf);
   if(my_address == 0) write(ofile,buf,len);

   scl = 40.0/(vmax);

   for(iy = 1; iy < My; iy++)
   {
      for(ix = 1; ix < Mx-1; ix++) /* fill buffer in parallel */
      {
        if(h[Nz-1][iy][ix+1] > 1.0)
	      work[iy][ix] = 0.5 * (a[iy][ix] + a[iy][ix+1]);
        else
          work[iy][ix] = a[iy][ix];
      }
   }

   for(iy = 1; iy < My-1; iy++)
   {
     for(ix = 1; ix < Mx-1; ix++) /* fill buffer in parallel */
     {
       if(h[Nz-1][iy+1][ix] > 1.0)
	     val = scl * 0.5 * (work[iy][ix] + work[iy+1][ix]);
       else
         val = scl * work[iy][ix];
	 
	   pixel_val = ((int) fabs(val+0.5)) % 40;

	   if(val > 0)
         pixel_val = 41 + pixel_val; /* red colors in std map */
	   else
         pixel_val = 81 + pixel_val; /* blue colrs in std map */

	   if(h[Nz-1][iy][ix] > 1.0)
         picbuf[iy-1][ix-1] = pixel_val;
       else
         picbuf[iy-1][ix-1] = 0; /* zero out dry cells */
     }
  }
  global_write();
}

global_write()
{
   int i, j, k, l, ik, jl, ierr;
   int tt,number;
   unsigned char t1, t2;
   MPI_Status status;



   if(my_address == 0) {
       write(ofile, picbuf, (Mx-2)*(My-2));
       for(i=0; i<Nrows; i++)
       {
	     for(j=0; j<Ncols; j++)
         {
          if (i != 0 || j!= 0) 
          {
            MPI_Recv(picbuf, (Mx-2)*(My-2), MPI_CHAR, i*Ncols+j,
		      99, split_comm, &status);
/*            MPI_Get_count(&status,MPI_CHAR,&number);
            printf("Source:%d\n",status.MPI_SOURCE);
            printf("Tag:%d\n",status.MPI_TAG);
            printf("Size:%d\n",number);
            printf("Receiving:8192\n"); */
       	    write(ofile, picbuf, (Mx-2)*(My-2));
          }
	     }
       }
   }
   else
   {
     ierr = MPI_Send(picbuf, (Mx-2)*(My-2), MPI_CHAR, 0, 99, split_comm);
/*     printf("Sending:8192\n"); */
     if(ierr != 0) {
       printf("Bsend error %d\n", ierr);
       exit(1);
     }
   }
}
#endif

/* History write routines */

void write_history(float a[Mz][My][Mx])
{
  int iz;

  for(iz = 0; iz < Mz; iz++)
  {
     write_history_slice(a,iz);
  }
}

unsigned char hist_out[My][Mx][2];

void write_history_slice(float a[Mz][My][Mx], int iz)
{
  int ix,iy,ival,ic,i,j;
  float xx,lxmin,lxmax,xmin,xmax,scale;
  int jrow,jcol,jy;
  unsigned char awhole1[Ny*Nx];
  unsigned char awhole2[Ny*Nx];
  unsigned char whist_out[Ny][Nx][2];
  unsigned char aprt1[(My-2)*(Mx-2)];
  unsigned char aprt2[(My-2)*(Mx-2)];

  unsigned char row_out[Mx-2][2];
  MPI_Status status;
  int ierr;


  lxmin = 1.0e10;
  lxmax = -1.0e10;

  for(iy = 1; iy < My-1; iy++)
  {
    for (ix = 1; ix < Mx-1; ix++)
    {
      if(h[iz][iy][ix] > 1.0)
      {
         xx = a[iz][iy][ix];
         if(xx < lxmin) lxmin = xx;
         if(xx > lxmax) lxmax = xx;
      }
    }
  }


   MPI_Allreduce(&lxmax, &xmax, 1, MPI_FLOAT, MPI_MAX, split_comm);
   MPI_Allreduce(&lxmin, &xmin, 1, MPI_FLOAT, MPI_MIN, split_comm);



  if(my_address == 0)
  {
     fprintf(hfile,"%f %f\n",xmin,xmax);
  }
  xx = xmax-xmin;
  if(xx < 1.0e-10) xx=1.0e-10;
  scale = 60000.0/xx;

  i=0;
  for(iy = 1; iy < My-1; iy++)
  {
    for (ix = 1; ix < Mx-1; ix++)
    {
      if(h[iz][iy][ix] > 1.0)
      {
         xx = a[iz][iy][ix];
         ival = 100 + scale*(xx-xmin);
      }
      else ival = 0;

      hist_out[iy][ix][0] = ival/256;
      hist_out[iy][ix][1] = ival % 256;
      aprt1[i]=hist_out[iy][ix][0];
      aprt2[i]=hist_out[iy][ix][1];
      i++;
    }
  }

   ierr=MPI_Gatherv(aprt1,(Mx-2)*(My-2),MPI_CHAR,awhole1,counts,displace,
        MPI_CHAR,0,split_comm);
   ierr=MPI_Gatherv(aprt2,(Mx-2)*(My-2),MPI_CHAR,awhole2,counts,displace,
        MPI_CHAR,0,split_comm);

  if(my_address == 0)
  {

   for(i = 0; i < Nrows*Ncols; i++)
   for(j = 0; j < (Mx-2)*(My-2); j++){
     iy=rows[i]*(My-2) + (j / (Mx-2));
     ix=cols[i]*(Mx-2) + (j % (Mx-2));
     /*fprintf(stderr,"All %d %d %d %d %d\n",i,j,iy,ix,(displace[i])+j);*/
     whist_out[iy][ix][0]=awhole1[(displace[i])+j];
     whist_out[iy][ix][1]=awhole2[(displace[i])+j];
     }

     for(iy = 0; iy < Ny; iy++)
           fwrite(&(whist_out[iy][0][0]),sizeof(char),2*Nx,hfile);

  }
  MPI_Barrier(split_comm);

}

void write_2dhistory_slice(float a[My][Mx], int iz)
{
  int ix,iy,ic,i,j;
  int jrow,jcol,jy;
  float awhole[Ny*Nx];
  float awhole2[Ny][Nx];
  float aprt[(My-2)*(Mx-2)];

  MPI_Status status;
  int ierr;

   i=0;
   for(iy = 1; iy < My-1; iy++)
   for(ix = 1; ix < Mx-1; ix++){
      if(h[iz][iy][ix] > 1.0) aprt[i]=a[iy][ix];
      else aprt[i]=MISSING;
      i++;
      }



   ierr=MPI_Gatherv(aprt,(Mx-2)*(My-2),MPI_FLOAT,awhole,counts,displace,
        MPI_FLOAT,0,split_comm);




  if(my_address == 0)
  {

   for(i = 0; i < Nrows*Ncols; i++)
   for(j = 0; j < (Mx-2)*(My-2); j++){
     iy=rows[i]*(My-2) + (j / (Mx-2));
     ix=cols[i]*(Mx-2) + (j % (Mx-2));
     /*fprintf(stderr,"All %d %d %d %d %d\n",i,j,iy,ix,(displace[i])+j);*/
     awhole2[iy][ix]=awhole[(displace[i])+j];
     }

     for(iy = 0; iy < Ny; iy++)
           fwrite(&(awhole2[iy]),sizeof(float),Nx,hfile);

  }
  MPI_Barrier(split_comm);

}

void read_geometry()
{
   int ix,iy,jx,jy,jz,iz;
   FILE *tfile;
   int tt;
   float lmaxdepth;
   unsigned char in[2*Nx], *inptr, t1, t2;

   if(my_address == 0)
   {
      sprintf(buff,"%s%s%d",time_inv,"/om3.geometry",Nz);
      printf("Opening geometry file %s\n",buff);
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on geometry failed. %s\n",buff);
      exit(89);
   }

   /* read depths of two types of grid levels */

   if(my_address == 0) fread(in,sizeof(char),2*Nz,tfile);
   MPI_Bcast(in,2*Nz,MPI_CHAR,0,split_comm);
   inptr = in;
   for(jz = 0; jz < Nz; jz++)
   {
      t1 = *inptr++; t2 = *inptr++;
      tt = t1*256 + t2;
      uvts_levels[jz] = tt;
   }

   if(my_address == 0) fread(in,sizeof(char),2*Nz,tfile);
   MPI_Bcast(in,2*Nz,MPI_CHAR,0,split_comm);
   inptr = in;
   for(jz = 0; jz < Nz; jz++)
   {
      t1 = *inptr++; t2 = *inptr++;
      tt = t1*256 + t2;
      w_levels[jz] = tt;
   }

   for(jz = Nz-1; jz >= 0; jz--) /* read 3-D topo field */
   {
     for(jy = 0; jy < Ny; jy++)
     {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0) fread(in,sizeof(char),2*Nx,tfile);
      MPI_Bcast(in,2*Nx,MPI_CHAR,0,split_comm);
      inptr = in;
      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
	     t1 = *inptr++; t2 = *inptr++;
         tt = t1*256 + t2;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
         {
            h[jz][iy][ix] = (float)tt;
         }
	 hw[jz][jy][jx] = (float)tt;
      }
     }
   }

   if(my_address == 0) fclose(tfile);

   wrap_qz(h);

/* make sure the north pole sees land to north and south pole
  sees land to south to prevent interaction from pole to pole
  caused by the torus communication topology */

/* see if this processor is handling the north pole. */
   if((My-2)*(row+1)==Ny){
    for(iz = 0; iz < Mz; iz++)
     for(ix = 0; ix < Mx; ix++)
      h[iz][My-1][ix]=0.;
   }

/* see if this processor is handling the south pole. */
   if(row==0){
    for(iz = 0; iz < Mz; iz++)
     for(ix = 0; ix < Mx; ix++)
      h[iz][0][ix]=0.;
   }

   lmaxdepth=0;

   for(iy = 1; iy < My-1; iy++) /* fill depth and reciprical depth arrays */
   {
     for(ix = 1; ix < Mx-1; ix++)
     {
       depth[iy][ix] = 0.0;
       uvdepth[iy][ix] = 0.0;
       for(jz = 0; jz < Nz; jz++)
       {
          depth[iy][ix] += h[jz][iy][ix];
	  if(depth[iy][ix] > lmaxdepth)lmaxdepth=depth[iy][ix];

	      if((h[jz][iy][ix+1] > 1.0) && (h[jz][iy][ix-1] > 1.0)
          &&(h[jz][iy+1][ix] > 1.0) && (h[jz][iy-1][ix] > 1.0))
          {
             uvdepth[iy][ix] += h[jz][iy][ix];
             uvh[jz][iy][ix] = h[jz][iy][ix];
          }
          else
          {
             uvh[jz][iy][ix] = 0.0;
          }
       }
       if(depth[iy][ix] > 1.0)
         rdepth[iy][ix] = 1.0/depth[iy][ix];
       else
         rdepth[iy][ix] = 0.0;

       if(uvdepth[iy][ix] > 1.0)
         ruvdepth[iy][ix] = 1.0/uvdepth[iy][ix];
       else
         ruvdepth[iy][ix] = 0.0;
     }
   }
   MPI_Allreduce(&lmaxdepth, &maxdepth, 1, MPI_FLOAT, MPI_MAX, split_comm);
   wrap_q(depth);
   wrap_q(uvdepth);
   wrap_q(rdepth);
   wrap_q(ruvdepth);
   wrap_qz(uvh);
}

void read_stress()
{
   int ix,iy,jx,jy;
   FILE *tfile;
   int tt;
   unsigned char in[2*Nx], *inptr, t1, t2;

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",time_inv,"/om3.stress");
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on stress failed.\n");
      exit(89);
   }


   for(jy = 0; jy < Ny; jy++) /* read x-stress field */
   {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0) fread(in,sizeof(char),2*Nx,tfile);
      MPI_Bcast(in,2*Nx,MPI_CHAR,0,split_comm);
      inptr = in;

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
         t1 = *inptr++; t2 = *inptr++;
         tt = t1*256 + t2;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
            strx[iy][ix] = ((float)tt/10000.0)-0.3;
      }
   }

   for(jy = 0; jy < Ny; jy++) /* read y-stress field */
   {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0) fread(in,sizeof(char),2*Nx,tfile);
      MPI_Bcast(in,2*Nx,MPI_CHAR,0,split_comm);
      inptr = in;

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
	     t1 = *inptr++; t2 = *inptr++;
         tt = t1*256 + t2;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
            stry[iy][ix] = ((float)tt/10000.0)-0.3;
      }
   }
   if(my_address == 0) fclose(tfile);
}

void read_annual_temp()
{
   int ix,iy,jx,jy,jz;
   FILE *tfile;
   int tt;
   unsigned char in[2*Nx], *inptr, t1, t2;
   float flin[Nx];

   if(my_address == 0)
   {
      sprintf(buff,"%s%s%d",time_inv,"/om3.temp",Nz);
      printf("Opening initial temp field %s\n",buff);
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on temp failed %s.\n",buff);
      exit(89);
   }

   for(jz = Nz-1; jz >= 0; jz--) /* read 3-D temp field */
   {
     for(jy = 0; jy < Ny; jy++)
     {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0)
      { float fxx;
	int tcount;
        fread(in,sizeof(char),2*Nx,tfile);
        inptr = in;
        fxx=0.0;
	tcount=0;
        for(jx = 0; jx < Nx; jx++)
        {
          t1 = *inptr++; t2 = *inptr++;
          tt  = t1*256 + t2;
          flin[jx] =((float)tt-10000.0)/100.0;
	  if(hw[jz][jy][jx] > 1.0)
	  {
          fxx += flin[jx];
	  tcount++;
	  }
        }
	if(tcount > 0)
        fxx = fxx/tcount;
        /*for(jx = 0; jx < Nx; jx++) flin[jx]=fxx;*/
        if(jy == 85) {
	base_t[jz] = flin[0];
	if(flin[0] == 0.) printf("ERROR in base_t\n");
			   
        }
      }

    MPI_Bcast(flin,Nx,MPI_FLOAT,0,split_comm);

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
         {
            annual_temp[jz][iy][ix] = flin[jx];
         }
      }
     }
   }
   if(my_address == 0) fclose(tfile);
 MPI_Bcast(base_t,Mz,MPI_FLOAT,0,split_comm);
   wrap_qz(annual_temp);
}

void read_annual_salt()
{
   int ix,iy,jx,jy,jz;
   FILE *tfile;
   int tt;
   unsigned char in[2*Nx], *inptr, t1, t2;
   float flin[Nx];

   if(my_address == 0)
   {
      sprintf(buff,"%s%s%d",time_inv,"/om3.salt",Nz);
      printf("Opening initial salt field %s\n",buff);
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on salt failed %s.\n",buff);
      exit(89);
   }

   for(jz = Nz-1; jz >= 0; jz--) /* read 3-D salinity field */
   {
     for(jy = 0; jy < Ny; jy++)
     {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0)
      { float fxx;
	int scount;
        fread(in,sizeof(char),2*Nx,tfile);
        inptr = in;
        fxx=0.0;
	scount=0;
        for(jx = 0; jx < Nx; jx++)
        {
          t1 = *inptr++; t2 = *inptr++;
          tt  = t1*256 + t2;
          flin[jx] =((float)tt-10000.0)/100.0;
	  if(hw[jz][jy][jx] > 1.0)
	  {
          fxx += flin[jx];
	  scount++;
	  }
        }
	if(scount > 0)
        fxx = fxx/scount;
        /*for(jx = 0; jx < Nx; jx++) flin[jx]=fxx;*/
        if(jy == 85) {
	base_s[jz] = flin[0];
	if(flin[0] == 0.) printf("ERROR in base_s\n");
			   
        }
      }

    MPI_Bcast(flin,Nx,MPI_FLOAT,0,split_comm);

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
         {
            annual_salt[jz][iy][ix] = flin[jx];
         }
      }
     }
   }
   if(my_address == 0) fclose(tfile);
 MPI_Bcast(base_s,Mz,MPI_FLOAT,0,split_comm);
   wrap_qz(annual_salt);
}

void init_state()
{
   int ix,iy,iz,jx,jy,mx,my,it;
   float x,y,yeq,lat,omega;

   Lathf = 333600.;  /* latent heat of fusion */
   Cp = 4.1e3;    /* specific heat of seawater */
   Tf = -1.8;  /* freezing point of seawater */
   Sref = 34.7;  /* reference salinity */
   rhoref = 1030.;  /* reference density */

   CtoK = 273.15;  /* celcius to kelvin */

   omega = (366.0/365.0) * 2.0*3.14159 / (3600.0*24.0);

/* initialize diagnostic fields */
    for (iz =  0; iz < Mz; iz++)
      for (iy = 0; iy < My; iy++)
       for (ix = 0; ix < Mx; ix++){
       diag6[0][iz][iy][ix]=0.;
       diag6[1][iz][iy][ix]=0.;
       diag7[0][iz][iy][ix]=0.;
       diag7[1][iz][iy][ix]=0.;
       /*diag0[0][iz][iy][ix]=0.;
       diag0[1][iz][iy][ix]=0.;
       diag8[0][iz][iy][ix]=0.;
       diag8[1][iz][iy][ix]=0.;*/
       }
/* initialize history fields */ 
      for (iy = 0; iy < My; iy++)
       for (ix = 0; ix < Mx; ix++)
       avtaux[iy][ix]=avtauy[iy][ix]=avheat[iy][ix]=avsalt[iy][ix]=convec_freq[iy][ix]=convec_freq2[iy][ix]=0;

/* enforce minima for initial fields */
   for (iz =  0; iz < Mz; iz++)
     for (iy = 0; iy < My; iy++)
      for (ix = 0; ix < Mx; ix++){
       if(annual_temp[iz][iy][ix] < Tf) annual_temp[iz][iy][ix]=Tf;
       if(annual_salt[iz][iy][ix] < 0.) annual_salt[iz][iy][ix]=0.;
       }

    for (iy = 0; iy < My; iy++) /* convective adjustment */
    {
      for (ix = 0; ix < Mx; ix++)
      { int mixing;

        mixing = 1;
        while(mixing)
        {
          mixing = 0;
	      for (iz = Mz-2; iz >= 0; iz--)
	      { float d1 ,d2 , snew, tnew;
            if((h[iz][iy][ix] > 1.0) && (h[iz+1][iy][ix]) > 1.0)
            {
               d1 = density(annual_salt[iz][iy][ix],annual_temp[iz][iy][ix],iz);
               d2 = density(annual_salt[iz+1][iy][ix],annual_temp[iz+1][iy][ix],iz);
               if(d2 > (d1 + 1.0e-7))
               {
                  mixing = 1;

                  snew = (annual_salt[iz][iy][ix]*h[iz][iy][ix] +
                          annual_salt[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);
                  annual_salt[iz][iy][ix] = annual_salt[iz+1][iy][ix] = snew;

                  tnew = (annual_temp[iz][iy][ix]*h[iz][iy][ix] +
                          annual_temp[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);
                  annual_temp[iz][iy][ix] = annual_temp[iz+1][iy][ix] = tnew;
               }
            }
	      }
	    }
      }
    }

   if(my_address == 0)
      printf("Base stratification:\n");
   fflush(of);

  for(iz = 0; iz < Mz; iz++)
  {
    for(iy = 0; iy < My; iy++)
    {
      for(ix = 0; ix < Mx; ix++)
      {
	    ul[iz][iy][ix] = u[iz][iy][ix] = ux[iz][iy][ix] = 0.0;
	    vl[iz][iy][ix] = v[iz][iy][ix] = vx[iz][iy][ix] = 0.0;
	    /*
	    tl[iz][iy][ix] = t[iz][iy][ix] = tx[iz][iy][ix] = base_t[iz];
	    sl[iz][iy][ix] = s[iz][iy][ix] = sx[iz][iy][ix] = base_s[iz];
	    
	    tl[iz][iy][ix] = t[iz][iy][ix] = tx[iz][iy][ix] = to[Nz-1-iz];
	    sl[iz][iy][ix] = s[iz][iy][ix] = sx[iz][iy][ix] =( so[Nz-1-iz]*1000.) + 35.;
	    */
	    tl[iz][iy][ix] = t[iz][iy][ix] = tx[iz][iy][ix] =annual_temp[iz][iy][ix];
	    sl[iz][iy][ix] = s[iz][iy][ix] = sx[iz][iy][ix] =annual_salt[iz][iy][ix];
       }
     }
   if(my_address == 0)
      printf("iz = %d T = %f S = %f\n",iz,to[iz],so[iz]);
   fflush(of);
  }

  for(iy = 0; iy < My; iy++)
  {
    for(ix = 0; ix < Mx; ix++)
    {
	   ptopl[iy][ix] = ptop[iy][ix] = ptopx[iy][ix] = 0.0;
	   qflx[iy][ix]=0.0;
     }
   }

   for (iz =  0; iz < Mz; iz++)
   for(iy = 0; iy < My; iy++)
    for(ix = 0; ix < Mx; ix++)
     qice[iz][iy][ix]=0.;

   for(iy = 0; iy < My; iy++)
   {  float x1,x2,x3,del_lat;

      jy = row*(My-2) + (iy-1);
      yeq = Dely * (Ny/2.0);
      y = Dely * (jy+0.5);
      lat = (2.0*3.14159/4.0e7) * (y-yeq);
      del_lat = 3.14159/Ny;
      f[iy] = 2.0*omega*sin(lat);

      x1 = cos(lat-del_lat);
      x2 = cos(lat);
      x3 = cos(lat+del_lat);

      geo1[iy] = x1/x2;
      geo2[iy] = x3/x2;

      geo3[iy]=1.0/cos(lat);
      if(geo3[iy] > 2.0) geo3[iy]=2.0;

      for(ix = 0; ix < Mx; ix++)
      {
	    if(h[Nz-1][iy][ix] > 1.0)
	    {
	       taux[iy][ix] = strx[iy][ix]/(rhoref * h[Nz-1][iy][ix]);
	       tauy[iy][ix] = stry[iy][ix]/(rhoref * h[Nz-1][iy][ix]);
	    }
	    else
	    {
	       taux[iy][ix] = 0.0;
	       tauy[iy][ix] = 0.0;
	    }
      }
   }
   wrap_q(taux);
   wrap_q(tauy);
}

void load_restart(char* resname)
{
   int jz;
   FILE *tfile;

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",prefix,resname);
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on restart read failed.\n");
      fflush(of);
      exit(89);
   }

   for(jz = 0; jz < Nz; jz++) /* model state fields */
   {
      read_restart_slice(&(u[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(ul[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(v[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(vl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(t[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(tl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(s[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(sl[jz][0]),tfile);
   }
   read_restart_slice(&(ptop[0]),tfile);
   read_restart_slice(&(ptopl[0]),tfile);
#ifdef READ1_5
   read_restart_slice(&(qflx[0]),tfile);
#endif

   if(my_address == 0) fclose(tfile);
}

void read_restart_slice(float a[My][Mx], FILE *tfile)
{
   int ix,iy,jx,jy;
   float row_in[Nx];

   for(jy = 0; jy < Ny; jy++) /* read xy field */
   {
     iy = jy - row*(My-2) + 1;
     if(my_address == 0) fread(row_in,sizeof(float),Nx,tfile);
     MPI_Bcast(row_in,Nx,MPI_FLOAT,0,split_comm);
     for(jx = 0; jx < Nx; jx++)
     {
       ix = jx - col*(Mx-2) + 1;
       if((ix > 0) && (ix < Mx-1) && (iy > 0) && (iy < My-1))
       {
          a[iy][ix] = row_in[jx];
       }
     }
   }
   wrap_q(a);
}

void save_restart(int suffix)
{
   int jz;
   FILE *tfile;

   if(my_address == 0)
   {
#ifdef SIMPLE_OUTPUT
# ifdef OCEAN_ALONE
      sprintf(buff,"restart/restart.ocean.%7.7d",suffix);
# else
      sprintf(buff,"restart/ocean/restart.ocean.%7.7d",suffix);
# endif
#else
      sprintf(buff,"%srestart.ocean.%7.7d",prefix,suffix);
#endif
      tfile = fopen(buff,"w");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on restart write failed. %s\n",buff);
      fflush(of);
      exit(1);
   }
   else if (my_address == 0)
   {
      printf("Opening restart file %s\n",buff);
      fflush(of);
   }

   for(jz = 0; jz < Nz; jz++) /* model state fields */
   {
      write_restart_slice(&(u[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(ul[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(v[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(vl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(t[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(tl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(s[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(sl[jz][0]),tfile);
   }
   write_restart_slice(&(ptop[0]),tfile);
   write_restart_slice(&(ptopl[0]),tfile);
   write_restart_slice(&(qflx[0]),tfile);

   if(my_address == 0) fclose(tfile);
}

void write_restart_slice(float a[My][Mx], FILE *tfile)
{
   int iy,ix,i,j;

   float awhole[Ny*Nx];
   float awhole2[Ny][Nx];
   float aprt[(My-2)*(Mx-2)];
   int ierr;

   i=0;
   for(iy = 1; iy < My-1; iy++)
   for(ix = 1; ix < Mx-1; ix++){
      aprt[i]=a[iy][ix];
      i++;
      }

   ierr=MPI_Gatherv(aprt,(Mx-2)*(My-2),MPI_FLOAT,awhole,counts,displace,
        MPI_FLOAT,0,split_comm);

  if(my_address == 0)
  {

   for(i = 0; i < Nrows*Ncols; i++)
   for(j = 0; j < (Mx-2)*(My-2); j++){
     iy=rows[i]*(My-2) + (j / (Mx-2));
     ix=cols[i]*(Mx-2) + (j % (Mx-2));
     /*fprintf(stderr,"All %d %d %d %d %d\n",i,j,iy,ix,(displace[i])+j);*/
     awhole2[iy][ix]=awhole[(displace[i])+j];
     }

     for(iy = 0; iy < Ny; iy++)
           fwrite(&(awhole2[iy]),sizeof(float),Nx,tfile);

  }
  MPI_Barrier(split_comm);

}

/* ----------------------------------------------------------------------
   Procedure ocean_to_coupler
      This procedure sends the set of fields to the coupler.
-----------------------------------------------------------------------*/


void ocean_to_coupler(int nfields,float data[][My-2][Mx-2],int tag)
{
   int size_array, ierr;

   size_array=(Mx-2)*(My-2)*nfields*sizeof(float);
#ifdef TCP_TEST
nexus_enable_tcp_poll();
#endif
#ifdef COLOCATED_OCEAN
   ierr=MPI_Bsend(data,size_array,MPI_CHAR,0,tag,coupler_ocean);
#else
   ierr=MPI_Send(data,size_array,MPI_CHAR,0,tag,coupler_ocean);
#endif

#ifdef TCP_TEST
nexus_disable_tcp_poll();
#endif
   if (ierr != MPI_SUCCESS) {
      fprintf(stderr,"Error in ocean_to_coupler\n");
   }
}



/* ----------------------------------------------------------------------
   Procedure ocean_from_coupler
      This procedure receives the set of fields from the coupler.
-----------------------------------------------------------------------*/

void ocean_from_coupler(int nfields,float data[][My-2][Mx-2],int tag)
{
   MPI_Status status;
   int size_array, ierr, wordsz;

   size_array=(Mx-2)*(My-2)*nfields*sizeof(float);
#ifdef TCP_TEST
nexus_enable_tcp_poll();
#endif
   ierr=MPI_Recv(data,size_array,MPI_CHAR,0,tag,coupler_ocean,&status);
#ifdef TCP_TEST
nexus_disable_tcp_poll();
#endif

   if (ierr != MPI_SUCCESS) {
      fprintf(stderr,"Error in ocean_from_coupler\n");
   }
}


void time_step()
{
   int step;
   int ix, iy, iz, isplit, isub_split;
   float delt_spl, tdx, tdy;
   float dx, dy;
   float tfor,sfor;
   float gdxdt, gdydt, rgdxdt, rgdydt;
   float dcx,dcy,dcxm,dcym;

   dcxm = Diff_Mks/(Delx * Delx);
   dcym = Diff_Mks/(Dely * Dely);
   dcx = Diff_Ratio*Diff_Mks/(Delx * Delx);
   dcy = Diff_Ratio*Diff_Mks/(Dely * Dely);

#ifdef ANNUAL_SST
   step = Start_step + finished * 4 + (istep + iblock * N_Steps);
#endif

/* Second order conservative dissipation on scalars*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         dts[iz][iy][ix] = 0.0;
         dtt[iz][iy][ix] = 0.0;

         if(h[iz][iy][ix] > 1.0)
         {
            if(h[iz][iy][ix-1] > 1.0)
            {
              dts[iz][iy][ix] -= dcx * (sl[iz][iy][ix] - sl[iz][iy][ix-1]);
              /*diag1[iz][iy][ix] -= Delt * dcx * (sl[iz][iy][ix] - sl[iz][iy][ix-1]);*/
              dtt[iz][iy][ix] -= dcx * (tl[iz][iy][ix] - tl[iz][iy][ix-1]);
            }

	        if(h[iz][iy][ix+1] > 1.0)
            {
	          dts[iz][iy][ix] -= dcx * (sl[iz][iy][ix] - sl[iz][iy][ix+1]);
	          /*diag1[iz][iy][ix] -= Delt * dcx * (sl[iz][iy][ix] - sl[iz][iy][ix+1]);*/
	          dtt[iz][iy][ix] -= dcx * (tl[iz][iy][ix] - tl[iz][iy][ix+1]);
            }

            if(h[iz][iy-1][ix] > 1.0)
            {
	          dts[iz][iy][ix] -= dcy * (sl[iz][iy][ix] - sl[iz][iy-1][ix]);
	          /*diag1[iz][iy][ix] -= Delt * dcy * (sl[iz][iy][ix] - sl[iz][iy-1][ix]);*/
	          dtt[iz][iy][ix] -= dcy * (tl[iz][iy][ix] - tl[iz][iy-1][ix]);
            }

	        if(h[iz][iy+1][ix] > 1.0)
            {
	          dts[iz][iy][ix] -= dcy * (sl[iz][iy][ix] - sl[iz][iy+1][ix]);
	          /*diag1[iz][iy][ix] -= Delt *dcy * (sl[iz][iy][ix] - sl[iz][iy+1][ix]);*/
	          dtt[iz][iy][ix] -= dcy * (tl[iz][iy][ix] - tl[iz][iy+1][ix]);
            }
         }
      }
   }
 }

 /* Second order conservative dissipation on momentum*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         dtu[iz][iy][ix] = 0.0;
         dtv[iz][iy][ix] = 0.0;

         if(uvh[iz][iy][ix] > 1.0)
         {
            if(uvh[iz][iy][ix-1] > 1.0)
            {
              dtu[iz][iy][ix] -= dcxm * (ul[iz][iy][ix] - ul[iz][iy][ix-1]);
              dtv[iz][iy][ix] -= dcxm * (vl[iz][iy][ix] - vl[iz][iy][ix-1]);
            }

	        if(uvh[iz][iy][ix+1] > 1.0)
            {
	          dtu[iz][iy][ix] -= dcxm * (ul[iz][iy][ix] - ul[iz][iy][ix+1]);
	          dtv[iz][iy][ix] -= dcxm * (vl[iz][iy][ix] - vl[iz][iy][ix+1]);
            }

            if(uvh[iz][iy-1][ix] > 1.0)
            {
	          dtu[iz][iy][ix] -= dcym * (ul[iz][iy][ix] - ul[iz][iy-1][ix]);
	          dtv[iz][iy][ix] -= dcym * (vl[iz][iy][ix] - vl[iz][iy-1][ix]);
            }

	        if(uvh[iz][iy+1][ix] > 1.0)
            {
	          dtu[iz][iy][ix] -= dcym * (ul[iz][iy][ix] - ul[iz][iy+1][ix]);
	          dtv[iz][iy][ix] -= dcym * (vl[iz][iy][ix] - vl[iz][iy+1][ix]);
            }
         }
      }
   }
 }

/* wind stress forcing on points with full dynamics */

 for (iy = 1; iy < My-1; iy++)
 {
   for (ix = 1; ix < Mx-1; ix++)
   {
	 if(uvh[Nz-1][iy][ix] >1.0)
	 {
#ifdef ANNUAL_SST
	    dtu[Nz-1][iy][ix] += tauxint(ix-1,iy-1,step)/(rhoref * h[Nz-1][iy][ix]);
	    dtv[Nz-1][iy][ix] += tauyint(ix-1,iy-1,step)/(rhoref * h[Nz-1][iy][ix]);
	    taux_for[iy][ix] += tauxint(ix-1,iy-1,step)/(rhoref * h[Nz-1][iy][ix]);
	    tauy_for[iy][ix] += tauyint(ix-1,iy-1,step)/(rhoref * h[Nz-1][iy][ix]);
	    avtaux[iy][ix] += tauxint(ix-1,iy-1,step)/(rhoref * h[Nz-1][iy][ix]);
	    avtauy[iy][ix] += tauyint(ix-1,iy-1,step)/(rhoref * h[Nz-1][iy][ix]);

#else
	    dtu[Nz-1][iy][ix] += taux[iy][ix];
	    dtv[Nz-1][iy][ix] += tauy[iy][ix];
	    avtaux[iy][ix] += taux[iy][ix];
	    avtauy[iy][ix] += tauy[iy][ix];
#endif
     }
   }
 }

/* Surface forcing on T,S */

 for (iy = 1; iy < My-1; iy++)
 {
   for (ix = 1; ix < Mx-1; ix++)
   {
	 if(h[Nz-1][iy][ix] >1.0)
	 {
#ifdef ANNUAL_SST

/* SET OCEAN-ONLY TEMP and SALT FORCING HERE
   Select a surface forcing by commenting/uncommenting the
   appropriate piece of code below */


/* Restoring forcing to the annual average contained
   in initial files om3.temp and om3.salt   */

/*        tfor = Restore*(annual_temp[Nz-1][iy][ix]-tl[Nz-1][iy][ix]);
	sfor = Restore * (annual_salt[Nz-1][iy][ix]-sl[Nz-1][iy][ix]);
        dtt[Nz-1][iy][ix] += tfor;
        dts[Nz-1][iy][ix] += sfor;
        t_forcing[iy][ix] += Delt * tfor;
        s_forcing[iy][ix] += Delt * sfor;
        avheat[iy][ix] += tfor * Cp * rhoref *h[Nz-1][iy][ix];
        avsalt[iy][ix] += sfor;*/

/* Restoring forcing using monthly SST from sstfile and
   annual SSS using sssom3file */

	tfor = Restore * (sstint(ix-1,iy-1,step)-tl[Nz-1][iy][ix]);
	sfor = Restore * (sssom3[iy-1][ix-1]-sl[Nz-1][iy][ix]);
        dtt[Nz-1][iy][ix] += tfor;
        dts[Nz-1][iy][ix] += sfor;
        t_forcing[iy][ix] += Delt * tfor;
        s_forcing[iy][ix] += Delt * sfor;
        avheat[iy][ix] += tfor * Cp * rhoref * h[Nz-1][iy][ix];
        avsalt[iy][ix] += sfor;


/* linearized-restoring forcing */
/* Use heatflux forcing from file heatfile and add a restoring force
   using SST data from sstfile.   For salinity, use freshwater flux
  in saltfile and also restore to monthly data in sstfile  */

/* 	tfor = (heatint(ix-1,iy-1,step)/(rhoref*h[Nz-1][iy][ix]* Cp)) +
			Restore * (sstint(ix-1,iy-1,step)-tl[Nz-1][iy][ix]);
        sfor = (-s[Nz-1][iy][ix]*saltint(ix-1,iy-1,step)/(rhoref * h[Nz-1][iy][ix])) + Restore * (sssint(ix-1,iy-1,step)-sl[Nz-1][iy][ix]);

        dtt[Nz-1][iy][ix] += tfor;
        dts[Nz-1][iy][ix] += sfor;

        t_forcing[iy][ix] += Delt * tfor;
        s_forcing[iy][ix] += Delt * sfor;
        avheat[iy][ix] += tfor* Cp * rhoref *h[Nz-1][iy][ix];
        avsalt[iy][ix] += sfor;*/
	
#else
/* use forcing from the coupler */
        dtt[Nz-1][iy][ix] += temp_for[iy][ix];
        dts[Nz-1][iy][ix] += salt_for[iy][ix];
	avheat[iy][ix] += temp_for[iy][ix]* Cp * rhoref *h[Nz-1][iy][ix];
	avsalt[iy][ix] += salt_for[iy][ix];
#endif
     }
   }
 }

/* Calculate Box u-advection dt's u,v,s,t.  Accumulate into dt arrays. */

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      double advc, advc1, advc2;

      advc = geo3[iy]*1.0/(4.0 *Delx);

      for (ix = 1; ix < Mx-1; ix++)
      {
	 if(h[iz][iy][ix] >1.0)
	 {
	    if(h[iz][iy][ix-1] >1.0)
           advc1 = advc * (u[iz][iy][ix-1] + u[iz][iy][ix]);
        else 
           advc1 = 0.0;

	    if(h[iz][iy][ix+1] >1.0)
  	       advc2 = advc * (u[iz][iy][ix] + u[iz][iy][ix+1]);
        else 
           advc2 = 0.0;

	    dtu[iz][iy][ix] += advc1 * (u[iz][iy][ix-1] - u[iz][iy][ix])
	        + advc2 * (u[iz][iy][ix] - u[iz][iy][ix+1]);

	    dtv[iz][iy][ix] += advc1 * (v[iz][iy][ix-1] - v[iz][iy][ix])
	        + advc2 * (v[iz][iy][ix] - v[iz][iy][ix+1]);

	    dts[iz][iy][ix] += advc1 * (s[iz][iy][ix-1] - s[iz][iy][ix])
	        + advc2 * (s[iz][iy][ix] - s[iz][iy][ix+1]);

	    /*diag2[iz][iy][ix] += Delt * advc1 * (s[iz][iy][ix-1] - s[iz][iy][ix])
	        + Delt * advc2 * (s[iz][iy][ix] - s[iz][iy][ix+1]);*/

	    dtt[iz][iy][ix] += advc1 * (t[iz][iy][ix-1] - t[iz][iy][ix])
	        + advc2 * (t[iz][iy][ix] - t[iz][iy][ix+1]);
	 }
      }
   }
 }

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      double advc, advc1, advc2;

      advc = 1.0/(4.0 *Dely);

      for (ix = 1; ix < Mx-1; ix++)
      {
	 if(h[iz][iy][ix] >1.0)
	 {
	    if(h[iz][iy-1][ix] >1.0)
           advc1 = advc * (v[iz][iy-1][ix] + v[iz][iy][ix]);
        else 
           advc1 = 0.0;

	    if(h[iz][iy+1][ix] >1.0)
  	       advc2 = advc * (v[iz][iy][ix] + v[iz][iy+1][ix]);
        else 
           advc2 = 0.0;

	    dtu[iz][iy][ix] += advc1 * (u[iz][iy-1][ix] - u[iz][iy][ix])
	        + advc2 * (u[iz][iy][ix] - u[iz][iy+1][ix]);

	    dtv[iz][iy][ix] += advc1 * (v[iz][iy-1][ix] - v[iz][iy][ix])
	        + advc2 * (v[iz][iy][ix] - v[iz][iy+1][ix]);

	    dts[iz][iy][ix] += advc1 * (s[iz][iy-1][ix] - s[iz][iy][ix])
	        + advc2 * (s[iz][iy][ix] - s[iz][iy+1][ix]);

	    /*diag2[iz][iy][ix] += Delt * advc1 * (s[iz][iy-1][ix] - s[iz][iy][ix])
	        + Delt * advc2 * (s[iz][iy][ix] - s[iz][iy+1][ix]);*/

	    dtt[iz][iy][ix] += advc1 * (t[iz][iy-1][ix] - t[iz][iy][ix])
	        + advc2 * (t[iz][iy][ix] - t[iz][iy+1][ix]);
	 }
      }
   }
 }

   /* setup vx, wx, tx, sx for time split step */

   for (iz = 0; iz < Mz; iz++)
   {
     for (iy = 0; iy < My; iy++)
     {
       for (ix = 0; ix < Mx; ix++)
       {
	      ux[iz][iy][ix] = ul[iz][iy][ix];

	      vx[iz][iy][ix] = vl[iz][iy][ix];

	      tx[iz][iy][ix] = tl[iz][iy][ix];

	      sx[iz][iy][ix] = sl[iz][iy][ix];
       }
     }
   }
   /* setup ptopx for time split step */

     for (iy = 0; iy < My; iy++)
     {
       for (ix = 0; ix < Mx; ix++)
       {
	      ptopx[iy][ix] = ptopl[iy][ix];
       }
     }

   /* caluculate mixing parameters before split */

   for (iz = Mz-1; iz > 0; iz--) /* calculate downward mixing  not iz=0 */
   {
     for (iy = 1; iy < My-1; iy++)
     {
       for (ix = 1; ix < Mx-1; ix++)
       {
	  if(h[iz][iy][ix] > 1.0)
          {  float shrsq,rr2,rr4,rr8;
             if(h[iz-1][iy][ix] > 1.0) /* interior layer */
             {
                nsq[iz][iy][ix] = 9.8*
                  (density(sx[iz-1][iy][ix],tx[iz-1][iy][ix],iz)-
                   density(sx[iz][iy][ix],tx[iz][iy][ix],iz))
                  /(0.5*(h[iz-1][iy][ix]+h[iz][iy][ix]));
             }
             else /* bottom layer */
             {
                if(iz < (Mz-1)) nsq[iz][iy][ix] = nsq[iz+1][iy][ix];
                else nsq[iz][iy][ix] = 3.0e-5; /* also top layer */
             }
             if(nsq[iz][iy][ix] < 1.0e-8) nsq[iz][iy][ix] = 1.0e-8; /* clamp */
             if(uvh[iz][iy][ix] > 1.0) /* cell with active vels */
             {
                if(uvh[iz-1][iy][ix] > 1.0)
                {
                   shrsq = (ux[iz][iy][ix]-ux[iz-1][iy][ix])*
                           (ux[iz][iy][ix]-ux[iz-1][iy][ix])
                          +(vx[iz][iy][ix]-vx[iz-1][iy][ix])*
                           (vx[iz][iy][ix]-vx[iz-1][iy][ix]);
                }
                else
                {
                   shrsq = ux[iz][iy][ix]*ux[iz][iy][ix]
                          +vx[iz][iy][ix]*vx[iz][iy][ix];
                }
                shrsq /= 0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]);
                shrsq /= 0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]);
             }
             else shrsq=0.0;
             if(shrsq < 0.01*nsq[iz][iy][ix]) shrsq = 0.01*nsq[iz][iy][ix]; /* clamp */
             risq[iz][iy][ix] = nsq[iz][iy][ix]/shrsq;
             if(risq[iz][iy][ix] < 0.01) risq[iz][iy][ix] = 0.01;
             rr2 = 1.0/risq[iz][iy][ix];
             rr4 = rr2*rr2;
             rr8 = rr4*rr4;
             kuv[iz][iy][ix] = (7.38e-8 * rr8)+1.0e-4;
             if(kuv[iz][iy][ix] > 0.01) kuv[iz][iy][ix] = 0.01;
             kts[iz][iy][ix] = (2.75e-8 * rr8)+1.0e-5;
             if(kts[iz][iy][ix] > 0.01) kts[iz][iy][ix] = 0.01;
          } 
       }
     }
   }

   iz=0; /* calcualte downward mixing, bottom layer */
   {
     for (iy = 1; iy < My-1; iy++)
     {
       for (ix = 1; ix < Mx-1; ix++)
       {
	  if(h[iz][iy][ix] > 1.0)
          {  float shrsq,rr2,rr4,rr8;

             nsq[iz][iy][ix] = nsq[iz+1][iy][ix];

             if(uvh[iz][iy][ix] > 1.0) /* cell with active vels */
             {
                shrsq = ux[iz][iy][ix]*ux[iz][iy][ix]
                       +vx[iz][iy][ix]*vx[iz][iy][ix];
                shrsq /= 0.5*uvh[iz][iy][ix];
                shrsq /= 0.5*uvh[iz][iy][ix];
             }
             else shrsq=0.0;
             if(shrsq < 0.01*nsq[iz][iy][ix]) shrsq = 0.01*nsq[iz][iy][ix]; /* clamp */
             risq[iz][iy][ix] = nsq[iz][iy][ix]/shrsq;
             if(risq[iz][iy][ix] < 0.01) risq[iz][iy][ix] = 0.01;
             rr2 = 1.0/risq[iz][iy][ix];
             rr4 = rr2*rr2;
             rr8 = rr4*rr4;
             kuv[iz][iy][ix] = (7.38e-8 * rr8)+1.0e-4;
             if(kuv[iz][iy][ix] > 0.01) kuv[iz][iy][ix] = 0.01;
             kts[iz][iy][ix] = (2.75e-8 * rr8)+1.0e-5;
             if(kts[iz][iy][ix] > 0.01) kts[iz][iy][ix] = 0.01;
          } 
       }
     }
   }

   delt_spl = 2.0 * Delt/((float) N_Split);

    for (iz = 0; iz < Mz-1; iz++) /* vertical advection of scalars - gradient calculation */
    { float dz;
      for (iy = 1; iy < My-1; iy++)
      {
	 for (ix = 1; ix < Mx-1; ix++)
	 {
        if((h[iz][iy][ix] >1.0) && (h[iz+1][iy][ix] >1.0))
        {
           dz = 1/(0.5*(h[iz][iy][ix]+h[iz+1][iy][ix]));

	       dzt[iz][iy][ix] = dz*(t[iz+1][iy][ix]-t[iz][iy][ix]);
	       dzs[iz][iy][ix] = dz*(s[iz+1][iy][ix]-s[iz][iy][ix]);
        }
        if((uvh[iz][iy][ix] >1.0) && (uvh[iz+1][iy][ix] >1.0))
        {
           dz = 1/(0.5*(h[iz][iy][ix]+h[iz+1][iy][ix]));

	       dzu[iz][iy][ix] = dz*(u[iz+1][iy][ix]-u[iz][iy][ix]);
	       dzv[iz][iy][ix] = dz*(v[iz+1][iy][ix]-v[iz][iy][ix]);
        }
	 }
      }
    }

   for (isplit = 0; isplit < N_Split; isplit++) /* time split hydrostatic integration */
   { float dxdt,dydt;

    for (iz = 0; iz < Mz; iz++) /* calculate vertical velocity */
    {  float cwx, cwy;

      cwy = 1.0/(2.0*Dely);

      for (iy = 1; iy < My-1; iy++)
      {
      cwx = geo3[iy]*1.0/(2.0*Delx);
	 for (ix = 1; ix < Mx-1; ix++)
	 {
        if(h[iz][iy][ix] >1.0)
        {
	      if(iz == 0)
          {
                w[iz][iy][ix] =
                    -cwx*(uvh[iz][iy][ix+1]*ux[iz][iy][ix+1] 
                        - uvh[iz][iy][ix-1]*ux[iz][iy][ix-1])
	                -cwy*(geo2[iy]*uvh[iz][iy+1][ix]*vx[iz][iy+1][ix]
                        - geo1[iy]*uvh[iz][iy-1][ix]*vx[iz][iy-1][ix]);
           } 
	       else
           {
                w[iz][iy][ix] = w[iz-1][iy][ix]
                     -cwx*(uvh[iz][iy][ix+1]*ux[iz][iy][ix+1] 
                         - uvh[iz][iy][ix-1]*ux[iz][iy][ix-1])
	                 -cwy*(geo2[iy]*uvh[iz][iy+1][ix]*vx[iz][iy+1][ix]
                         - geo1[iy]*uvh[iz][iy-1][ix]*vx[iz][iy-1][ix]);
           }
        }
        else w[iz][iy][ix] = 0.0;
	 }
      }
    }

    for (iz = 0; iz < Mz; iz++) /* Baro correction for vertical velocity */
    {
      float thick[My][Mx];

      for (iy = 1; iy < My-1; iy++)
      {
        for (ix = 1; ix < Mx-1; ix++)
        {
           if(iz==0) thick[iy][ix]=h[iz][iy][ix];
           else thick[iy][ix] += h[iz][iy][ix];

        if(h[iz][iy][ix] >1.0)
           w[iz][iy][ix] -= (thick[iy][ix]/depth[iy][ix]) * w[Nz-1][iy][ix];
        }
      }
    }

    for (iz = 0; iz < Mz-1; iz++) /* vertical advection of scalars - update */
    { float dt;
      for (iy = 1; iy < My-1; iy++)
      {
	 for (ix = 1; ix < Mx-1; ix++)
	 {
        if((h[iz][iy][ix] >1.0) && (h[iz+1][iy][ix] >1.0))
        {
           dt = 0.5*delt_spl;

	       tx[iz][iy][ix] -= dt*w[iz][iy][ix]*dzt[iz][iy][ix];
	       tx[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzt[iz][iy][ix];

	       sx[iz][iy][ix] -= dt*w[iz][iy][ix]*dzs[iz][iy][ix];
	       /*diag3[iz][iy][ix] -= 0.5*dt*w[iz][iy][ix]*dzs[iz][iy][ix];*/
	       sx[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzs[iz][iy][ix];
	       /*diag3[iz+1][iy][ix] -= 0.5*dt*w[iz][iy][ix]*dzs[iz][iy][ix];*/
        }
        if((uvh[iz][iy][ix] >1.0) && (uvh[iz+1][iy][ix] >1.0))
        {
           dt = 0.5*delt_spl;

	       ux[iz][iy][ix] -= dt*w[iz][iy][ix]*dzu[iz][iy][ix];
	       ux[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzu[iz][iy][ix];

	       vx[iz][iy][ix] -= dt*w[iz][iy][ix]*dzv[iz][iy][ix];
	       vx[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzv[iz][iy][ix];
        }
	 }
      }
    }

    for (iz = 0; iz < Mz; iz++) /* time tendancies for scalars */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	   for (ix = 1; ix < Mx-1; ix++)
	   {
         if(h[iz][iy][ix] >1.0)
         {
	       tx[iz][iy][ix] += delt_spl * dtt[iz][iy][ix];
	       sx[iz][iy][ix] += delt_spl * dts[iz][iy][ix];
         }
	   }
      }
    }

    wrap_qz(ux);
    wrap_qz(vx);


    for (iy = 0; iy < My; iy++) /* split barotropic mode u,v */
    {
      for (ix = 0; ix < Mx; ix++)
      {
        ubaro[iy][ix] = 0.0;
        vbaro[iy][ix] = 0.0;
	    for (iz = 0; iz < Mz; iz++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
	          ubaro[iy][ix] += h[iz][iy][ix]*ux[iz][iy][ix];
	          vbaro[iy][ix] += h[iz][iy][ix]*vx[iz][iy][ix];
           }
	    }
        ubaro[iy][ix] = ruvdepth[iy][ix] * ubaro[iy][ix];
        vbaro[iy][ix] = ruvdepth[iy][ix] * vbaro[iy][ix];
	    for (iz = 0; iz < Mz; iz++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
              ux[iz][iy][ix] -= ubaro[iy][ix];
              vx[iz][iy][ix] -= vbaro[iy][ix];
           }
	    }
      }
    }

    for (iy = 1; iy < My-1; iy++)
    {
	  for (ix = 1; ix < Mx-1; ix++)
	  {
         p[Nz-1][iy][ix] = 0.0;
	  }
    }

    for (iz = Nz-2; iz >= 0; iz--) /* hydrostatic pressure integration */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	    for (ix = 1; ix < Mx-1; ix++)
	    {
	      p[iz][iy][ix] = p[iz+1][iy][ix]+
             9.8*0.5*(
               h[iz][iy][ix]*density(sx[iz][iy][ix],tx[iz][iy][ix],iz)
             + h[iz+1][iy][ix]*density(sx[iz+1][iy][ix],tx[iz+1][iy][ix],iz+1));
	    }
      }
    }

    wrap_qz(p);

    dy = 1.0/(2.0*Dely);

    for (iy = 1; iy < My-1; iy++) /* split barotropic mode dtu, dtv including pgf */
    { float pgfx, pgfy;
      dx = geo3[iy]*1.0/(2.0*Delx);
      for (ix = 1; ix < Mx-1; ix++)
      {
        dtubaro[iy][ix] = 0.0;
        dtvbaro[iy][ix] = 0.0;
	    for (iz = 0; iz < Mz; iz++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
	          pgfx = -dx * (p[iz][iy][ix+1] - p[iz][iy][ix-1]);
	          dtubaro[iy][ix] += h[iz][iy][ix] * (pgfx + dtu[iz][iy][ix]);

	          pgfy = -dy * (p[iz][iy+1][ix] - p[iz][iy-1][ix]);
	          dtvbaro[iy][ix] += h[iz][iy][ix] * (pgfy + dtv[iz][iy][ix]);
           }
	    }
        dtubaro[iy][ix] = ruvdepth[iy][ix] * dtubaro[iy][ix];
        dtvbaro[iy][ix] = ruvdepth[iy][ix] * dtvbaro[iy][ix];
      }
    }


    dydt = delt_spl/(2.0*Dely);

    for (iz = 0; iz < Mz; iz++)
    {
      for (iy = 1; iy < My-1; iy++)
      { float dtf, a, b, r, pgfx, pgfy;

     dxdt = geo3[iy]*delt_spl/(2.0*Delx);

	 dtf = delt_spl*f[iy];
     r = 1.0/(1.0+dtf*dtf); /* norm for semi-implicit coriolis */

	 for (ix = 1; ix < Mx-1; ix++)
	 {
	    if(uvh[iz][iy][ix] >1.0) /* no velocities in dry boxes */
	    {
	         pgfx = -dxdt * (p[iz][iy][ix+1] - p[iz][iy][ix-1]);
             a = delt_spl * (dtu[iz][iy][ix] - dtubaro[iy][ix]);

	         pgfy = -dydt * (p[iz][iy+1][ix] - p[iz][iy-1][ix]);
             b = delt_spl * (dtv[iz][iy][ix] - dtvbaro[iy][ix]);

	         ux[iz][iy][ix] = r*(ux[iz][iy][ix] + pgfx + 
                                 dtf*(vx[iz][iy][ix]+ b + pgfy) + a);

 	         vx[iz][iy][ix] += pgfy - dtf*ux[iz][iy][ix] + b;
	    }
	 }
      }
    }

 for (isub_split = 0; isub_split < Sub_Split; isub_split++) /* barotropic sub-cycle */
 { float delt_sub;

    delt_sub = delt_spl/Sub_Split;

   for (iy = 1; iy < My-1; iy++) /* Barotropic Pressure field predict*/
   { float cwx, cwy;

     cwx = geo3[iy]*1.0/(2.0*Delx);
     cwy = 1.0/(2.0*Dely);

	 for (ix = 1; ix < Mx-1; ix++)
	 {
	    ptstar[iy][ix] = ptopx[iy][ix]-G*delt_sub*
                           (cwx*(uvdepth[iy][ix+1]*ubaro[iy][ix+1] 
                               - uvdepth[iy][ix-1]*ubaro[iy][ix-1])
	                       +cwy*(geo2[iy]*uvdepth[iy+1][ix]*vbaro[iy+1][ix]
                               - geo1[iy]*uvdepth[iy-1][ix]*vbaro[iy-1][ix]));
	 }
   }

   dydt = delt_sub/(2.0*Dely);

   for (iy = 1; iy < My-1; iy++) /* barotropic u,v predict*/
   { float dtf, a, b, r, pgfx, pgfy;

     dxdt = geo3[iy]*delt_sub/(2.0*Delx);

	 dtf = delt_sub*f[iy];
     r = 1.0/(1.0 + dtf*dtf); /* norm for semi-implicit coriolis */

	 for (ix = 1; ix < Mx-1; ix++)
	 {

	    if(uvdepth[iy][ix] > 1.0)
        {
	         pgfx = -dxdt * (ptopx[iy][ix+1] - ptopx[iy][ix-1]);
             a = delt_sub * dtubaro[iy][ix];

	         pgfy = -dydt * (ptopx[iy+1][ix] - ptopx[iy-1][ix]);
             b = delt_sub * dtvbaro[iy][ix];

	         ubstar[iy][ix] = ubaro[iy][ix] + pgfx + dtf*vbaro[iy][ix] + a;

 	         vbstar[iy][ix] = vbaro[iy][ix] + pgfy - dtf*ubaro[iy][ix] + b;
         }
	  }
   }

   wrap_q3(ubstar, vbstar, ptstar);

   for (iy = 1; iy < My-1; iy++) /* Barotropic Pressure field correct*/
   { float cwx, cwy;

     cwx = geo3[iy]*1.0/(2.0*Delx);
     cwy = 1.0/(2.0*Dely);

	 for (ix = 1; ix < Mx-1; ix++)
	 {
	    ptopx[iy][ix] = ptopx[iy][ix]-G*delt_sub*
                           (cwx*(uvdepth[iy][ix+1]*ubstar[iy][ix+1] 
                               - uvdepth[iy][ix-1]*ubstar[iy][ix-1])
	                       +cwy*(geo2[iy]*uvdepth[iy+1][ix]*vbstar[iy+1][ix]
                               - geo1[iy]*uvdepth[iy-1][ix]*vbstar[iy-1][ix]));
	 }
   }

   dydt = delt_sub/(2.0*Dely);

   for (iy = 1; iy < My-1; iy++) /* barotropic u,v correct */
   { float dtf, a, b, r, pgfx, pgfy;

     dxdt = geo3[iy]*delt_sub/(2.0*Delx);

	 dtf = delt_sub*f[iy];

	 for (ix = 1; ix < Mx-1; ix++)
	 {

	    if(uvdepth[iy][ix] > 1.0)
        {
	         pgfx = -dxdt * (ptstar[iy][ix+1] - ptstar[iy][ix-1]);
             a = delt_sub * dtubaro[iy][ix];

	         pgfy = -dydt * (ptstar[iy+1][ix] - ptstar[iy-1][ix]);
             b = delt_sub * dtvbaro[iy][ix];

	         ubaro[iy][ix] = ubaro[iy][ix] + pgfx + dtf*vbstar[iy][ix] + a;

 	         vbaro[iy][ix] = vbaro[iy][ix] + pgfy - dtf*ubstar[iy][ix] + b;
         }
	  }
   }

  wrap_q3(ubaro, vbaro, ptopx);
  filt4p(ptopx);

 } /* end barotropic sub cycle */

    for (iz = 0; iz < Mz; iz++) /* unsplit barotropic mode u,v */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	    for (ix = 1; ix < Mx-1; ix++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
              ux[iz][iy][ix] += ubaro[iy][ix];
              vx[iz][iy][ix] += vbaro[iy][ix];
           }
	    }
      }
    }

 /* Vertical mixing */

 for (iz = 1; iz < Mz-1; iz++) /* Interior levels */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { float u_up_str,u_down_str,v_up_str,v_down_str,mix;

            if(uvh[iz-1][iy][ix] > 1.0)
            {
              u_down_str = (ux[iz][iy][ix]-ux[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
              v_down_str = (vx[iz][iy][ix]-vx[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
            }
            else
            {
              u_down_str = ux[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
              v_down_str = vx[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
            }
            u_up_str = (ux[iz+1][iy][ix]-ux[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));
            v_up_str = (vx[iz+1][iy][ix]-vx[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));

            u_down_str *= kuv[iz][iy][ix];
            v_down_str *= kuv[iz][iy][ix];
            u_up_str *= kuv[iz+1][iy][ix];
            v_up_str *= kuv[iz+1][iy][ix];

            mix = delt_spl/uvh[iz][iy][ix];
            mix_u[iz][iy][ix] = mix*(u_up_str-u_down_str);
            mix_v[iz][iy][ix] = mix*(v_up_str-v_down_str);
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { float t_up_str,t_down_str,s_up_str,s_down_str,scmix;

            if(h[iz-1][iy][ix] > 1.0)
            {
              t_down_str = (tx[iz][iy][ix]-tx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
              s_down_str = (sx[iz][iy][ix]-sx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
            }
            else
            {
              t_down_str = 0.0;
              s_down_str = 0.0;
            }
            t_up_str = (tx[iz+1][iy][ix]-tx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));
            s_up_str = (sx[iz+1][iy][ix]-sx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));

            t_down_str *= kts[iz][iy][ix];
            s_down_str *= kts[iz][iy][ix];
            t_up_str *= kts[iz+1][iy][ix];
            s_up_str *= kts[iz+1][iy][ix];

            scmix = delt_spl/h[iz][iy][ix];
            mix_t[iz][iy][ix] = scmix*(t_up_str-t_down_str);
            mix_s[iz][iy][ix] = scmix*(s_up_str-s_down_str);
         }
      }
   }
 }

 iz = 0; /* Bottom level */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { float u_up_str,u_down_str,v_up_str,v_down_str,mix;

            u_down_str = ux[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
            v_down_str = vx[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));

            u_up_str = (ux[iz+1][iy][ix]-ux[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));
            v_up_str = (vx[iz+1][iy][ix]-vx[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));

            u_down_str *= kuv[iz][iy][ix];
            v_down_str *= kuv[iz][iy][ix];
            u_up_str *= kuv[iz+1][iy][ix];
            v_up_str *= kuv[iz+1][iy][ix];

            mix = delt_spl/uvh[iz][iy][ix];
            mix_u[iz][iy][ix] = mix*(u_up_str-u_down_str);
            mix_v[iz][iy][ix] = mix*(v_up_str-v_down_str);
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { float t_up_str,s_up_str,scmix;

            t_up_str = (tx[iz+1][iy][ix]-tx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));
            s_up_str = (sx[iz+1][iy][ix]-sx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));

            t_up_str *= kts[iz+1][iy][ix];
            s_up_str *= kts[iz+1][iy][ix];

            scmix = delt_spl/h[iz][iy][ix];
            mix_t[iz][iy][ix] = scmix*(t_up_str);
            mix_s[iz][iy][ix] = scmix*(s_up_str);
         }
      }
   }
 }

 iz=Nz-1; /* Top level */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { float u_down_str,v_down_str,mix;

            if(uvh[iz-1][iy][ix] > 1.0)
            {
              u_down_str = (ux[iz][iy][ix]-ux[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
              v_down_str = (vx[iz][iy][ix]-vx[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
            }
            else
            {
              u_down_str = ux[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
              v_down_str = vx[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
            }

            u_down_str *= kuv[iz][iy][ix];
            v_down_str *= kuv[iz][iy][ix];

            mix = delt_spl/uvh[iz][iy][ix];
            mix_u[iz][iy][ix] = mix*(-u_down_str);
            mix_v[iz][iy][ix] = mix*(-v_down_str);
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { float t_down_str,s_down_str,scmix;

            if(h[iz-1][iy][ix] > 1.0)
            {
              t_down_str = (tx[iz][iy][ix]-tx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
              s_down_str = (sx[iz][iy][ix]-sx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
            }
            else
            {
              t_down_str = 0.0;
              s_down_str = 0.0;
            }

            t_down_str *= kts[iz][iy][ix];
            s_down_str *= kts[iz][iy][ix];

            scmix = delt_spl/h[iz][iy][ix];
            mix_t[iz][iy][ix] = scmix*(-t_down_str);
            mix_s[iz][iy][ix] = scmix*(-s_down_str);
         }
      }
   }
 }

 for (iz = 0; iz < Mz; iz++) /* mixing update */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { 
            ux[iz][iy][ix] += mix_u[iz][iy][ix];
            vx[iz][iy][ix] += mix_v[iz][iy][ix];
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { 
            tx[iz][iy][ix] += mix_t[iz][iy][ix];
            sx[iz][iy][ix] += mix_s[iz][iy][ix];
            diag6[0][iz][iy][ix] += 0.5 * mix_t[iz][iy][ix];
            diag6[1][iz][iy][ix] += 0.5 * mix_s[iz][iy][ix];
         }
      }
   }
 }
    wrap_qz(ux);
    wrap_qz(vx);
    wrap_qz(tx);
    wrap_qz(sx);
   } /* end timesplit part */

    for (iz = 0; iz < Mz; iz++) /* calculate full pressure field for diagnostics */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	    for (ix = 1; ix < Mx-1; ix++)
	    {
	      p[iz][iy][ix] += ptopx[iy][ix];
	    }
      }
    }
   /* for (iz = 0; iz < Mz; iz++)
     for (iy = 0; iy < My; iy++)
      for (ix = 0; ix < Mx; ix++){
            diag0[0][iz][iy][ix] = 0.5 * (tx[iz][iy][ix]-tl[iz][iy][ix]);
            diag0[1][iz][iy][ix] = 0.5 * (sx[iz][iy][ix]-sl[iz][iy][ix]);
	    }
   */


/* ************************************************************
SEA ICE.   
    this algorithm computes ocean heat flux to the sea-ice. it forms
    the necessary ice in the ocean and adjusts the potential 
    temperature and salinity fields accordingly. the logic of this 
    subroutine is based on William Large''s 1-d model and is based
    on a version from the NCOM model written by Ghokan Danabasoglu. 

    If ice forms in lower layers but layers above are warm - the heat is
    used to melt the ice.  total ice heat flux is accumulated.
    WARNING: unless a monotone advection scheme is in place,
    advective errors could lead to temps that are far below freezing
    in some locations and this scheme will form lots of ice.
    ice formation should be limited to the top layer (izmxice=1)
    if the advection scheme is not monotone.

    NOTE:  conversion to C and incorporation into Om3 by R. Jacob
********************************************************** */

/* do sea ice model izmxice levels below and including surface */
     for (iy = 0; iy < My; iy++)
     {
      for (ix = 0; ix < Mx; ix++)
      {

   for(iz = Mz-izmxice; iz < Mz; iz++)
    {
        if (h[iz][iy][ix] > 1.0)
        {

/* find water flux due to ice formation in kg/m2s */
/* positive when  tx < Tf */
           potice = (rhoref*Cp/Lathf)*
              h[iz][iy][ix]*(Tf - tx[iz][iy][ix]);

              if( potice < qice[iz-1][iy][ix])potice=qice[iz-1][iy][ix];

              qice[iz][iy][ix] += qice[iz-1][iy][ix]-potice;

/* convert water flux to temperature forcing */
              tx[iz][iy][ix] +=  potice*
                 Lathf/(h[iz][iy][ix]*rhoref*Cp);
              /*diag8[0][iz][iy][ix]=potice*Lathf/(h[iz][iy][ix]*rhoref*Cp);*/

/* convert water flux to salinity forcing */
              sx[iz][iy][ix] += potice*(Sref)/(rhoref*h[iz][iy][ix]);
              /*diag8[1][iz][iy][ix]=potice*(Sref)/(rhoref*h[iz][iy][ix]);*/
        }
      }
     }
    }

/* let any residual heat in the upper layer melt formed ice */
     for (iy = 0; iy < My; iy++)
     {
      for (ix =0; ix < Mx; ix++)
      {
        if(h[Nz-1][iy][ix] > 1.0)
        {
           potice = (rhoref*Cp/Lathf)*
              h[Nz-1][iy][ix]*(Tf - tx[Nz-1][iy][ix]);

              if( potice < qice[Nz-1][iy][ix])potice=qice[Nz-1][iy][ix];

              tx[Nz-1][iy][ix] +=  potice*
                 Lathf/(h[Nz-1][iy][ix]*rhoref*Cp);
              /*diag8[0][Nz-1][iy][ix] = potice*Lathf/(h[Nz-1][iy][ix]*rhoref*Cp);*/

              sx[Nz-1][iy][ix] += potice*(Sref)/(rhoref*h[Nz-1][iy][ix]);
              /*diag8[1][Nz-1][iy][ix]=potice*(Sref)/(rhoref*h[Nz-1][iy][ix]);*/
              qice[Nz-1][iy][ix] -= potice;
        }
      }
     }

/* tx may still be above freezing so determine melt potential */
     for (iy = 0; iy < My; iy++)
     {
      for (ix = 0; ix < Mx; ix++)
      {
        if(h[Nz-1][iy][ix] > 1.0)
        {

           potice = (rhoref*Cp/Lathf)*
              h[Nz-1][iy][ix]*(Tf - tx[Nz-1][iy][ix]);
           qflx[iy][ix]= Lathf*(potice-qice[Nz-1][iy][ix])/Delt;
        }
      }
     }

/*   end of sea ice */   
/********************************************************** */
	

/* Fourth order space filter */

  filt4m(ux);
  filt4m(vx);
  filt4s(tx);
  filt4s(sx);

/*  Salinity clamp */
/*  Temp warning */

   for (iz =  0; iz < Mz; iz++)
   {
     for (iy = 0; iy < My; iy++)
     {
      for (ix = 0; ix < Mx; ix++)
      {

      if(h[iz][iy][ix] > 1.0){
      if(tx[iz][iy][ix] < Tf){

/* issue a warning to stderr if sub-freezing water is found below ice levels
   and at the bottom of a water column */
      if((iz < Mz-izmxice)  && ((iz == 0) || (h[iz-1][iy][ix] <= 1.0)))
      fprintf(stderr,"OM3 temp warning: %d %d %d %d %d %d %f Reset\n",istep,iz,iy,ix,(row*(My-2))+iy,(col*(Mx-2))+ix,tx[iz][iy][ix]);

      tx[iz][iy][ix] = Tf;
      }

      if(sx[iz][iy][ix] < 0.0){
      fprintf(stderr,"Negative salinity warning: %d %d %d %d %f Reset\n",istep,iz,iy,ix,sx[iz][iy][ix]);
      sx[iz][iy][ix] = 0.0;
      }
      }

      }
     }
   }

/* convective adjustment */
    for (iy = 0; iy < My; iy++) 
    {
      for (ix = 0; ix < Mx; ix++)
      { int mixing, pass_count;

        mixing = 1;
        pass_count = 0;
        while(mixing && (pass_count++ < Nz))
        {
          mixing = 0;
	      for (iz = Mz-2; iz >= 0; iz--)
	      { float d1 ,d2 , snew, tnew;
            if((h[iz][iy][ix] > 1.0) && (h[iz+1][iy][ix]) > 1.0)
            {
               d1 = density(sx[iz][iy][ix],tx[iz][iy][ix],iz);
               d2 = density(sx[iz+1][iy][ix],tx[iz+1][iy][ix],iz);
               if(d2 > (d1 + 1.0e-5))
               {
                  mixing = 1;

                  snew = (sx[iz][iy][ix]*h[iz][iy][ix] +
                          sx[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);

                  diag7[1][iz][iy][ix] += snew - sx[iz][iy][ix];
                  diag7[1][iz+1][iy][ix] += snew - sx[iz+1][iy][ix];

                  sx[iz][iy][ix] = sx[iz+1][iy][ix] = snew;

                  tnew = (tx[iz][iy][ix]*h[iz][iy][ix] +
                          tx[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);

                  diag7[0][iz][iy][ix] += tnew - tx[iz][iy][ix];
                  diag7[0][iz+1][iy][ix] += tnew - tx[iz+1][iy][ix];

                  tx[iz][iy][ix] = tx[iz+1][iy][ix] = tnew;
               if(iz >= Mz/2)
	       convec_freq[iy][ix] += 2;
	       else
	       convec_freq2[iy][ix] += 2;

               }
            }
	      }
	    }
      }
    }



/* Time filter */

   for (iz =  0; iz < Mz; iz++)
   {
     for (iy = 0; iy < My; iy++)
     {
      for (ix = 0; ix < Mx; ix++)
      {
	 ul[iz][iy][ix] = 0.65 * u[iz][iy][ix]
	     + 0.175 * (ul[iz][iy][ix] + ux[iz][iy][ix]);
	 u[iz][iy][ix] = ux[iz][iy][ix];

	 vl[iz][iy][ix] = 0.65 * v[iz][iy][ix]
	     + 0.175 * (vl[iz][iy][ix] + vx[iz][iy][ix]);
	 v[iz][iy][ix] = vx[iz][iy][ix];

	 /*diag0[0][iz][iy][ix] += diag7[0][iz][iy][ix]+diag8[0][iz][iy][ix];
	 diag0[1][iz][iy][ix] += diag7[1][iz][iy][ix]+diag8[0][iz][iy][ix];*/

	 tl[iz][iy][ix] = 0.65 * t[iz][iy][ix]
	     + 0.175 * (tl[iz][iy][ix] + tx[iz][iy][ix]);
	 t[iz][iy][ix] = tx[iz][iy][ix];

	 sl[iz][iy][ix] = 0.65 * s[iz][iy][ix]
	     + 0.175 * (sl[iz][iy][ix] + sx[iz][iy][ix]);
	 s[iz][iy][ix] = sx[iz][iy][ix];
      }
     }
   }

   for (iy = 0; iy < My; iy++)
   {
     for (ix = 0; ix < Mx; ix++)
     {
	    ptopl[iy][ix] = 0.65 * ptop[iy][ix]
	      + 0.175 * (ptopl[iy][ix] + ptopx[iy][ix]);
	    ptop[iy][ix] = ptopx[iy][ix];
      }
    }

}

float density(float s, float t, int iz)
{
   float den,sq,tq;
   int k;

/*
 * use a 3rd order polynomail approx to the UNESCO eqn of
   state.  See Bryan and Cox, 1972.  The coeffs for our
   levels were generated from a utility program provided with
   the MOM2 code.
*/

 /* the coefficients all assume that the top of the
 ocean is level 0 */
   k=(Nz-1)-iz;

 /* convert s to MOM units */
   sq=(s-35.)/1000.;

 /*subtract base */
   sq=sq-so[k];
   tq=t-to[k];

   den = (c[k][0] + (c[k][3] + c[k][6]* sq)* sq +
	 (c[k][2] + c[k][7]* sq + c[k][5]* tq)* tq)* tq +
	 (c[k][1] + (c[k][4] + c[k][8]* sq)* sq)* sq;

   return den;

/*  The old equation
   den = 0.2*(s-base_s[iz]) - 0.2*(t-base_t[iz]);
   return den/rhoref; */
}

void filt4s(float q[Mz][My][Mx])
{
  int ix,iy,iz;
  float damp,work[Mz][My][Mx];

  damp = 0.5/32.0;

 /* Fourth order filter on scalars - pass 1*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         work[iz][iy][ix] = 0.0;

         if(h[iz][iy][ix] > 1.0)
         {
            if(h[iz][iy][ix-1] > 1.0)
            {
              work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix-1]);
            }

	        if(h[iz][iy][ix+1] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix+1]);
            }

            if(h[iz][iy-1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy-1][ix]);
            }

	        if(h[iz][iy+1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy+1][ix]);
            }
         }
      }
   }
 }

 wrap_qz(work);

 /* Fourth order conservative dissipation on scalars - pass2 */

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(h[iz][iy][ix] > 1.0)
         {
            if(h[iz][iy][ix-1] > 1.0)
            {
              q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix-1]);
            }

	        if(h[iz][iy][ix+1] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix+1]);
            }

            if(h[iz][iy-1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy-1][ix]);
            }

	        if(h[iz][iy+1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy+1][ix]);
            }
         }
      }
   }
 }
wrap_qz(q);
}

void filt4m(float q[Mz][My][Mx])
{
  int ix,iy,iz;
  float damp,work[Mz][My][Mx];

  damp = 0.5/32.0;

 /* Fourth order filter on momentum - pass 1*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         work[iz][iy][ix] = 0.0;

         if(uvh[iz][iy][ix] > 1.0)
         {
            if(uvh[iz][iy][ix-1] > 1.0)
            {
              work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix-1]);
            }

	        if(uvh[iz][iy][ix+1] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix+1]);
            }

            if(uvh[iz][iy-1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy-1][ix]);
            }

	        if(uvh[iz][iy+1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy+1][ix]);
            }
         }
      }
   }
 }

 wrap_qz(work);

 /* Fourth order conservative dissipation on momentum - pass2 */

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         {
            if(uvh[iz][iy][ix-1] > 1.0)
            {
              q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix-1]);
            }

	        if(uvh[iz][iy][ix+1] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix+1]);
            }

            if(uvh[iz][iy-1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy-1][ix]);
            }

	        if(uvh[iz][iy+1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy+1][ix]);
            }
         }
      }
   }
 }
wrap_qz(q);
}

void filt4p(float q[My][Mx])
{
  int ix,iy;
  float damp,work[My][Mx];

  damp = 0.5/32.0;

 /* Fourth order filter on pressure - pass 1*/

   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         work[iy][ix] = 0.0;

         if(depth[iy][ix] > 1.0)
         {
            if(depth[iy][ix-1] > 1.0)
            {
              work[iy][ix] -= (q[iy][ix] - q[iy][ix-1]);
            }

	        if(depth[iy][ix+1] > 1.0)
            {
	          work[iy][ix] -= (q[iy][ix] - q[iy][ix+1]);
            }

            if(depth[iy-1][ix] > 1.0)
            {
	          work[iy][ix] -= (q[iy][ix] - q[iy-1][ix]);
            }

	        if(depth[iy+1][ix] > 1.0)
            {
	          work[iy][ix] -= (q[iy][ix] - q[iy+1][ix]);
            }
         }
      }
   }

 wrap_q(work);

 /* Fourth order conservative dissipation on pressure - pass2 */

   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(depth[iy][ix] > 1.0)
         {
            if(depth[iy][ix-1] > 1.0)
            {
              q[iy][ix] += damp * (work[iy][ix] - work[iy][ix-1]);
            }

	        if(depth[iy][ix+1] > 1.0)
            {
	          q[iy][ix] += damp * (work[iy][ix] - work[iy][ix+1]);
            }

            if(depth[iy-1][ix] > 1.0)
            {
	          q[iy][ix] += damp * (work[iy][ix] - geo1[iy]*work[iy-1][ix]);
            }

	        if(depth[iy+1][ix] > 1.0)
            {
	          q[iy][ix] += damp * (work[iy][ix] - geo2[iy]*work[iy+1][ix]);
            }
         }
      }
   }
wrap_q(q);
}


/* Begin sp2 communications routines. */

float in_xz[Mz][Mx], out_xz[Mz][Mx];
float in_yz[Mz][My], out_yz[Mz][My];

void wrap_qz(float a[Mz][My][Mx])
{
   int ix, iy, iz;
   int tag=0;

/* rec from south, send north */
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       out_xz[iz][ix] = a[iz][My-2][ix];
     }
   }
   send_and_receive(south, tag, in_xz, Mx*Mz*sizeof(float), north, out_xz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       a[iz][0][ix] = in_xz[iz][ix];
     }
   }

/* rec from north, send south */
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       out_xz[iz][ix] = a[iz][1][ix];
     }
   }
   send_and_receive(north, tag, in_xz, Mx*Mz*sizeof(float), south, out_xz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       a[iz][My-1][ix] = in_xz[iz][ix];
     }
   }

/* rec from west, send east */
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       out_yz[iz][iy] = a[iz][iy][Mx-2];
     }
   }
   send_and_receive(west, tag, in_yz, My*Mz*sizeof(float), east, out_yz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       a[iz][iy][0] = in_yz[iz][iy];
     }
   }

/* rec from east, send west */
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       out_yz[iz][iy] = a[iz][iy][1];
     }
   }
   send_and_receive(east, tag, in_yz, My*Mz*sizeof(float), west, out_yz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       a[iz][iy][Mx-1] = in_yz[iz][iy];
     }
   }
}

float in_3x[3][Mx], out_3x[3][Mx];
float in_3y[3][My], out_3y[3][My];

void wrap_q3(float a[My][Mx], float b[My][Mx], float c[My][Mx])
{
   int ix, iy;
   int tag=0;

/* rec from south, send north */
   for(ix = 0; ix < Mx; ix++)
   {
     out_3x[0][ix] = a[My-2][ix];
     out_3x[1][ix] = b[My-2][ix];
     out_3x[2][ix] = c[My-2][ix];
   }
   send_and_receive(south, tag, in_3x, Mx*3*sizeof(float), north, out_3x);
   for(ix = 0; ix < Mx; ix++)
   {
     a[0][ix] = in_3x[0][ix];
     b[0][ix] = in_3x[1][ix];
     c[0][ix] = in_3x[2][ix];
   }

/* rec from north, send south */
   for(ix = 0; ix < Mx; ix++)
   {
     out_3x[0][ix] = a[1][ix];
     out_3x[1][ix] = b[1][ix];
     out_3x[2][ix] = c[1][ix];
   }
   send_and_receive(north, tag, in_3x, Mx*3*sizeof(float), south, out_3x);
   for(ix = 0; ix < Mx; ix++)
   {
     a[My-1][ix] = in_3x[0][ix];
     b[My-1][ix] = in_3x[1][ix];
     c[My-1][ix] = in_3x[2][ix];
   }

/* rec from west, send east */
   for(iy = 0; iy < My; iy++)
   {
     out_3y[0][iy] = a[iy][Mx-2];
     out_3y[1][iy] = b[iy][Mx-2];
     out_3y[2][iy] = c[iy][Mx-2];
   }
   send_and_receive(west, tag, in_3y, My*3*sizeof(float), east, out_3y);
   for(iy = 0; iy < My; iy++)
   {
     a[iy][0] = in_3y[0][iy];
     b[iy][0] = in_3y[1][iy];
     c[iy][0] = in_3y[2][iy];
   }

/* rec from east, send west */
   for(iy = 0; iy < My; iy++)
   {
     out_3y[0][iy] = a[iy][1];
     out_3y[1][iy] = b[iy][1];
     out_3y[2][iy] = c[iy][1];
   }
   send_and_receive(east, tag, in_3y, My*3*sizeof(float), west, out_3y);
   for(iy = 0; iy < My; iy++)
   {
     a[iy][Mx-1] = in_3y[0][iy];
     b[iy][Mx-1] = in_3y[1][iy];
     c[iy][Mx-1] = in_3y[2][iy];
   }
}

void wrap_q(a)
float a[My][Mx];
{
   int src, dest, msg_len, tag=0, ellen, elcnt, stride;
   char *msg_in, *msg_out;

   msg_len = Mx * sizeof(float);

   msg_in = (void *) &(a[0][0]); /* rec from south, send north */
   msg_out = (void *) &(a[My-2][0]);

   send_and_receive(south, tag, msg_in, msg_len, north, msg_out);

   msg_in = (void *) &(a[My-1][0]); /* rec from north, send south */
   msg_out = (void *) &(a[1][0]);

   send_and_receive(north, tag, msg_in, msg_len, south, msg_out);

   msg_in = (void *) &(a[0][0]);  /* rec from west, send east */
   msg_out = (void *) &(a[0][Mx-2]);
/* 
 * defined in mpi_edge_type
 *
 * ellen = sizeof(float);
 * elcnt = My;
 * stride = Mx * ellen;
*/

   send_and_receive_v(west, tag, msg_in, ellen, stride, elcnt, east, msg_out);

   msg_in = (void *) &(a[0][Mx-1]);
   msg_out = (void *) &(a[0][1]);
/* 
 * defined in mpi_edge_type
 *
 * ellen = sizeof(float);
 * elcnt = My;
 * stride = Mx * ellen;
*/

   send_and_receive_v(east, tag, msg_in, ellen, stride, elcnt, west, msg_out);
}

send_and_receive(
      int source_node, int tag, void *inbuffer, int buf_len,
      int dest_node, void *outbuffer)
{
    MPI_Status status;
    int result;

#ifdef DEBUG
    if(source_node != dest_node) {
	printf("SEND_RECEIVE: Source and dest node differ: %d %d\n",
	       source_node, dest_node);
	exit(1);
    }
#endif

    result = MPI_Sendrecv(outbuffer, buf_len, MPI_CHAR, dest_node, tag,
			  inbuffer, buf_len, MPI_CHAR, source_node, tag,
			  split_comm, &status);

    if(result != MPI_SUCCESS) {
	printf("SEND_RECEIVE: Return code %d\n",result);
	exit(1);
    }
}

send_and_receive_v(
      int source_node, int tag, void *inbuffer,
      int elem_len, int stride, int elem_count,
      int dest_node, void *outbuffer)
{
    MPI_Status status;
    int result;

#ifdef DEBUG
    if(source_node != dest_node) {
	printf("SEND_RECEIVE_V: Source and dest node differ: %d %d\n",
	       source_node, dest_node);
	exit(1);
    }
#endif

    result = MPI_Sendrecv(outbuffer, 1, mpi_edge_type, dest_node, tag,
			  inbuffer, 1, mpi_edge_type, source_node, tag,
			  split_comm, &status);

    if(result != MPI_SUCCESS) {
	printf("SEND_RECEIVE_V: Return code %d\n",result);
	exit(1);
    }
}

/* End sp2 communications routines */




int global_open(char *file, int flags, int mode)
{
    int fd;

    if(my_address == 0) {
	if((fd = open(file, flags, mode)) < 0) {
	    printf("File open error: %s\n",file);
	    exit(1);
	}
	printf("File open: %s\n",file);
	return(fd);
    }
    else
	return(-1);
}


/* Begin sp2 timing routines. */
/*********** DUMMY ************************************************/
int timers_initialized = 0;

void node_timer_clear(int n)
{
    if(!timers_initialized) {
      UTP_init(100);
      timers_initialized = 1;
    }
    UTP_reset_timer(n);
}

void node_timer_start(int n)
{
    if(!timers_initialized) {
      UTP_init(100);
      timers_initialized = 1;
    }
    UTP_start_timer(n);
}

void node_timer_stop(int n)
{
    if(!timers_initialized) {
      UTP_init(100);
      timers_initialized = 1;
    }
    UTP_stop_timer(n);
}

float node_timer_busy(int n)
{
    double time;
    float result;
    int precision;
    UTP_get_accum_time(n, &time, &precision);
    result = (float) time;
    return(result);
}
/* End sp2 timing routines */

/* Routines for the inclusion of the reading of sst in an annual cycle */

#ifdef ANNUAL_SST

void read_sst_data()

{
  int iy,ix,month,advance;
  FILE *sstfile;
  FILE *sssfile;
  FILE *sssom3file;
  FILE *tauxfile;
  FILE *tauyfile;
  FILE *heatfile;
  FILE *saltfile;

  sprintf(buff,"%s%s",time_inv,"/sstshea5c");
  sstfile = fopen(buff,"r");
  sprintf(buff,"%s%s",time_inv,"/sssom3c");
  sssom3file = fopen(buff,"r");
  sprintf(buff,"%s%s",time_inv,"/tauxhcf");
  tauxfile = fopen(buff,"r");
  sprintf(buff,"%s%s",time_inv,"/tauyhcf");
  tauyfile = fopen(buff,"r");
  sprintf(buff,"%s%s",time_inv,"/sssmodm3ph2c");
  /*sssfile = fopen(buff,"r");
  sprintf(buff,"%s%s",time_inv,"/saltmask3c");
  saltfile = fopen(buff,"r");
  sprintf(buff,"%s%s",time_inv,"/heatmask3c");
  heatfile = fopen(buff,"r");*/

  for (month=0; month < 12; month++)
  {
    advance = (month*Nx*Ny) + (col*(Mx-2)+row*(My-2)*Nx);
    fseek(sstfile,advance*sizeof(float),SEEK_SET);
    if(month==0)fseek(sssom3file,advance*sizeof(float),SEEK_SET);
    fseek(tauxfile,advance*sizeof(float),SEEK_SET);
    fseek(tauyfile,advance*sizeof(float),SEEK_SET);
    /*fseek(sssfile,advance*sizeof(float),SEEK_SET);
    fseek(heatfile,advance*sizeof(float),SEEK_SET);
    fseek(saltfile,advance*sizeof(float),SEEK_SET);*/
    for (iy = 0; iy < My-2; iy++)
    {
      fread(&(sstm[month][iy][0]),sizeof(float),Mx-2,sstfile);
      if(month==0)fread(&(sssom3[iy][0]),sizeof(float),Mx-2,sssom3file);
      fread(&(tauxm[month][iy][0]),sizeof(float),Mx-2,tauxfile);
      fread(&(tauym[month][iy][0]),sizeof(float),Mx-2,tauyfile);

      /*fread(&(sssm[month][iy][0]),sizeof(float),Mx-2,sssfile);
      fread(&(heatm[month][iy][0]),sizeof(float),Mx-2,heatfile);
      fread(&(saltm[month][iy][0]),sizeof(float),Mx-2,saltfile);*/
      advance = Nx - (Mx-2);
      fseek(sstfile,advance*sizeof(float),SEEK_CUR);
      if(month==0)fseek(sssom3file,advance*sizeof(float),SEEK_CUR);
      fseek(tauxfile,advance*sizeof(float),SEEK_CUR);
      fseek(tauyfile,advance*sizeof(float),SEEK_CUR);
/*
      fseek(sssfile,advance*sizeof(float),SEEK_CUR);
      fseek(heatfile,advance*sizeof(float),SEEK_CUR);
      fseek(saltfile,advance*sizeof(float),SEEK_CUR);*/
    }
  }
  fclose(sstfile);
  fclose(sssom3file);
  fclose(tauxfile);
  fclose(tauyfile);
/*
  fclose(sssfile);
  fclose(heatfile);
  fclose(saltfile);*/
}


float sstint(int j, int i, int timestep)

{
  int prevstep, nextstep;
  float sst_value;
  float fact;
  int month, sid;

  sid = 86400 / Delt;
  timestep = timestep - 15*sid;
  if (timestep < 0) {
    timestep = timestep + 360*sid;
  }
  timestep = timestep % (360 * sid);
  month = timestep / (30 * sid);
  prevstep = month * 30 * sid;
  nextstep = prevstep + 30 * sid;

  fact = ((float)timestep - (float)prevstep) / ((float)nextstep - (float)prevstep);
  sst_value = sstm[month][i][j] + fact * (sstm[(month+1)%12][i][j] - sstm[month][i][j]);
  return(sst_value);
}
float sssint(int j, int i, int timestep)

{
  int prevstep, nextstep;
  float sss_value;
  float fact;
  int month, sid;

  sid = 86400 / Delt;
  timestep = timestep - 30*sid;
  if (timestep < 0) {
    timestep = timestep + 360*sid;
  }
  timestep = timestep % (360 * sid);
  month = timestep / (30 * sid);
  prevstep = month * 30 * sid;
  nextstep = prevstep + 30 * sid;

  fact = ((float)timestep - (float)prevstep) / ((float)nextstep - (float)prevstep);
  sss_value = sssm[month][i][j] + fact * (sssm[(month+1)%12][i][j] - sssm[month][i][j]);
  return(sss_value);
}

float tauxint(int j, int i, int timestep)

{
  int prevstep, nextstep;
  float taux_value;
  float fact;
  int month, sid;

  sid = 86400 / Delt;
  timestep = timestep - 15*sid;
  if (timestep < 0) {
    timestep = timestep + 360*sid;
  }
  timestep = timestep % (360 * sid);
  month = timestep / (30 * sid);
  prevstep = month * 30 * sid;
  nextstep = prevstep + 30 * sid;

  fact = ((float)timestep - (float)prevstep) / ((float)nextstep - (float)prevstep);
  taux_value = tauxm[month][i][j] + fact * (tauxm[(month+1)%12][i][j] - tauxm[month][i][j]);
  return(taux_value);
}
float tauyint(int j, int i, int timestep)

{
  int prevstep, nextstep;
  float tauy_value;
  float fact;
  int month, sid;

  sid = 86400 / Delt;
  timestep = timestep - 15*sid;
  if (timestep < 0) {
    timestep = timestep + 360*sid;
  }
  timestep = timestep % (360 * sid);
  month = timestep / (30 * sid);
  prevstep = month * 30 * sid;
  nextstep = prevstep + 30 * sid;

  fact = ((float)timestep - (float)prevstep) / ((float)nextstep - (float)prevstep);
  tauy_value = tauym[month][i][j] + fact * (tauym[(month+1)%12][i][j] - tauym[month][i][j]);
  return(tauy_value);
}
float heatint(int j, int i, int timestep)

{
  int prevstep, nextstep;
  float heat_value;
  float fact;
  int month, sid;

  sid = 86400 / Delt;
  timestep = timestep - 15*sid;
  if (timestep < 0) {
    timestep = timestep + 360*sid;
  }
  timestep = timestep % (360 * sid);
  month = timestep / (30 * sid);
  prevstep = month * 30 * sid;
  nextstep = prevstep + 30 * sid;

  fact = ((float)timestep - (float)prevstep) / ((float)nextstep - (float)prevstep);
  heat_value = heatm[month][i][j] + fact * (heatm[(month+1)%12][i][j] - heatm[month][i][j]);
  return(heat_value);
}

float saltint(int j, int i, int timestep)

{
  int prevstep, nextstep;
  float salt_value;
  float fact;
  int month, sid;

  sid = 86400 / Delt;
  timestep = timestep - 15*sid;
  if (timestep < 0) {
    timestep = timestep + 360*sid;
  }
  timestep = timestep % (360 * sid);
  month = timestep / (30 * sid);
  prevstep = month * 30 * sid;
  nextstep = prevstep + 30 * sid;

  fact = ((float)timestep - (float)prevstep) / ((float)nextstep - (float)prevstep);
  salt_value = saltm[month][i][j] + fact * (saltm[(month+1)%12][i][j] - saltm[month][i][j]);
  return(salt_value);
}

#endif

