#include "params.h"
      subroutine srfint(lat     ,ioro    ,psm1    ,sndpth  ,ubot    ,
     $                  vbot    ,tbot    ,qbot    ,thbot   ,zbot    ,
     $                  pmidm1  ,srfrad  ,ts      ,tssub   ,shflx   ,
     $                  lhflx   ,qflx    ,ws      ,taux    ,tauy    )
C-----------------------------------------------------------------------
C 
C Compute temperatures for land/sea-ice surfaces for specified radiative
C heating and for current atmospheric conditions.  In addition, for both
C this routine, as well as later use in the atmospheric boundary
C layer/diffusion calculations, the sensible/latent heat fluxes, stress
C components, and exchange coefficiants are computed.
C 
C Temperatures over land/sea-ice surfaces are specified in 'plevmx'
C layers of fixed thickness and thermal properties.  The forecast
C temperatures are determined from a backward/implicit diffusion
C calculation using linearized sensible/latent heat fluxes.  Over land,
C a zero bottom boundary flux condition is assumed; over sea-ice, the
C bottom ocean temperature is fixed at -2C, allowing heat flux exchange
C with underlying ocean.
C 
C For the standard fixed surface hydrology, the top layer forecast
C temperature over snow covered land, as well as all layers over
C permanent land ice and sea-ice, are not allowed to exceed melting
C temperature.
C 
C Note that qflx is the surface water flux, and that surface fluxes of
C all other constituents (if present) are currently ignored.
C
C Note:  CCM1 'TSCALC'  Modified for CCM2 by B. Briegleb
C
C---------------------------Code history--------------------------------
C
C Original version:  CCM1
C Standardized:      L. Buja, June 1992
C Reviewed:          B. Briegleb, J. Hack, August 1992
C
C-----------------------------------------------------------------------
      implicit none
#if ( defined RS6000 )
c      implicit automatic (a-z)
#endif
C------------------------------Parameters-------------------------------
#include "pmgrid.com"
C------------------------------Commons----------------------------------
#include "comctl.com"
C-----------------------------------------------------------------------
#include "comtim.com"
C-----------------------------------------------------------------------
#include "comtsc.com"
C-----------------------------------------------------------------------
#include "crdalb.com"
C------------------------------Arguments--------------------------------
C
C Input arguments
C
      integer lat               ! Latitude index
      integer ioro(plond)       ! Integer array for oro (surface type)
      real psm1(plond),         ! Surface pressure
     $     sndpth(plond),       ! Snow depth (liquid water equivalent)
     $     ubot(plond),         ! Bottom level u wind
     $     vbot(plond),         ! Bottom level v wind
     $     tbot(plond),         ! Bottom level temperature
     $     qbot(plond),         ! Bottom level specific humidity
     $     thbot(plond),        ! Bottom level potential temperature
     $     zbot(plond),         ! Bottom level height above surface
     $     pmidm1(plond,plev),  ! Level pressures
     $     srfrad(plond)        ! Srf solar abs flux plus down longwave
C
C Input/output arguments
C
      real ts(plond),           ! Surface temperature
     $     tssub(plond,plevmx)  ! Surface/sub-surface temperatures
C
C Output arguments
C
      real shflx(plond),       ! Surface sensible heat flux (J/m2/s)
     $     lhflx(plond),       ! Surface latent   heat flux (J/m2/s)
     $     qflx(plond),        ! Surface water flux (kg/m2/s)
     $     ws(plond),          ! Soil moisture (m)
     $     taux(plond),        ! X surface stress (N)
     $     tauy(plond)         ! Y surface stress (N)
C
C---------------------------Local variables-----------------------------
C
      integer i,               ! Longitude index
     $        k,               ! Sub-surface level index
     $        isrftyp(plond)   ! Integer type for surface
C
      real delts(plond),       ! Forecast minus original surface temp
     $     dfntdt(plond),      ! ts partial derivative of net srf flx
     $     dlhf(plond),        ! ts partial deriv for latent heat flux
     $     dshf(plond),        ! ts partial deriv for sensible heat flux
     $     fnt(plond),         ! Net surface flux for input conditions
     $     frsnow,             ! Fractional snow cover (horizontal area)
     $     lhf(plond),         ! Init estimate of sfc latent heat flux
     $     ltheat(plond),      ! Latent heat for given sfc conditions
     $     rghsnw,             ! Roughness used for fraction snow comp
     $     shf(plond),         ! Init estimate of sfc sensible heat flux
     $     snwhgt,             ! Snow height (physical height)             
     $     tsbsf(plond,plevmx),! Non-adjusted srfc/sub-srfc temperatures
     $     tsin(plond),        ! Initial surface temperature
     $     wet(plond)          ! Sfc wetness factor(fract saturated sfc)
C
C------------------------------Externals--------------------------------
C
      external srfflx,         ! Surface flux computation
     $         srftsb          ! Sub-surface temperature computation
C
C-----------------------------------------------------------------------
C
C 
C For fixed surface hydrology, adjust surface wetness for snow cover;
C calculation of fraction of snow cover is identical to that in sfc
C albedo routine 'radalb'; see that routine for further reference.
C
CDIR$ IVDEP
       do i=1,plon
          if(sndpth(i) .le. 0.) then
             frsnow    = 0.0
          else
C
C snwedp is *2 here to match plx18 control. This is inconsistent with
C srftsb.for.  This 2*snwedp also occurs in radalb.for
C
            snwhgt    = 2*snwedp*sndpth(i)
            rghsnw    = amax1(rghnss(i,lat),0.25)
            frsnow    = snwhgt/(rghsnw + snwhgt)
         end if
         vevapf(i,lat) = frsnow + (1.-frsnow)*evapf(i,lat)
         wet(i)        = vevapf(i,lat)
      end do
C
C Define surface type
C
      do i=1,plon
         if (ioro(i).eq.0) then   
            isrftyp(i) = 1
         else if (ioro(i).eq.2) then
            isrftyp(i) = 2            ! Account for variable sea-ice
         else                        
            isrftyp(i) = int(vegtyp(i,lat))
         end if
C  
C Set latent heat for evaporation over snow covered or sea-ice sfcs
C
         if ( sndpth(i).gt.0. .or. ioro(i).eq.2 ) then
            ltheat(i) = latvap + latice
         else
            ltheat(i) = latvap
         end if
      end do
C
C Compute surface fluxes, derviatives, and exchange coefficiants
C
      call srfflx(lat     ,ioro    ,psm1    ,pmidm1  ,ubot    ,
     $            vbot    ,tbot    ,qbot    ,thbot   ,zbot    ,
     $            srfrad  ,ts      ,ltheat  ,wet     ,fnt     ,
     $            dfntdt  ,shf     ,dshf    ,lhf     ,dlhf    ,
     $            taux    ,tauy    )
C
C Compute surface/subsurface temperatures
C
      do k=1,plevmx
         do i=1,plon
            tsin(i)    = ts(i) 
            tsbsf(i,k) = tssub(i,k)
         end do
      end do
C
      call srftsb(isrftyp ,fnt     ,dfntdt  ,sndpth ,tsbsf   )
C
C 
C Fixed surface hydrology modification to surface temperatures: if over
C land covered with snow, and ts exceeds tmelt, reset to tmelt:
C
      do i=1,plon
         if(isrftyp(i).ge.4 .and. sndpth(i) .gt. 0. ) then
            tsbsf(i,1) = amin1(tsbsf(i,1),tmelt)
         end if
      end do
C 
C Modification to updated surface temperatures: if either sea-ice(2)or
C land (permanent) ice (3), and if temperatures of any layer is over
C melting point, reset temperature to melting point.
C
      do i=1,plon
         if(isrftyp(i).eq.2 .or. isrftyp(i).eq.3) then
            do k=1,plevmx
               tsbsf(i,k) = amin1(tsbsf(i,k),tmelt)
            end do
         end if
C
C Update surface temperature over land/ice surfaces only:
C
         if( isrftyp(i) .gt. 1 ) then
           ts(i) = tsbsf(i,1)
         end if
         delts(i) = ts(i) - tsin(i)
      end do
C
      do  k=1,plevmx
         do i=1,plon
            if( isrftyp(i) .gt. 1 ) then
              tssub(i,k) = tsbsf(i,k)
            end if
         end do
      end do
C
C Fixed surface hydrology
C
       do  i=1,plon
          ws(i) = vevapf(i,lat)/qwssat
       end do
C
C Evaluate sensible and latent heat fluxes using forecasted ts
C
CDIR$ IVDEP
       do i=1,plon
         shflx(i)  = shf(i) + dshf(i)*delts(i)
         lhflx(i)  = lhf(i) + dlhf(i)*delts(i)
         qflx(i)   = lhflx(i)/ltheat(i)
       end do
C    
      return
      end
