#include "params.h"
      subroutine settau(zdt)
C-----------------------------------------------------------------------
C
C Set time invariant hydrostatic matrices, which depend on the reference
C temperature and pressure in the semi-implicit time step. Note that
C this subroutine is actually called twice, because the time step
C changes between step 0 and step 1.
C
C zdt = delta t for next semi-implicit time step.
C
C---------------------------Code history--------------------------------
C
C Original version:  CCM1
C Standardized:      J. Rosinski, June 1992
C Reviewed:          B. Boville, D. Williamson,  August 1992
C
C-----------------------------------------------------------------------
      implicit none
#if ( defined RS6000 )
c      implicit automatic (a-z)
#endif
C------------------------------Parameters-------------------------------
#include "pmgrid.com"
#include "pspect.com"
C------------------------------Commons----------------------------------
#include "comcon.com"
C-----------------------------------------------------------------------
#include "comhyb.com"
C-----------------------------------------------------------------------
#include "commap.com"
C-----------------------------------------------------------------------
#include "comspe.com"
C------------------------------Arguments--------------------------------
C
C Input arguments
C
      real zdt                   ! time step (or dt/2 at time 0)
C
C---------------------------Local workspace-----------------------------
C
      real aq(plev,plev)
#if ( defined CRAY )
      real work(plev,2),         ! work array for matrix inverter
     $     zdtr                  ! used by minv
#else
      real rcond,z(plev),det(2),work(plev)
      integer ipvt(plev)
#endif
      real zcr(plev),            ! gravity wave equivalent depth
     $     zci(plev),            ! dummy, used to print phase speeds
     $     zdt2,                 ! zdt**2
     $     factor,               ! intermediate workspace
     $     zdt0u,                ! vertical diff. of ref. temp (above)
     $     zshu,                 ! interface "sigma" (above)
     $     zr2ds,                ! 1./(2.*hypd(k))
     $     zdt0d,                ! vertical diff. of ref. temp (below)
     $     zshd,                 ! interface "sigma" (below)
     $     ztd,                  ! temporary accumulator
     $     zcn                   ! sq(n)
      real zb(plev,plev)         ! semi-implicit matrix in d equation
      integer
     $     k,kk,kkk,             ! level indices
     $     n,                    ! n-wavenumber index
     $     nneg                  ! number of unstable mean temperatures
C
C-----------------------------------------------------------------------
C
      zdt2 = zdt*zdt
C
C Set mean temperature
C NOTE: Making t0 an actual function of height ***DOES NOT WORK***
C
      do k=1,plev
         t0(k) = 300.
      end do
C
C Calculate hydrostatic matrix tau using the formulation of the hydrostatic
C integral from CCM1 converted to hybrid coordinate form
C
      zdt0u = 0.
      zshu = 0.
      do k=1,plev
         zr2ds = 1./(2.*hypd(k))
         if (k.lt.plev) then
            zdt0d = t0(k+1) - t0(k)
            zshd = hybi(k+1)
         else
            zdt0d = 0.
            zshd = 0.
         end if
C
         factor = ((zdt0u*zshu + zdt0d*zshd) - (zdt0d + zdt0u))*zr2ds
         do kk=1,k-1
            tau(kk,k) = factor*hypd(kk) + cappa*t0(k)*ecref(kk,k)
         end do
C
         factor = (zdt0u*zshu + zdt0d*zshd - zdt0d)*zr2ds
         tau(k,k) = factor*hypd(k) + cappa*t0(k)*ecref(k,k)
C
         factor = (zdt0u*zshu + zdt0d*zshd)*zr2ds
         do kk=k+1,plev
            tau(kk,k) = factor*hypd(kk)
         end do
         zdt0u = zdt0d
         zshu = zshd
      end do
C
C Vector for linear surface pressure term in divergence
C Pressure gradient and diagonal term of hydrostatic components
C
      do k=1,plev
         bps(k) = t0(k)
         bps(k) = bps(k)*rair
      end do
      do k=1,plev
         do kk=1,plev
            ztd = bps(k) * hypd(kk)/hypi(plevp)
            do kkk=1,plev
               ztd = ztd + href(kkk,k)*tau(kk,kkk)
            end do
            zb(kk,k) = ztd
            aq(kk,k) = ztd
         end do
      end do
C
C Compute and print gravity wave equivalent depths and phase speeds
C
      call qreig(zb      ,plev    ,plev    ,plev    ,zcr     ,
     $           zci     )
      do k=1,plev
         zci(k) = sign(1.,zcr(k))*sqrt(abs(zcr(k)))
         zcr(k) = zcr(k) / gravit
      end do
      write(6,910) (t0(k),k=1,plev)
      write(6,920) (zci(k),k=1,plev)
      write(6,930) (zcr(k),k=1,plev)
C
C Test for unstable mean tempuratures (negative phase speed and eqivalent
C depth) for at least one gravity wave.
C
      nneg = 0
      do k=1,plev
         if (zcr(k).le.0.) nneg = nneg + 1
      end do
      if (nneg.ne.0) then
         write(6,*)'---------------------------------------------------'
         write(6,*)'SETTAU: UNSTABLE MEAN TEMPERATURE. STOP'
         call endrun
      end if
C
C Compute and invert matrix a(n)=(i+sq*b*delt**2)
C
      do k=1,plev
         do kk=1,plev
            aq(kk,k) = aq(kk,k)*zdt2
            bm1(kk,k,1) = 0.
         end do
      end do
      do n=2,pnmax
         zcn = sq(n)
         do k=1,plev
            do kk=1,plev
               zb(kk,k) = zcn*aq(kk,k)
               if(kk.eq.k) zb(kk,k) = zb(kk,k) + 1.
            end do
         end do
C
#if ( defined CRAY )
         call minv(zb,plev,plev,work,zdtr,1.0e-100,0,1)
#else
         call dgeco(zb,plev,plev,ipvt,rcond,z)
         call dgedi(zb,plev,plev,ipvt,det,work,01)
#endif
C
         do k=1,plev
            do kk=1,plev
               bm1(kk,k,n) = zb(kk,k)
            end do
         end do
      end do
      return
C
C Formats
C 
  910 format(' REFERENCE TEMPERATURES FOR SEMI-IMPLICIT SCHEME = ',
     $       /(1x,12f9.3))
  920 format(' GRAVITY WAVE PHASE SPEEDS (M/S) FOR MEAN STATE = '
     $       /(1x,12f9.3))
  930 format(' GRAVITY WAVE EQUIVALENT DEPTHS (M) FOR MEAN STATE = '
     $       /(1x,12f9.3))
      end
