#include "params.h"
c Series of routines to pack the buffer that will be used to asynchronously
c write the restart data to file.   The packing is done in a special way
c to allow reassembly of the data from individual processors simply by
c concatenation of each of the files.  Thus the model can be restarted
c on a different number of processors, because reassembly data is coded
c in with the restart data.
c
c This particular design was chosen because it pushes most of the work
c of disassembly and reassemply back onto the instance of  the program
c that is restarting rather than the program writing restart data as it
c runs.  Reassembly of the data by a restarting program is intended to be
c more time consuming than disassemply by the writing program.
c
c This is done by reorganizing the data into columns, each of which is pre
c pended with four word descriptor containing:
c
c    sn		the section number (arbitrary tag)
c    nwords	number of words (4 byte) to follow
c    glon	global longitude of this column
c    glat	global latitude of this column
c
c The buffer to be packed is passed in to these routines as buf, and
c assumed to be organized as follows:
c    
c      buf( local longitude + pad, level, field )
c
c Latitude is passed in to the routines, which handle packing one latitude
c at a time.  The resulting buffer has the following organization:
c
c      rstrt_buf( field/level , lon, lat )
c
c Assuming the package has been initialized with the proper column length,
c all columns will be in contiguous memory when everything has been packed.
c The set of routines are initialized by pck_rstrt_init(), which establishes
c the length of each column in the restart data set.  This routine also
c resets the m,n array of pointers to the beginning of each column.
c
c The routines pck_rstrt_buf and pck_rstrt_rep are called to pack the
c data into the buffer.  The difference is that pck_rstrt_buf writes
c a different i,j value into the appropriate column of the restart 
c dataset.  Pck_rstrt_buf replicates a value into all columns.
c 
      
ccccccccccccccccccc

       subroutine pck_rstrt_buf( lat, buf, nflds, pad )
       implicit none
#include "pmgrid.com"
#include "pagrid.com"
#include "pspect.com"
#include "comparll.com"
#include "restart_fil.com"
#include "restart_buf.com"
       real buf(*)
       integer lon, lat, nflds, pad

#ifdef PP_ERROR
       write(0,*)'pck_rstrt_buf: ',lat,nflds,pad
#endif

       call pck_rstrt_buf1( lat, buf, rstrt_buf, ip,
     $                      clen, nflds, plon, pad, plat )
       return
       end

       subroutine pck_rstrt_buf1( lat, buf, rstrt_buf, ip,
     $                            clen, nflds, plon, pad, plat )
       implicit none
       integer lat, clen, nflds, plon, pad, plat
       real buf(*), rstrt_buf( clen, plon, plat )
       integer ip( plon, plat )
       integer lp, field, lon
       do lon = 1, plon
           lp = lon
           do field = 1, nflds
              rstrt_buf(ip(lon,lat),lon,lat) = buf(lp)
              ip(lon,lat) = ip(lon,lat) + 1
              lp = lp + plon + pad
           enddo
       enddo
       return
       end

ccccccccccccccccccc

       subroutine pck_rstrt_rep( lat, item )
       implicit none
#include "pmgrid.com"
#include "pagrid.com"
#include "pspect.com"
#include "comparll.com"
#include "restart_fil.com"
#include "restart_buf.com"
       integer lat
       real item
#ifdef PP_ERROR
       write(0,*)'pck_rstrt_rep ',lat,item
#endif
       call pck_rstrt_rep1( lat, item, rstrt_buf, ip,
     $                      clen, 1, plon, plat )
       return
       end

       subroutine pck_rstrt_rep1( lat, buf, rstrt_buf, ip,
     $                            clen, nflds, plon, plat )
       implicit none
       integer lat, clen, nflds, plon, plat
       real buf(*), rstrt_buf( clen, plon, plat )
       integer ip( plon, plat )
       integer field, lon
       do field = 1, nflds
           do lon = 1, plon
               rstrt_buf(ip(lon,lat),lon,lat) = buf(field)
               ip(lon,lat) = ip(lon,lat) + 1
           enddo
       enddo
       return
       end

ccccccccccccccccccc

       subroutine pck_rstrt_init( numsect, collen )
       implicit none
#include "pmgrid.com"
#include "pagrid.com"
#include "pspect.com"
#include "comparll.com"
#include "restart_fil.com"
#include "restart_buf.com"
       integer numsect, collen
       integer lat, lon

       clen = collen		! length of column stored in common
       nsect = numsect		! number of sections allowed
       isect = 0		! reset counter of sections
       do lat = 1, plat
          do lon = 1, plon
	      ip(lon,lat) = 1
          enddo
       enddo
       return
       end

ccccccccccccccccccc

       subroutine pck_rstrt_sect(sn,nwords)
       implicit none
#include "pmgrid.com"
#include "pagrid.com"
#include "pspect.com"
#include "comparll.com"
#include "restart_fil.com"
#include "restart_buf.com"
       integer sn, nwords

       isect = isect+1
       if ( isect .gt. nsect ) then
#ifdef PP_ERROR
	   write(0,*)'pck_rstrt_sect',isect,nsect
#endif
	   call abortf('pck_rstrt_sect')
       endif

#ifdef PP_ERROR
       write(0,*)'pck_rstrt_sect: ',sn,nwords
#endif
       call pck_rstrt_sec1( sn, nwords, rstrt_buf, ip, clen, plon, plat,
     $                      m_pin_ln, m_pin_lt, p_glat, p_nproc_lt )
       return
 100   format( 'pck_rstrt_sect: too many sections ',I4,
     $  '.  Only ',I4,' sections initialized (see pck_rstrt_init()).')
       end

       subroutine pck_rstrt_sec1(sn,nwords,rstrt_buf,ip,clen,plon,plat,
     $                           m_pin_ln, m_pin_lt, p_glat, p_nproc_lt)
       implicit none
       integer sn, nwords, clen, plat, plon
       integer ip( plon, plat )
       real rstrt_buf( clen, plon, plat )
       integer lat, lon, glat, glon
       integer m_pin_ln, m_pin_lt, p_glat, p_nproc_lt

       do lat = 1, plat
          do lon = 1, plon
              if ( lat .le. plat/2 ) then
                  glat = lat + m_pin_lt * plat/2
              else
                  glat = p_glat/2 + lat +
     $                   (plat/2)*(p_nproc_lt-m_pin_lt-2)
              endif
	      glon = m_pin_ln * plon + lon
              rstrt_buf(ip(lon,lat)  ,lon,lat) = sn
              rstrt_buf(ip(lon,lat)+1,lon,lat) = nwords
              rstrt_buf(ip(lon,lat)+2,lon,lat) = glat
	      rstrt_buf(ip(lon,lat)+3,lon,lat) = glon
	      ip(lon,lat) = ip(lon,lat)+4
          enddo
       enddo
       return
       end

ccccccccccccccccccc
ccccccccccccccccccc
ccccccccccccccccccc
