C#######################################################################
C PSTSWM Version 4.0 (12/1/94)                                         #
C  A message-passing benchmark code and parallel algorithm testbed     #
C  that solves the nonlinear shallow water equations using the spectral#
C  transform method.                                                   #
C Written by:                                                          #
C  Patrick Worley of Oak Ridge National Laboratory                     #
C  Ian Foster of Argonne National Laboratory                           #
C Based on the sequential code STSWM 2.0 by James Hack and Ruediger    #
C  Jakob of the National Center for Atmospheric Research.              #
C Research and development funded by the Computer Hardware, Advanced   #
C  Mathematics, and Model Physics (CHAMMP) program of the U.S.         #
C  Department of Energy.                                               # 
C                                                                      #
C Questions and comments should be directed to worley@msr.epm.ornl.gov #
C Please notify and acknowledge the authors in any research or         #
C publications utilizing PSTSWM or any part of the code.               #
C                                                                      #
C NOTICE: Neither the institutions nor the authors make any            #
C representations about the suitability of this software for any       #
C purpose. This software is provided "as is", without express or       #
C implied warranty.                                                    #
C#######################################################################
C include precision declaration definitions                            #
#include "params.h"
C#######################################################################
      SUBROUTINE HALFIMAX(COMMOPT, BUFFERS, PROTOPT, MAPSIZE,
     &                   MAP, MYINDEX, BASE1, BASE2, EXCHSIZE, LTH, WS,
     &                   SUP) 
C
C HALFIMAX calls routines that calculate a vector maximum over a 
C specified subset of processors using a recursive halving algorithm, 
C generalized to handle nonpowers of two number of processors. As
C messages become small in the recursive halving process, the algorithm
C switches to an exchange algorithm (idea courtesty of R. van de Geign).
C The MAP array defines the subset and the processor ordering to use. 
C The results are duplicated across all processors in the subset.
C
C Communication options (COMMOPT) for HALFIMAX include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/max
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/max
C Communication protocol options (PROTOPT) for HALFIMAX include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5)     nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  nonblocking receive
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5) .AND. (BUFFERS .GT. 1)
C    recv-ahead
C  IF (PROTOPT .EQ. 4 .OR. 5)         forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C called by: FLTSUM
C calls: HIMAX1, HIMAX2
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C number of communication buffers (to use in recv-ahead algorithms)
      INTEGER BUFFERS
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in MAP array
      INTEGER MYINDEX
C message type offsets to use in interprocessor communication
      INTEGER BASE1, BASE2
C message size (in reals) below which an exchange max algorithm is
C to be used instead of a recursive halving max algorithm
      INTEGER EXCHSIZE
C length of vectors
      INTEGER LTH
C
C     Work Space
C
C message buffers
      INTEGER WS(LTH*BUFFERS)
C
C     Input/Output
C
C On entry, contains (local) data. On exit contains vector max.
      INTEGER SUP(LTH)
C
C---- Executable Statements --------------------------------------------
C
      IF (MAPSIZE .GT. 1) THEN
C       Calculate vector max.
C
        IF (BUFFERS .EQ. 1) THEN
C         single buffer algorithms
          CALL HIMAX1(COMMOPT, PROTOPT, MAPSIZE,
     &               MAP, MYINDEX, BASE1, BASE2, EXCHSIZE, LTH, WS, SUP) 
C
        ELSEIF (BUFFERS .GE. 2) THEN
C         multiple buffer algorithms
          CALL HIMAX2(COMMOPT, BUFFERS, PROTOPT, MAPSIZE,
     &               MAP, MYINDEX, BASE1, BASE2, EXCHSIZE, LTH, WS, SUP)
C
        ELSE
C         illegal number of buffers specified
          WRITE(0,100) MAP(MYINDEX), BUFFERS
  100     FORMAT (/,' PSTSWM: FATAL ERROR IN SUBROUTINE HALFIMAX ',/,
     &            ' ILLEGAL COMMUNICATION BUFFER SIZE SPECIFIED',/,
     &            ' PROCID = ',I4,' BUFFERS = ',I4)
          STOP
        ENDIF
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE HIMAX1(COMMOPT, PROTOPT, MAPSIZE,
     &                 MAP, MYINDEX, BASE1, BASE2, EXCHSIZE, LTH, WS,
     &                 SUP) 
C
C HIMAX1 calculates a vector max over a specified subset of processors
C when one communication buffer is provided, using a recursive halving
C algorithm generalized to handle nonpowers of two number of processors.
C As messages become small in the recursive halving process, the
C algorithm switches to an exchange algorithm (idea courtesty of R. van
C de Geign). The MAP array defines the subset and the processor ordering
C to use. The results are duplicated across all processors in the
C subset. 
C
C Communication options (COMMOPT) for HIMAX1 include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/max
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/max
C Communication protocol options (PROTOPT) for HIMAX1 include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5)     nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  nonblocking receive
C  IF (PROTOPT .EQ. 4 .OR. 5)         forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C called by: HALFIMAX
C calls: HALF_INIT, SWAP, SWAP_RECV, SWAP_SEND
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C include precision declaration definitions                            #
#     include "palgor.com"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in MAP array
      INTEGER MYINDEX
C message type offsets to use in interprocessor communication
      INTEGER BASE1, BASE2
C message size (in reals) below which an exchange max algorithm is
C to be used instead of a recursive halving max algorithm
      INTEGER EXCHSIZE
C length of vectors
      INTEGER LTH
C
C     Work Space
C
C message buffers
      INTEGER WS(LTH)
C
C     Input/Output
C
C On entry, SUP contains (local) data. On exit, it contains the vector 
C max.
      INTEGER SUP(LTH)
C
C---- Local Variables --------------------------------------------------
C
C true processor id for "me"
      INTEGER ME
C flag indicating whether "me" is in the logical (partial) upper
C half cube of a non-power of two hypercube
      LOGICAL UPPER
C If MAPSIZE is not a power of two, then TWIN is the (true) processor 
C id for the neighboring processor in the other half cube of 
C the partial logical hypercube.
      INTEGER TWIN
C number of swaps in the recursive halving max and number of swaps
C in both the recursive halving and exchange max portions of the
C algorithm. MAXSTEP-HALFSTEP represents the number of swaps in the
C recursive doubling broadcast.
      INTEGER HALFSTEP, MAXSTEP
C arrays indicating the destination of the message sent during a
C given swap, and whether this processor sends or receives first
      INTEGER SWAPNODE(LGPROCSX)
      INTEGER ORDER(LGPROCSX)
C arrays indicating the size of the message being sent/received 
C during a given swap, and the corresponding index offsets
      INTEGER SENDSIZE(LGPROCSX)
      INTEGER RECVSIZE(LGPROCSX)
      INTEGER SENDINDEX(LGPROCSX)
      INTEGER RECVINDEX(LGPROCSX)
C temporaries for current message sizes and offsets
      INTEGER SNDLTH, RCVLTH, SNDDEX, RCVDEX
C loop indices
      INTEGER I, STEP
C
C---- Executable Statements --------------------------------------------
C
C     Identify who i am.
      ME = MAP(MYINDEX)
C
C     Precalculate swap partners and other information needed by vector
C     max algorithm.
      CALL HALF_INIT(MAPSIZE, MAP, MYINDEX, LTH, LGPROCSX, ALIGN, 
     &               EXCHSIZE, UPPER, TWIN, MAXSTEP, HALFSTEP, ORDER, 
     &               SENDSIZE, RECVSIZE, SENDINDEX, RECVINDEX, SWAPNODE)
C
      IF (UPPER) THEN
C       In upper half cube, so send data to TWIN and wait for results.
C       - ordered send/recv, with upper half cube processor going 
C         first
C       - safe to use same buffer for send and recv because using
C         ordered swap and going first
C
        CALL SWAP(1, PROTOPT, 1, ME, BASE1, TWIN, IBYTES*LTH, 
     &            SUP, IBYTES*LTH, SUP)
C
      ELSE
C       In lower half cube, so use usual power of two algorithm.
C
        IF (TWIN .NE. -1) THEN
C
C         Get data from TWIN in upper half cube using recv part of 
C         ordered swap.
          CALL SWAP_RECV(1, PROTOPT, -1, ME, BASE1, TWIN, 
     &                   IBYTES*LTH, WS)
C
C         Combine vectors.
          DO I = 1,LTH
            SUP(I) = MAX(SUP(I),WS(I))
          ENDDO
C
        ENDIF
C 
C       Calculate vector max in lower half cube using recursive halving
C       and/or exchange algorithm (leaving result partially
C       distributed). 
        DO STEP=1,MAXSTEP
C
          SNDLTH = SENDSIZE(STEP)
          RCVLTH = RECVSIZE(STEP)
          SNDDEX = SENDINDEX(STEP)
          RCVDEX = RECVINDEX(STEP)
C
C         Swap vectors.
          CALL SWAP(COMMOPT, PROTOPT, ORDER(STEP),
     &              ME, BASE1, SWAPNODE(STEP), IBYTES*SNDLTH,
     &              SUP(SNDDEX), IBYTES*RCVLTH, WS(RCVDEX))
C
C         Combine vectors.
          DO I = RCVDEX, RCVDEX+(RCVLTH-1)
            SUP(I) = MAX(SUP(I),WS(I))
          ENDDO
C
        ENDDO
C
C       Broadcast distributed vector by reversing recursive halving part
C       of the algorithm (recursize doubling).
        DO STEP=HALFSTEP,1,-1
C
          RCVLTH = SENDSIZE(STEP)
          SNDLTH = RECVSIZE(STEP)
          RCVDEX = SENDINDEX(STEP)
          SNDDEX = RECVINDEX(STEP)
C
C         Swap vectors.
          CALL SWAP(COMMOPT, PROTOPT, ORDER(STEP),
     &              ME, BASE2, SWAPNODE(STEP), IBYTES*SNDLTH,
     &              SUP(SNDDEX), IBYTES*RCVLTH, SUP(RCVDEX))
C
        ENDDO
C
        IF (TWIN .NE. -1) THEN
C
C         Send results to TWIN in upper half cube using send part of 
C         ordered swap.
          CALL SWAP_SEND(1, PROTOPT, -1, ME, BASE1, TWIN,
     &                   IBYTES*LTH, SUP, IBYTES*LTH, SUP)
C
        ENDIF
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE HIMAX2(COMMOPT, BUFFERS, PROTOPT, MAPSIZE,
     &                 MAP, MYINDEX, BASE1, BASE2, EXCHSIZE, LTH, WS,
     &                 SUP) 
C
C HIMAX2 calculates a vector max over a specified subset of processors
C when more than one communication buffer is provided, using a recursive
C halving algorithm generalized to handle nonpowers of two number of
C processors. As messages become small in the recursive halving process,
C the algorithm switches to an exchange algorithm (idea courtesty of R.
C van de Geign). The MAP array defines the subset and the processor
C ordering to use. The results are duplicated across all processors in
C the subset. 
C
C Communication options (COMMOPT) for HIMAX2 include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/max
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/max
C Communication protocol options (PROTOPT) for HIMAX2 include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5) nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  
C    nonblocking receive and recv-ahead
C  IF (PROTOPT .EQ. 4 .OR. 5) forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C called by: HALFIMAX
C calls: HALF_INIT, SWAP, SWAP1, SWAP2, SWAP3, SWAP_RECV, SWAP_RECVBEGIN, 
C        SWAP_RECVEND, SWAP_SEND
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C include precision declaration definitions                            #
#     include "palgor.com"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C number of communication buffers (to use in recv-ahead algorithms)
      INTEGER BUFFERS
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in MAP array
      INTEGER MYINDEX
C message type offsets to use in interprocessor communication
      INTEGER BASE1, BASE2
C message size (in reals) below which an exchange max algorithm is
C to be used instead of a recursive halving max algorithm
      INTEGER EXCHSIZE
C length of vectors
      INTEGER LTH
C
C     Work Space
C
C message buffers
      INTEGER WS(LTH*BUFFERS)
C
C     Input/Output
C
C on entry, contains (local) data, on exit contains vector max
      INTEGER SUP(LTH)
C
C---- Local Variables --------------------------------------------------
C
C true processor id for "me"
      INTEGER ME
C flag indicating whether "me" is in the logical (partial) upper
C half cube of a non-power of two hypercube
      LOGICAL UPPER
C If MAPSIZE is not a power of two, then TWIN is the (true) processor 
C id for the neighboring processor in the other half cube of 
C the partial logical hypercube.
      INTEGER TWIN
C number of swaps in the recursive halving max and number of swaps
C in both the recursive halving and exchange max portions of the
C algorithm. MAXSTEP-HALFSTEP represents the number of swaps in the
C recursive doubling broadcast.
      INTEGER HALFSTEP, MAXSTEP
C arrays indicating the destination of the message sent during a
C given swap, and whether this processor sends or receives first
      INTEGER SWAPNODE(LGPROCSX)
      INTEGER ORDER(LGPROCSX)
C arrays indicating the size of the message being sent/received 
C during a given swap, and the corresponding index offsets
      INTEGER SENDSIZE(LGPROCSX)
      INTEGER RECVSIZE(LGPROCSX)
      INTEGER SENDINDEX(LGPROCSX)
      INTEGER RECVINDEX(LGPROCSX)
      INTEGER WSINDEX(LGPROCSX)
C temporaries for current message sizes and offsets
      INTEGER SNDLTH, RCVLTH, SNDDEX, RCVDEX, WSDEX
C loop indices
      INTEGER I, STEP
C
C---- Executable Statements --------------------------------------------
C
C     Identify who i am.
      ME = MAP(MYINDEX)
C
C     Precalculate swap partners and other information needed by vector
C     max algorithm.
      CALL HALF_INIT(MAPSIZE, MAP, MYINDEX, LTH, LGPROCSX, ALIGN, 
     &               EXCHSIZE, UPPER, TWIN, MAXSTEP, HALFSTEP, ORDER, 
     &               SENDSIZE, RECVSIZE, SENDINDEX, RECVINDEX, SWAPNODE)
C
      IF (UPPER) THEN
C       In upper half cube, so send data to TWIN and wait for results.
C       - ordered send/recv, with upper half cube processor going 
C         first
C       - safe to use same buffer for send and recv because using
C         ordered swap and going first
C
        CALL SWAP(1, PROTOPT, 1, ME, BASE1, TWIN, IBYTES*LTH,
     &            SUP, IBYTES*LTH, SUP)
C
      ELSE
C       In lower half cube, so use usual power of two algorithm.
C
        IF (TWIN .NE. -1) THEN
C
C         Get/prepare for data from TWIN.
          IF (BUFFERS .GT. 2) THEN
C
C           Space available for recv-ahead for both TWIN and lower 
C           subcube interprocessor communication, so use recvbegin part
C           of ordered swap.
            CALL SWAP_RECVBEGIN(1, PROTOPT, -1, ME, BASE1,
     &                          TWIN, IBYTES*LTH, WS(2*LTH+1)) 
C
          ELSE
C
C           Space not available for recv-ahead for both TWIN and lower 
C           subcube interprocessor communication, so use recv part of 
C           ordered swap.
            CALL SWAP_RECV(1, PROTOPT, -1, ME, BASE1, TWIN,
     &                     IBYTES*LTH, WS)
C
C           Combine vectors.
            DO I = 1,LTH
              SUP(I) = MAX(SUP(I),WS(I))
            ENDDO
C
          ENDIF
C
        ENDIF
C
C       Calculate indices into work space to be used for recv-ahead 
C       algorithm, and post receive requests for recursive halving
C       vector max algorithm.  
        WSINDEX(1) = 1
        DO STEP=2,HALFSTEP
          WSINDEX(STEP) = WSINDEX(STEP-1) + RECVSIZE(STEP-1)
        ENDDO
        DO STEP=1,HALFSTEP
          CALL SWAP1(COMMOPT, PROTOPT, .TRUE., ORDER(STEP),
     &               ME, BASE1, SWAPNODE(STEP), IBYTES*RECVSIZE(STEP),
     &               WS(WSINDEX(STEP))) 
        ENDDO
C
C       Post receive requests for recursive doubling broadcast algorithm
C       (since know that requests will be posted before messages sent,
C       can disable safe forcetype protocol)
        DO STEP=HALFSTEP,1,-1
          CALL SWAP1(COMMOPT, PROTOPT, .FALSE., ORDER(STEP),
     &               ME, BASE2, SWAPNODE(STEP), IBYTES*SENDSIZE(STEP),
     &               SUP(SENDINDEX(STEP))) 
        ENDDO
C
        IF ((TWIN .NE. -1) .AND. (BUFFERS .GT. 2)) THEN
C
C         Get data from TWIN in upper half cube using recvend part of 
C         ordered swap.
          CALL SWAP_RECVEND(1, PROTOPT, -1, ME, BASE1, TWIN,
     &                      IBYTES*LTH, WS(2*LTH+1))
C
C         Combine vectors.
          DO I = 1,LTH
            SUP(I) = MAX(SUP(I),WS(2*LTH+I))
          ENDDO
C
        ENDIF
C 
C       Calculate vector max in lower half cube using recursive halving
C       algorithm (leaving result distributed).
        DO STEP=1,HALFSTEP
C
          SNDLTH = SENDSIZE(STEP)
          RCVLTH = RECVSIZE(STEP)
          SNDDEX = SENDINDEX(STEP)
          RCVDEX = RECVINDEX(STEP)
          WSDEX = WSINDEX(STEP)
C
C         Swap vectors.
          CALL SWAP2(COMMOPT, PROTOPT, .TRUE., ORDER(STEP), 
     &               ME, BASE1, SWAPNODE(STEP), IBYTES*SNDLTH, 
     &               SUP(SNDDEX), IBYTES*RCVLTH, WS(WSDEX))
C
C         Combine vectors.
          DO I = 0, RCVLTH-1
            SUP(RCVDEX+I) = MAX(SUP(RCVDEX+I),WS(WSDEX+I))
          ENDDO
C
        ENDDO
C
C       When message sizes get small, finish calculating vector max
C       using exchange algorithm (using previous RCVLTH, etc.).
C       Not using recv-ahead logic for exchange, since recv-ahead
C       is unlikely to be efficient for small messages.
        DO STEP=HALFSTEP+1,MAXSTEP
C
          SNDLTH = SENDSIZE(STEP)
          RCVLTH = RECVSIZE(STEP)
          SNDDEX = SENDINDEX(STEP)
          RCVDEX = RECVINDEX(STEP)
C
C         Swap vectors.
          CALL SWAP(COMMOPT, PROTOPT, ORDER(STEP),
     &              ME, BASE1, SWAPNODE(STEP), IBYTES*SNDLTH,
     &              SUP(SNDDEX), IBYTES*RCVLTH, WS)
C
C         Combine vectors.
          DO I = 0, RCVLTH-1
            SUP(RCVDEX+I) = MAX(SUP(RCVDEX+I),WS(I+1))
          ENDDO
C
        ENDDO
C
C       Broadcast distributed vector by reversing recursive halving 
C       algorithm (recursize doubling).
        DO STEP=HALFSTEP,1,-1
C
          RCVLTH = SENDSIZE(STEP)
          SNDLTH = RECVSIZE(STEP)
          RCVDEX = SENDINDEX(STEP)
          SNDDEX = RECVINDEX(STEP)
C
C         Swap vectors (with safe forcetype protocol turned off, to 
C         match swapbegin).
          CALL SWAP2(COMMOPT, PROTOPT, .FALSE., ORDER(STEP),
     &               ME, BASE2, SWAPNODE(STEP), IBYTES*SNDLTH,
     &               SUP(SNDDEX), IBYTES*RCVLTH, SUP(RCVDEX)) 
C
        ENDDO
C
        IF (TWIN .NE. -1) THEN
C
C         Send results to TWIN in upper half cube using send part of 
C         ordered swap.
          CALL SWAP_SEND(1, PROTOPT, -1, ME, BASE1, TWIN,
     &                   IBYTES*LTH, SUP, IBYTES*LTH, SUP)
C
        ENDIF
C
C       Clean up outstanding sends. Note that send buffers are never in
C       danger of being overwritten. During the calculation of the 
C       maximum, send buffer space is never reused. During the broadcast
C       stage, the processor that might overwrite a send buffer from the 
C       calculation stage was the destination of the message sent, so 
C       know that send has been completed.
        DO I = 1,HALFSTEP
          CALL SWAP3(COMMOPT, PROTOPT, ME, BASE1, 
     &               SWAPNODE(I), 0, WS)
        ENDDO
        DO I = HALFSTEP,1,-1
          CALL SWAP3(COMMOPT, PROTOPT, ME, BASE2, 
     &               SWAPNODE(I), 0, WS)
        ENDDO
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
