#include "params.h"
      subroutine pflxoce(pmidm1  ,ubot    ,vbot    ,
     $                  tbot    ,qbot    ,thbot   ,zbot    ,
     $                  shf     ,lhf     ,taux    ,tauy    )
C
C Compute ocean to atmosphere surface fluxes of sensible, latent heat
C and stress components:
C
C Assume:
C   1) Neutral 10m drag coeff: 
C         cdn = .0027/U10N + .000142 + .0000764 U10N
C   2) Neutral 10m stanton number: 
C         ctn = .0327 sqrt(cdn), unstable
C         ctn = .0180 sqrt(cdn), stable
C   3) Neutral 10m dalton number:  
C         cen = .0346 sqrt(cdn)
C   4) The saturation humidity of air at T(K): 
C         qsat(T)  (kg/m^3)
C
C Note:
C   1) here, tstar = <WT>/U*, and qstar = <WQ>/U*.
C   2) wind speeds should all be above a minimum speed (umin)
C
C--------------------------Code History---------------------------------
C
C Original:      Bill Large/M.Vertenstein, Sep. 1995
C Standardized:  L. Buja,     Feb 1996
C Reviewed:      B. Briegleb, March 1996
C
C-----------------------------------------------------------------------
      implicit none
C------------------------------Parameters-------------------------------
#include "pmgrid.com"
#include "parpbl.com"
C------------------------------Commons----------------------------------
#include "pcoupled.com"
c  put in pcplrstr.foam1.5PCsrc.com for PC runs   PCrgg
c  but use the name below                         PCrgg
#include "pcplrstr.com"
#include "comcon.com"
C------------------------------Arguments--------------------------------
C
C Input arguments
C
C                         
      real pmidm1(p_glon,p_glat)  ! Bottom level pressure
      real ubot(p_glon,p_glat)    ! Bottom level u wind
      real vbot(p_glon,p_glat)    ! Bottom level v wind
      real tbot(p_glon,p_glat)    ! Bottom level temperature
      real qbot(p_glon,p_glat)    ! Bottom level specific humidity
      real thbot(p_glon,p_glat)   ! Bottom level potential temperature
      real zbot(p_glon,p_glat)    ! Bottom level height above surface
C                         
C Output arguments        
C                         
      real shf(uwlon,uwlat,maxoverlaps)     ! Initial sensible heat flux (W/m2)
      real lhf(uwlon,uwlat,maxoverlaps)     ! Initial latent heat flux (W/m2)
      real taux(uwlon,uwlat,maxoverlaps)    ! X surface stress (N/m2)
      real tauy(uwlon,uwlat,maxoverlaps)    ! Y surface stress (N/m2)
C
C---------------------------Local variables-----------------------------
C
      real ts(uwlon,uwlat)  ! surface temp (use sst)
      integer i,j,k,l,m,n     ! Longitude indices
      real ssq            ! Surface saturation specific humidity
      real ustar          ! ustar
      real tstar          ! tstar
      real qstar          ! qstar
      real u10n           ! neutral 10 m wind speed over ocean
      real vmag           ! Surface wind magnitude
      real thvbot         ! Bottom lev virtual potential temp
      real delt           ! potential T difference (K)
      real delq           ! specific humidity difference (kg/kg)
      real rdn            ! sqrt of neutral exchange coeff (momentum)
      real rhn            ! sqrt of neutral exchange coeff (heat)
      real ren            ! sqrt of neutral exchange coeff (tracers)          
      real rd             ! sqrt of exchange coeff (momentum)
      real rh             ! sqrt of exchange coeff (heat)
      real re             ! sqrt of exchange coeff (tracers)
      real hol            ! Ref hgt (10m) / monin-obukhov length
      real xsq            ! Temporary variable
      real xqq            ! Temporary variable
      real alz            ! ln(zbot/z10)
      real cp             ! Specific heat of moist air
      real tau            ! Reference height stress
      real psimh          ! Stability funct at ref lev (momentum)
      real psixh          ! Stability funct at ref lev (heat & tracers) 
      real stable         ! Stability factor
      real rbot(p_glon,p_glat)    ! Density at bottom model level
      integer indx(uwlon)       ! Longitude index array (land)
      integer npts              ! Number of land points
      integer ii              ! Longitude indices 

C
C--------------------------Statement functions--------------------------
C
      real psimhu         ! Unstable part of psimh
      real psixhu         ! Unstable part of psixh
      real qsat           ! Saturation specific humidty of air
      real cdn            ! Neutral drag coeff at bottom model level
      real xd             ! Dummy argument
      real Tk             ! Temperature (K)
      real Umps           ! Wind velocity (m/sec)
C
      qsat(Tk)   = 640380. / exp(5107.4/Tk)
      cdn(Umps)  = 0.0027 / Umps + .000142 + .0000764 * Umps
      psimhu(xd) = log((1.+xd*(2.+xd))*(1.+xd*xd)/8.)
     $              - 2.*atan(xd) + 1.571
      psixhu(xd) = 2. * log((1. + xd*xd)/2.)
C
C-----------------------------------------------------------------------
C
      npts = 0
      do n=1,2
      do j=plats(n),plate(n)
       do i=plons,plone
         if(uwmask(i,j) .eq. 0) then
          npts =npts +1
         endif
       enddo
      enddo
      enddo
      if(npts .eq. 0) return

      do n=1,2
      do j=plats(n),plate(n)

       npts = 0
       do i=plons,plone
         if(uwmask(i,j) .eq. 0) then
          npts =npts +1
          indx(npts)=i
         endif
       enddo

      do ii=1,npts
          i=indx(ii)
	  ts(i,j)=sst(i,j)

      do k=1,overlaps(j,i)
	 l = ctoulon(j,i,k)
	 m = ctoulat(j,i,k)

C
C---------------------------------------------------------------
C Set up necessary variables
C---------------------------------------------------------------
C
         rbot(l,m)= pmidm1(l,m) / (rair*tbot(l,m))
         vmag   = max(umin, sqrt(ubot(l,m)**2+vbot(l,m)**2))
         thvbot = thbot(l,m) * (1.0 + zvir*qbot(l,m))
         ssq    = 0.98 * qsat(ts(i,j)) / rbot(l,m)
         delt   = thbot(l,m) - ts(i,j)
         delq   = qbot(l,m) - ssq  
         alz    = log(zbot(l,m)/zref) 
         cp     = cpair*(1. + cpvir*ssq) 
C
C---------------------------------------------------------------
C First iteration to converge on Z/L and hence the fluxes
C---------------------------------------------------------------
C
C Initial guess for roots of neutral exchange coefficients, 
C assume z/L=0. and u10n is approximated by vmag.
C Stable if (thbot > ts ).
C
         stable = 0.5 + sign(0.5 , delt)
         rdn  = sqrt(cdn(vmag))
         rhn  = (1.-stable) * 0.0327 + stable * 0.018 
         ren  = 0.0346 
C
C Initial guess of ustar,tstar and qstar
C
         ustar = rdn*vmag
         tstar = rhn*delt
         qstar = ren*delq
C
C Compute stability and evaluate all stability functions
C Stable if (thbot > ts or hol > 0 )
C
         hol = xkar *gravit *zbot(l,m)
     $         * (tstar/thvbot + qstar/(1./zvir+qbot(l,m)))
     $         / ustar**2
         hol = sign( min(abs(hol),10.), hol )
         stable = 0.5 + sign(0.5 , hol)
         xsq   = max(sqrt(abs(1. - 16.*hol)) , 1.)
         xqq   = sqrt(xsq)
         psimh = -5. * hol * stable + (1.-stable)*psimhu(xqq)
         psixh = -5. * hol * stable + (1.-stable)*psixhu(xqq)
C
C Shift 10m neutral wind speed using old rdn coefficient
C
         rd   = rdn / (1.+rdn/xkar*(alz-psimh))
         u10n = vmag * rd / rdn
C
C Update the neutral transfer coefficients at 10m and neutral stability
C
         rdn = sqrt(cdn(u10n))
         ren = 0.0346
         rhn = (1.-stable) * 0.0327 + stable * 0.018 
C
C Shift all coeffs to measurement height and stability
C
         rd = rdn / (1.+rdn/xkar*(alz-psimh)) 
         rh = rhn / (1.+rhn/xkar*(alz-psixh)) 
         re = ren / (1.+ren/xkar*(alz-psixh))
C
C Update ustar, tstar, qstar using updated, shifted coeffs 
C
         ustar = rd * vmag 
         tstar = rh * delt 
         qstar = re * delq 
C
C---------------------------------------------------------------
C Second iteration to converge on Z/L and hence the fluxes
C---------------------------------------------------------------
C
C Recompute stability & evaluate all stability functions  
C Stable if (thbot > ts or hol > 0 )
C 
         hol = xkar * gravit * zbot(l,m)
     $        * (tstar/thvbot + qstar/(1./zvir+qbot(l,m)))
     $        / ustar**2
         hol = sign( min(abs(hol),10.), hol )
         stable = 0.5 + sign(0.5 , hol)
         xsq   = max(sqrt(abs(1. - 16.*hol)) , 1.)
         xqq   = sqrt(xsq)
         psimh = -5. * hol * stable + (1.-stable)*psimhu(xqq)
         psixh = -5. * hol * stable + (1.-stable)*psixhu(xqq)
C
C Shift 10m neutral wind speed using old rdn coefficient
C
         rd   = rdn / (1.+rdn/xkar*(alz-psimh))
         u10n = vmag * rd / rdn
C
C Update the neutral transfer coefficients at 10m and neutral stability
C
        rdn = sqrt(cdn(u10n))
        ren = 0.0346
        rhn = (1.-stable) * 0.0327 + stable * 0.018 
C
C Shift all coeffs to measurement height and stability

        rd = rdn / (1.+rdn/xkar*(alz-psimh)) 
        rh = rhn / (1.+rhn/xkar*(alz-psixh)) 
        re = ren / (1.+ren/xkar*(alz-psixh))
C
C Update ustar, tstar, qstar using updated, shifted coeffs 
C
        ustar = rd * vmag 
        tstar = rh * delt 
        qstar = re * delq 
C
C---------------------------------------------------------------
C Done, compute the fluxes
C---------------------------------------------------------------
C
C Compute surface stress components
C
        tau     =  rbot(l,m) * ustar * ustar 
        taux(i,j,k) = -tau * ubot(l,m) / vmag 
        tauy(i,j,k) = -tau * vbot(l,m) / vmag 
C
C Compute heat flux components at current surface temperature
C (Define positive latent and sensible heat as upwards into the atm)
C
        shf(i,j,k) = -cp * tau * tstar / ustar 
        lhf(i,j,k) = -latvap * tau * qstar / ustar

      enddo
      enddo
      enddo
      enddo
C
      return
      end

