C#######################################################################
C PSTSWM Version 4.0 (12/1/94)                                         #
C  A message-passing benchmark code and parallel algorithm testbed     #
C  that solves the nonlinear shallow water equations using the spectral#
C  transform method.                                                   #
C Written by:                                                          #
C  Patrick Worley of Oak Ridge National Laboratory                     #
C  Ian Foster of Argonne National Laboratory                           #
C Based on the sequential code STSWM 2.0 by James Hack and Ruediger    #
C  Jakob of the National Center for Atmospheric Research.              #
C Research and development funded by the Computer Hardware, Advanced   #
C  Mathematics, and Model Physics (CHAMMP) program of the U.S.         #
C  Department of Energy.                                               # 
C                                                                      #
C Questions and comments should be directed to worley@msr.epm.ornl.gov #
C Please notify and acknowledge the authors in any research or         #
C publications utilizing PSTSWM or any part of the code.               #
C                                                                      #
C NOTICE: Neither the institutions nor the authors make any            #
C representations about the suitability of this software for any       #
C purpose. This software is provided "as is", without express or       #
C implied warranty.                                                    #
C#######################################################################
C include precision declaration definitions                            #
#include "params.h"
C#######################################################################
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C seqfor.F                                                             C
C                                                                      C
C The following sequential routine calculates the forward block        C
C complex Fourier transform.                                           C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE CFFTF(N, CJUMP, MVECS, TRIG, Y)
C
C CFFTF calculates in-place (unnormalized) forward Fourier transforms
C of an array of complex vectors. Each vector is of length N, where N
C is a power of twos, threes, and fives. 
C
C called by: CUBCTF
C calls: FAX235
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C length of complex vector
      INTEGER N
C length of first dimension of data/results array
      INTEGER CJUMP
C number of vectors to be transformed.
      INTEGER MVECS
C trigonometric function values used in the transform
      COMPLEX TRIG(N)
C
C     Input/Output
C
C On input, Y contains the data to be transformed. On output, Y
C contains the results of the transforms.
      COMPLEX Y(CJUMP, MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop limits for each stage of the forward fast Fourier transform
      INTEGER NSTEPS, NBLOCK, INCREM
C factor of N currently being used in FFT
      INTEGER FACTOR
C index offset used to move through the trigonometric function values
C in the correct order
      INTEGER IC
C index offsets used to access the twiddle factors needed in a given 
C butterfly update
      INTEGER TOFF1, TOFF2, TOFF3, TOFF4, TOFF5
C index offsets used to access the data needed in a given butterfly
C update
      INTEGER YOFF1, YOFF2, YOFF3, YOFF4, YOFF5
C loop indices
      INTEGER I, J, K, L
C temporaries used in computing the forward butterfly updates
      COMPLEX C1, C2, C3, C4, C5
C previous vector length
      INTEGER NSAVE
      SAVE NSAVE
C array containing number of factors and factorization of N
      INTEGER NFAX(0:12), FOURFAX, TWOFAX, THREEFAX, FIVEFAX
      SAVE FOURFAX, TWOFAX, THREEFAX, FIVEFAX
C "local" twiddle factors, used in power-of-three and power-of-five
C butterfly updates
      REAL PI
      COMPLEX A1, A2, B1, B2, B3, B4
      SAVE A1, A2, B1, B2, B3, B4
C remember whether this is the first time routine called. If yes,
C then compute local values.
      LOGICAL FIRST
      SAVE FIRST
C
C---- Data Statements --------------------------------------------------
C
      DATA NSAVE /0/
      DATA FIRST /.TRUE./
C                                                                             
C---- Executable Statements --------------------------------------------
C
      IF (FIRST) THEN
        PI = 4.0D0*ATAN(1.0D0)
        A1 = EXP(CMPLX(0.0D0,-PI/3.0D0))
        A2 = EXP(CMPLX(0.0D0,-2.0D0*(PI/3.0D0)))
        B1 = EXP(CMPLX(0.0D0,-PI/5.0D0))
        B2 = EXP(CMPLX(0.0D0,-2.0D0*(PI/5.0D0)))
        B3 = EXP(CMPLX(0.0D0,-3.0D0*(PI/5.0D0)))
        B4 = EXP(CMPLX(0.0D0,-4.0D0*(PI/5.0D0)))
        FIRST = .FALSE.
      ENDIF
C
      IF (NSAVE .NE. N) THEN
C       Following logic assumes that all factors are grouped, with
C       4s first, followed by 2s, 3s, and 5s, respectively.
C
        CALL FAX235(N, NFAX)
C
        FOURFAX=0
        TWOFAX=0
        THREEFAX=0
        FIVEFAX=0
        DO K=1,NFAX(0)
          IF (NFAX(K) .EQ. 4) THEN
            FOURFAX = FOURFAX + 1
          ELSEIF (NFAX(K) .EQ. 2) THEN
            TWOFAX = TWOFAX + 1
          ELSEIF (NFAX(K) .EQ. 3) THEN
            THREEFAX = THREEFAX + 1
          ELSEIF (NFAX(K) .EQ. 5) THEN
            FIVEFAX = FIVEFAX + 1
          ELSE
            WRITE(6,*) 'CFFTF: N contains factors other than 2,3,4,5.'
            STOP
          ENDIF
        ENDDO
C
        NSAVE = N
      ENDIF
C
      DO L=1,MVECS
C
        INCREM = N
        NBLOCK = 1
        IC = 0
C
        DO K=1,FOURFAX
C
          INCREM = INCREM/4
C
          TOFF2 = IC
          TOFF3 = TOFF2 + INCREM
          TOFF4 = TOFF3 + INCREM
          DO J=1,NBLOCK
C
            YOFF1 = 4*(J-1)*INCREM
            YOFF2 = YOFF1 + INCREM
            YOFF3 = YOFF2 + INCREM
            YOFF4 = YOFF3 + INCREM
            DO I=1,INCREM
C
C             Calculate power-of-four butterfly update.
              C1 = Y(YOFF1+I,L) + Y(YOFF3+I,L)
              C2 = Y(YOFF1+I,L) - Y(YOFF3+I,L)
              C3 = Y(YOFF2+I,L) + Y(YOFF4+I,L)
C C4 = CMPLX(0.0, 1.0)*(Y(YOFF2+I,L) - Y(YOFF4+I,L))
              C4 = CMPLX(AIMAG(Y(YOFF4+I,L))-AIMAG(Y(YOFF2+I,L)),
     &                   AREAL(Y(YOFF2+I,L))-AREAL(Y(YOFF4+I,L)))
C
              Y(YOFF1+I,L) = (C1 + C3)
              Y(YOFF2+I,L) = (C2 - C4)*TRIG(TOFF2+I)
              Y(YOFF3+I,L) = (C1 - C3)*TRIG(TOFF3+I)
              Y(YOFF4+I,L) = (C2 + C4)*TRIG(TOFF4+I)
C
            ENDDO
C
          ENDDO
C
          IC = IC + 3*INCREM
	  NBLOCK = NBLOCK*4
C
        ENDDO
C
        DO K=1,TWOFAX
C
          INCREM = INCREM/2
C
          TOFF2 = IC
          DO J=1,NBLOCK
C
            YOFF1 = 2*(J-1)*INCREM
            YOFF2 = YOFF1 + INCREM
            DO I=1,INCREM
C
C             Calculate power-of-two butterfly update.
              C2 = Y(YOFF1+I,L) - Y(YOFF2+I,L)
              Y(YOFF1+I,L) = (Y(YOFF1+I,L) + Y(YOFF2+I,L))
              Y(YOFF2+I,L) = C2*TRIG(TOFF2+I)
C
            ENDDO
C
          ENDDO
C
          IC = IC + INCREM
	  NBLOCK = NBLOCK*2
C
        ENDDO
C
        DO K=1,THREEFAX
C
          INCREM = INCREM/3
C
          TOFF2 = IC
          TOFF3 = TOFF2 + INCREM
          DO J=1,NBLOCK
C
            YOFF1 = 3*(J-1)*INCREM
            YOFF2 = YOFF1 + INCREM
            YOFF3 = YOFF2 + INCREM
            DO I=1,INCREM
C
C             Calculate power-of-three butterfly update.
              C2 = Y(YOFF1+I,L) + A2*Y(YOFF2+I,L) - A1*Y(YOFF3+I,L)
              C3 = Y(YOFF1+I,L) - A1*Y(YOFF2+I,L) + A2*Y(YOFF3+I,L)
              Y(YOFF1+I,L) = Y(YOFF1+I,L) + Y(YOFF2+I,L)
     &                     + Y(YOFF3+I,L)
              Y(YOFF2+I,L) = C2*TRIG(TOFF2+I)
              Y(YOFF3+I,L) = C3*TRIG(TOFF3+I)
C
            ENDDO
C
          ENDDO
C
          IC = IC + 2*INCREM
	  NBLOCK = NBLOCK*3
C
        ENDDO
C
        DO K=1,FIVEFAX
C
          INCREM = INCREM/5
C
          TOFF2 = IC
          TOFF3 = TOFF2 + INCREM
          TOFF4 = TOFF3 + INCREM
          TOFF5 = TOFF4 + INCREM
          DO J=1,NBLOCK
C
            YOFF1 = 5*(J-1)*INCREM
            YOFF2 = YOFF1 + INCREM
            YOFF3 = YOFF2 + INCREM
            YOFF4 = YOFF3 + INCREM
            YOFF5 = YOFF4 + INCREM
            DO I=1,INCREM
C
C             Calculate power-of-five butterfly update.
              C2 = Y(YOFF1+I,L) + B2*Y(YOFF2+I,L) + B4*Y(YOFF3+I,L)
     &                          - B1*Y(YOFF4+I,L) - B3*Y(YOFF5+I,L)
              C3 = Y(YOFF1+I,L) + B4*Y(YOFF2+I,L) - B3*Y(YOFF3+I,L)
     &                          + B2*Y(YOFF4+I,L) - B1*Y(YOFF5+I,L)
              C4 = Y(YOFF1+I,L) - B1*Y(YOFF2+I,L) + B2*Y(YOFF3+I,L)
     &                          - B3*Y(YOFF4+I,L) + B4*Y(YOFF5+I,L)
              C5 = Y(YOFF1+I,L) - B3*Y(YOFF2+I,L) - B1*Y(YOFF3+I,L)
     &                          + B4*Y(YOFF4+I,L) + B2*Y(YOFF5+I,L)
              Y(YOFF1+I,L) = Y(YOFF1+I,L) + Y(YOFF2+I,L)
     &                     + Y(YOFF3+I,L) + Y(YOFF4+I,L)
     &                     + Y(YOFF5+I,L)
              Y(YOFF2+I,L) = C2*TRIG(TOFF2+I)
              Y(YOFF3+I,L) = C3*TRIG(TOFF3+I)
              Y(YOFF4+I,L) = C4*TRIG(TOFF4+I)
              Y(YOFF5+I,L) = C5*TRIG(TOFF5+I)
C
            ENDDO
C 
          ENDDO
C
          IC = IC + 4*INCREM
	  NBLOCK = NBLOCK*5
C
        ENDDO
C
      ENDDO
C
      RETURN
      END
