C#######################################################################
C PSTSWM Version 6.2 (12/1/97)                                         #
C  A message-passing benchmark code and parallel algorithm testbed     #
C  that solves the nonlinear shallow water equations using the spectral#
C  transform method.                                                   #
C Written by:                                                          #
C  Patrick Worley of Oak Ridge National Laboratory                     #
C  Ian Foster of Argonne National Laboratory                           #
C Based on the sequential code STSWM 2.0 by James Hack and Ruediger    #
C  Jakob of the National Center for Atmospheric Research.              #
C Research and development funded by the Computer Hardware, Advanced   #
C  Mathematics, and Model Physics (CHAMMP) program of the U.S.         #
C  Department of Energy.                                               # 
C                                                                      #
C Questions and comments should be directed to worley@msr.epm.ornl.gov #
C Please notify and acknowledge the authors in any research or         #
C publications utilizing PSTSWM or any part of the code.               #
C                                                                      #
C NOTICE: Neither the institutions nor the authors make any            #
C representations about the suitability of this software for any       #
C purpose. This software is provided "as is", without express or       #
C implied warranty.                                                    #
C#######################################################################
C include precision declaration definitions                            #
#include "params.h"
C#######################################################################
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C mpi.F                                                                C
C                                                                      C
C The following routines support the use of MPI nonblocking and        C
C buffered communication routines in swap.F and sendrecv.F. Where      C
C possible, these should be in-lined for efficiency.                   C
C                                                                      C
C COMMINIT - used to initialize any data structures used by native     C
C            implementations of sendrecv and swap.                     C
C COMMEXIT - used to clean up after use of native commands in PSTSWM   C
C ADDSEND  - used to add a nonblocking send request id to the list of  C
C            outstanding requests.                                     C
C RMVSEND  - used to remove a nonblocking send request id from the listC
C            of outstanding requests.                                  C
C ADDRECV  - used to add a nonblocking receive request id to the list  C
C            of outstanding requests.                                  C
C RMVRECV  - used to remove a nonblocking receive request id from the  C
C            list of outstanding requests.                             C
C GETDATAF - query datatype of message being communicated              C
C SETDATAF - set datatype of message being communicated                C
C SYSTEM_MESH_MAP - used to number processes                           C
C NATSUM   - native distributed vector sum routine                     C
C NATTRS   - native distributed matrix transpose routine               C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
#ifdef COUPLED
      SUBROUTINE COMMINIT(SYSMSGS, SYSVOL, COMBUF, BUFSIZ, NEW_COMM)
#else
      SUBROUTINE COMMINIT(SYSMSGS, SYSVOL, COMBUF, BUFSIZ)
#endif
C
C This subroutine initializes the data structures used to support the
C nonblocking communication options in swap.F and sendrecv.F in the
C native MPI implemenations. 
C
C called by: PSTSWM
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C mpi implementation-specific parameters
#     include "mpif.h"
C pstswm-specific parameters
#     include "params.i"
C
C---- Common Blocks ----------------------------------------------------
C
C definition for communicator and datatype for native MPI implementation
#     include "picl.i"
C data and data structures for native MPI implementation
#     include "mpi.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input 
C
C number of blocking asynchronous send requests that the system needs 
C to be able to handle 
      INTEGER SYSMSGS
C size (in bytes) of blocking asynchronous send requests that the system 
C needs to be able to handle 
      INTEGER SYSVOL
C size of user buffer space
      INTEGER BUFSIZ
C User space to use for communication buffering 
      REAL COMBUF(BUFSIZ)
#ifdef COUPLED
      INTEGER NEW_COMM
#endif
C
C---- Local Variables --------------------------------------------------
C
C PROCESSOR ID
      INTEGER ME
C index variable
      INTEGER I
C error return variable
      INTEGER IERR
C
C---- Executable Statements --------------------------------------------
C
C     Generate new communicator to use in swap and sendrecv, disabling
C     MPI_ABORT for errors.
#ifdef COUPLED
      CALL MPI_COMM_DUP(NEW_COMM, COMM, IERR)
      CALL MPI_COMM_DUP(NEW_COMM, PSTSWM_COMM, IERR)
#else
      CALL MPI_COMM_DUP(MPI_COMM_WORLD, COMM, IERR)
      CALL MPI_COMM_DUP(MPI_COMM_WORLD, PSTSWM_COMM, IERR)
#endif
      CALL MPI_ERRHANDLER_SET(COMM, MPI_ERRORS_RETURN, IERR)
      CALL MPI_ERRHANDLER_SET(PSTSWM_COMM, MPI_ERRORS_RETURN, IERR)
C
C     Define MPI collective communicators to be NULL
      SUM_COMM = MPI_COMM_NULL
      DO I=1,6
        TRS_COMM(I) = MPI_COMM_NULL
      ENDDO
C
C     Provide user buffer for buffered communication
      IF (BUFSIZ*RBYTES .LT. SYSVOL) THEN
        CALL MPI_COMM_RANK(PSTSWM_COMM, ME, IERR)
c       IF (ME .EQ. 0) THEN
          WRITE(6,1) BUFSIZ*RBYTES, SYSVOL
 1        FORMAT(/,' -- warning warning warning warning -- ',/,
     & ' USER BUFFER SIZE INSUFFICIENT FOR SYSTEM BUFFER ',/,
     & ' REQUIREMENTS. RUN MAY DEADLOCK.',/,
     & ' USER SPACE: ',I10,' BUFFER REQUIREMENT: ',I10,/,
     &     ' -- warning warning warning warning -- ',/)
c       ENDIF
      ENDIF
      CALL MPI_BUFFER_ATTACH(COMBUF, BUFSIZ*RBYTES, IERR)
C
C     Initialize data structures used to hold nonblocking communication
C     request IDs.
      FIRSTRCV = 0
      NXTRCV   = 0
      FIRSTSND = 0
      NXTSND   = 0
      DO I = 0,NPROCSX-1
       RCVIDS(I) = -1
       RCVTAGS(I) = -2
       RCVSRCS(I) = -2
       SNDIDS(I) = -1
       SNDTAGS(I) = -2
       SNDDSTS(I) = -2
      ENDDO
C
      RETURN
C
      END
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
      SUBROUTINE COMMEXIT
C
C This subroutine does any clean up needed after using native commands
C commands in PSTSWM. 
C
C called by: PSTSWM
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C mpi implementation-specific parameters
#     include "mpif.h"
C
C---- Common Blocks ----------------------------------------------------
C
C definition for communicator and datatype for native MPI implementation
#     include "picl.i"
C
C---- Local Variables --------------------------------------------------
C
C Pointer to user space used for communication buffering 
      INTEGER DUMMY
C size of returned user buffer space
      INTEGER BUFSIZ
C index variable
      INTEGER I
C error return variable
      INTEGER IERR
C
C---- Executable Statements --------------------------------------------
C
C     Clean-up and "return" user buffer used for buffered communication
      CALL MPI_BUFFER_DETACH(DUMMY, BUFSIZ, IERR)
C
C     Free SWAP/SENDRECV communicator
      CALL MPI_COMM_FREE(COMM, IERR)
C
C     Free topology communicator
      IF (CARTFLAG) THEN
        CALL MPI_COMM_FREE(MESHCOMM, IERR)
        CARTFLAG = .FALSE.
      ENDIF
C
C     Free collective communicators
      IF (SUM_COMM .NE. MPI_COMM_NULL) THEN
        CALL MPI_COMM_FREE(SUM_COMM, IERR)
      ENDIF
      DO I=1,6
        IF (TRS_COMM(I) .NE. MPI_COMM_NULL) THEN
          CALL MPI_COMM_FREE(TRS_COMM(I), IERR)
        ENDIF
      ENDDO
C
      RETURN
C
      END
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE ADDSEND(SNDID, SNDTAG, SNDDST)
C
C This subroutine adds a nonblocking send request id and the associated
C message tag to the list of outstanding requests.
C
C called by: SRBEGIN, SR2
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Common Blocks ----------------------------------------------------
C
C data and data structures for native MPI implementation
#     include "mpi.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input 
C
C nonblocking send request id
      INTEGER SNDID
C associated message tag value
      INTEGER SNDTAG
C associated message destination
      INTEGER SNDDST
C
C---- Executable Statements --------------------------------------------
C
C     If there is space, add to the list
      IF (SNDIDS(NXTSND) .EQ. -1) THEN
        SNDIDS(NXTSND) = SNDID
        SNDTAGS(NXTSND) = SNDTAG
        SNDDSTS(NXTSND) = SNDDST
        NXTSND = MOD(NXTSND+1,NPROCSX)
      ELSE
        WRITE (0,101)
  101   FORMAT(/,' PSTSWM: FATAL ERROR IN SUBROUTINE ADDSEND:',
     &         /,' NOT ENOUGH SPACE FOR NONBLOCKING SEND REQUEST IDS')
        STOP                                                   
      ENDIF
C
      RETURN
      END
C                                                                      
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION RMVSEND(SNDTAG, SNDDST)
C
C This subroutine finds and returns a nonblocking send request id that
C matches the indicated message tag, removing it from the list of 
C outstanding requests in the process.
C
C called by: SREND, SR3
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Common Blocks ----------------------------------------------------
C
C data and data structures for native MPI implementation
#     include "mpi.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input 
C
C message tag value
      INTEGER SNDTAG
C message destination
      INTEGER SNDDST
C
C---- Local Variables --------------------------------------------------
C
C index variable
      INTEGER I
C
C---- Executable Statements --------------------------------------------
C
C     Find first id associated with indicated message tag.
      I = FIRSTSND
      DO WHILE ((I .NE. NXTSND) .AND. 
     &          ((SNDTAGS(I) .NE. SNDTAG) .OR. (SNDDSTS(I) .NE. SNDDST)
     &                                    .OR. (SNDIDS(I) .EQ. -1)))
        I = MOD(I+1,NPROCSX)
      ENDDO
C     If found anything, return it.
      IF (I .NE. NXTSND) THEN
        RMVSEND = SNDIDS(I)
        SNDIDS(I) = -1
C       Update "first" (oldest id) pointer.
        DO WHILE ((FIRSTSND .NE. NXTSND) .AND. 
     &            (SNDIDS(FIRSTSND) .EQ. -1))
          FIRSTSND = MOD(FIRSTSND+1,NPROCSX)
        ENDDO
      ELSE
        WRITE (0,101) SNDTAG, FIRSTSND, NXTSND
  101   FORMAT(/,' PSTSWM: FATAL ERROR IN SUBROUTINE RMVSEND:',
     &         /,' COULD NOT FIND OUTSTANDING SEND REQUEST WITH',
     &         /,' INDICATED MESSAGE TAG',
     &         /,' SNDTAG = ',I10,' FIRSTSND = ',I10,' NXTSND = ',I10)
        STOP                                                   
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE ADDRECV(RCVID, RCVTAG, RCVSRC)
C
C This subroutine adds a nonblocking send request id and the associated
C message tag to the list of outstanding requests.
C
C called by: SRBEGIN, SR1
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Common Blocks ----------------------------------------------------
C
C data and data structures for native MPI implementation
#     include "mpi.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input 
C
C nonblocking send request id
      INTEGER RCVID
C associated message tag value
      INTEGER RCVTAG
C associated message source
      INTEGER RCVSRC
C
C---- Executable Statements --------------------------------------------
C
C     If there is space, add to the list
      IF (RCVIDS(NXTRCV) .EQ. -1) THEN
        RCVIDS(NXTRCV) = RCVID
        RCVTAGS(NXTRCV) = RCVTAG
        RCVSRCS(NXTRCV) = RCVSRC
        NXTRCV = MOD(NXTRCV+1,NPROCSX)
      ELSE
        WRITE (0,101)
  101   FORMAT(/,' PSTSWM: FATAL ERROR IN SUBROUTINE ADDRECV:',
     &         /,' NOT ENOUGH SPACE FOR NONBLOCKING RECEIVE REQUEST',
     &         /,' IDS')
        STOP                                                   
      ENDIF
C
      RETURN
      END
C                                                                      
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION RMVRECV(RCVTAG, RCVSRC)
C
C This subroutine finds and returns a nonblocking receive request id 
C that matches the indicated message tag, removing it from the list of 
C outstanding requests in the process.
C
C called by: SREND, SR2, SR3
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Common Blocks ----------------------------------------------------
C
C data and data structures for native MPI implementation
#     include "mpi.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input 
C
C message tag value
      INTEGER RCVTAG
C message source
      INTEGER RCVSRC
C
C---- Local Variables --------------------------------------------------
C
C index variable
      INTEGER I
C
C---- Executable Statements --------------------------------------------
C
C     Find first id associated with indicated message tag.
      I = FIRSTRCV
      DO WHILE ((I .NE. NXTRCV) .AND. 
     &          ((RCVTAGS(I) .NE. RCVTAG) .OR. (RCVSRCS(I) .NE. RCVSRC)
     &                                    .OR. (RCVIDS(I) .EQ. -1)))
        I = MOD(I+1,NPROCSX)
      ENDDO
C     If found anything, return it.
      IF (I .NE. NXTRCV) THEN
        RMVRECV = RCVIDS(I)
        RCVIDS(I) = -1
C       Update "first" (oldest id) pointer.
        DO WHILE ((FIRSTRCV .NE. NXTRCV) .AND. 
     &            (RCVIDS(FIRSTRCV) .EQ. -1))
          FIRSTRCV = MOD(FIRSTRCV+1,NPROCSX)
        ENDDO
      ELSE
        WRITE (0,101) RCVTAG, FIRSTRCV, NXTRCV
  101   FORMAT(/,' PSTSWM: FATAL ERROR IN SUBROUTINE RMVRECV:',
     &         /,' COULD NOT FIND OUTSTANDING RECEIVE REQUEST WITH',
     &         /,' INDICATED MESSAGE TAG',
     &         /,' RCVTAG = ',I10,' FIRSTRCV = ',I10,' NXTRCV = ',I10)
        STOP                                                   
      ENDIF
C
      RETURN
      END
C                                                                      
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE GETDATAF(SAVETYPE)
C
C This function returns the data type specified by the last call to
C SETDATAF. 
C
C called by: MESH_MAP_INIT, RING_MAP_INIT
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Output
C 
C Current data type
      CHARACTER*1 SAVETYPE(1)
C
C----- Executable Statements -------------------------------------------
C
      CALL GETDATA0F(SAVETYPE)
C
      RETURN
C
      END
C                                                                      
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE SETDATAF(NEWTYPE)
C
C This function remembers the data type for subsequent interprocess 
C communication calls. 
C
C called by: ALGINP, FLTSUM, ILTCAST, MESH_MAP_INIT, PRBINP,
C            PRFINP, PSTSWM, RFTLON, RING_MAP_INIT, TIMING
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C mpi implementation-specific parameters
#     include "mpif.h"
C pstswm-specific parameters
#     include "params.i"
C
C---- Common Blocks ----------------------------------------------------
C
C definition for communicator and datatype for native MPI implementation
#     include "picl.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C 
C Current data type
      CHARACTER*1 NEWTYPE(1)
C
C---- Executable Statements --------------------------------------------
C
      CALL SETDATA0F(NEWTYPE)
      IF ((NEWTYPE(1) .EQ. 'I') .OR. (NEWTYPE(1) .EQ. 'i')) THEN
        DATATYPE = MPI_INTEGER
        DATALTH  = IBYTES
      ELSEIF ((NEWTYPE(1) .EQ. 'R') .OR. (NEWTYPE(1) .EQ. 'r') .OR.
     &        (NEWTYPE(1) .EQ. 'F') .OR. (NEWTYPE(1) .EQ. 'f')) THEN
        DATATYPE = MPI_REAL4
        DATALTH  = RBYTES
      ELSEIF ((NEWTYPE(1) .EQ. 'D') .OR. (NEWTYPE(1) .EQ. 'd')) THEN
        DATATYPE = MPI_REAL8
        DATALTH  = DBYTES
      ELSEIF ((NEWTYPE(1) .EQ. 'C') .OR. (NEWTYPE(1) .EQ. 'c')) THEN
        DATATYPE = MPI_BYTE
        DATALTH  = 1
      ENDIF
C
      RETURN
C
      END
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
      SUBROUTINE NATSUM(INITIALIZE, MAPSIZE, MAP, MYINDEX, BASE, LTH, 
     &                  INITDATA, WS, SUM) 
C
C NATSUM calls MPI_ALLREDUCE to calculate a vector sum over a 
C specified subset of processors. The MAP array defines the subset and 
C the processor ordering to use. The results are duplicated across all 
C processors in the subset.
C
C called by: FLTSUM
C calls: MPI_ALLREDUCE
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C mpi implementation-specific parameters
#     include "mpif.h"
C
C---- Common Blocks ----------------------------------------------------
C
C definition for communicator and datatype for native MPI implementation
#     include "picl.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C switch indicating whether to compute intialization data (1) or not (0)
      INTEGER INITIALIZE
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in MAP array
      INTEGER MYINDEX
C message type offset to use in interprocessor communication
      INTEGER BASE
C length of vectors to be summed
      INTEGER LTH
C array containing initialization data (unused)
      INTEGER INITDATA(1)
C
C     Work Space
C
C message buffers
      REAL WS(LTH)
C
C     Input/Output
C
C On entry, contains (local) data. On exit, contains vector sum.
      REAL SUM(LTH)
C
C---- Local Variables --------------------------------------------------
C
C error return variable
      INTEGER IERR
C "world" group
      INTEGER GROUP
C "summation" group
      INTEGER SUM_GROUP
C loop index 
      INTEGER I
C
C---- Executable Statements --------------------------------------------
C
      IF (MAPSIZE .GT. 1) THEN
C       Calculate vector sum.
C
C       Calculate group for vector sum. Assuming that this is fixed
C       from call to call.
        IF (SUM_COMM .EQ. MPI_COMM_NULL) THEN
          CALL MPI_COMM_GROUP(PSTSWM_COMM, GROUP, IERR)
          CALL MPI_GROUP_INCL(GROUP, MAPSIZE, MAP, SUM_GROUP, IERR)
          CALL MPI_COMM_CREATE(PSTSWM_COMM, SUM_GROUP, SUM_COMM,
     &                         IERR)
        ENDIF
C
        CALL MPI_ALLREDUCE(SUM, WS, LTH, DATATYPE, MPI_SUM, 
     &                     SUM_COMM, IERR)
        DO I=1,LTH
          SUM(I) = WS(I)
        ENDDO
C
      ENDIF
C
      RETURN
      END
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
      SUBROUTINE SYSTEM_MESH_MAP(ISIZE, JSIZE, MESHMAP)
C
C This subroutine defines the logical processor mesh.
C
C called by: MESH_MAP_INIT
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C mpi implementation-specific parameters
#     include "mpif.h"
C
C---- Common Blocks ----------------------------------------------------
C
C definition for communicator and datatype for native MPI implementation
#     include "picl.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input 
C
C dimensions of logical processor mesh
      INTEGER ISIZE, JSIZE
C
C     Output
C
C logical-to-physical processor mapping
      INTEGER MESHMAP(0:ISIZE-1,0:JSIZE-1)
C
C---- Local Variables --------------------------------------------------
C
C dimensions or coordinates array
      INTEGER DIMS(2)
C periodicity array
      LOGICAL PERIODS(2)
C error return variable
      INTEGER IERR
C index variables
      INTEGER I, J
C
C---- Executable Statements --------------------------------------------
C
C     Generate Cartesian communicator
      DIMS(1) = ISIZE
      DIMS(2) = JSIZE
      PERIODS(1) = .FALSE.
      PERIODS(2) = .FALSE.
      CALL MPI_CART_CREATE(PSTSWM_COMM, 2, DIMS, PERIODS,
     &                     .TRUE., MESHCOMM, IERR)
      CARTFLAG = .TRUE.
C
C     Fill map array
      DO J=0,JSIZE-1
        DO I=0,ISIZE-1
          DIMS(1)=I
          DIMS(2)=J
          CALL MPI_CART_RANK(MESHCOMM, DIMS, MESHMAP(I,J), IERR)
        ENDDO
      ENDDO
C
      RETURN
C
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE NATTRS(MAPID, MAPSIZE, MAP, MYINDEX, BASE, DIR, W, M, 
     &                  N, H1, H2, LM, LN, MX, A, WS, B) 
C
C This subroutine calls routines that compute B = transpose(A) using
C the MPI_ALLTOALL routine. Here
C  A is a matrix of size (W,M,H1,H2,N) distributed by rows and
C  B is a matrix of size (W,N,H1,H2,M) distributed by rows
C over MAPSIZE processors, and each processor has part of A and B as 
C follows:
C  A(W,LM(MYINDEX),H1,H2,N): Processor I has LM(I) rows of A; 
C                            LM(0) + ... + LM(MAPSIZE-1) = M.
C  B(W,LN(MYINDEX),H1,H2,M): Processor I has LN(I) rows of B; 
C                            LN(0) + ... + LN(MAPSIZE-1) = N.
C W is 1 or 2, depending on whether the arrays are REAL or COMPLEX.
C
C Alternative reorganizations of B are also supported, and are 
C determined by the parameter DIR. DIR specifies where TRANSPOSE is 
C called from, allowing the routine to order B as required for 
C subsequent stages in PSTSWM: 
C  DIR=-1: B(W,MX,LN,H1,H2)  (Used after real forward transpose.)
C  DIR=+1: B(W,MX,M,H1,H2)   (Used after real backward transpose.)
C  DIR=-2: B(W,MX,H1,M,H2)   (Used after complex forward transpose
C                             in transpose FFT/transpose LT algorithm)
C  DIR=+2: B(W,MX,H1,LN,H2)  (Used after complex backward transpose
C                             in transpose FFT/transpose LT algorithm)
C  DIR=-3: B(W,MX,LN,M,H2)   (Used after complex forward transpose
C                             in distributed FFT/transpose LT algorithm)
C  DIR=+3: B(W,MX,M,LN,H2)   (Used after complex backward transpose
C                             in distributed FFT/transpose LT algorithm)
C The MAP array defines the processor subset and ordering to use.
C
C called by: TRANSPOSE
C calls: NATTR1
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C a unique integer (1<= MAPID <= 6) for each map array passed
C to transpose routine (used to avoid regeneration of existing groups)
      INTEGER MAPID
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in MAP array
      INTEGER MYINDEX
C message type offset to use in interprocessor communication
      INTEGER BASE
C context in which transpose occurs, and hence required data organization
      INTEGER DIR
C number of reals in datatype (1: REAL, 2: COMPLEX)
      INTEGER W
C dimensions of input and output arrays on processors in MAP array
      INTEGER M, N, H1, H2, MX
      INTEGER LM(0:MAPSIZE-1)
      INTEGER LN(0:MAPSIZE-1)
C local component of the array that is to be transposed, of size
C REAL (W,LM(MYINDEX),H1,H2,N)
      REAL A(1)
C
C     Work Space
C
C message buffers, of size
C REAL (W,LN(MYINDEX),H1,H2,M)
      REAL WS(1)
C
C     Output
C
C local component of the transposed array.
C (organized as REAL (W,MX,LN,H1,H2), (W,MX,M,H1,H2), (W,MX,H1,M,H2),
C  (W,MX,H1,LN,H2), (W,MX,LN,M,H2), or (W,MX,M,LN,H2)) 
      REAL B(1)
C
C---- Executable Statements --------------------------------------------
C
      IF (MAPSIZE .GT. 1) THEN
C       Compute transpose.
C
        CALL NATTR1(MAPID, MAPSIZE, MAP, MYINDEX, BASE, DIR, W, M, N, 
     &              H1, H2, LM, LN, LM(MYINDEX), LN(MYINDEX), MX, A, 
     &              WS, B) 
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE NATTR1(MAPID, MAPSIZE, MAP, MYINDEX, BASE, DIR, W, M, 
     &                  N, H1, H2, LM, LN, ML, NL, MX, A, WS, B) 
C
C This routine computes B = transpose(A) using the MPI_ALLTOALLV 
C routine. Here
C  A is a matrix of size (W,M,H1,H2,N) distributed by rows and
C  B is a matrix of size (W,N,H1,H2,M) distributed by rows
C over MAPSIZE processors, and each processor has part of A and B as 
C follows:
C  A(W,LM(MYINDEX),H1,H2,N): Processor I has LM(I) rows of A; 
C                            LM(0) + ... + LM(MAPSIZE-1) = M.
C  B(W,LN(MYINDEX),H1,H2,M): Processor I has LN(I) rows of B; 
C                            LN(0) + ... + LN(MAPSIZE-1) = N.
C W is 1 or 2, depending on whether the arrays are REAL or COMPLEX.
C
C Alternative reorganizations of B are also supported, and are 
C determined by the parameter DIR. DIR specifies where TRANSPOSE is 
C called from, allowing the routine to order B as required for 
C subsequent stages in PSTSWM: 
C  DIR=-1: B(W,MX,LN,H1,H2)  (Used after real forward transpose.)
C  DIR=+1: B(W,MX,M,H1,H2)   (Used after real backward transpose.)
C  DIR=-2: B(W,MX,H1,M,H2)   (Used after complex forward transpose
C                             in transpose FFT/transpose LT algorithm)
C  DIR=+2: B(W,MX,H1,LN,H2)  (Used after complex backward transpose
C                             in transpose FFT/transpose LT algorithm)
C  DIR=-3: B(W,MX,LN,M,H2)   (Used after complex forward transpose
C                             in distributed FFT/transpose LT algorithm)
C  DIR=+3: B(W,MX,M,LN,H2)   (Used after complex backward transpose
C                             in distributed FFT/transpose LT algorithm)
C The MAP array defines the processor subset and ordering to use.
C
C called by: SRTRANS
C calls: MPI_ALLTOALLV
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C mpi implementation-specific parameters
#     include "mpif.h"
C pstswm-specific parameters
#     include "params.i"
C
C---- Common Blocks ----------------------------------------------------
C
C definition for communicator and datatype for native MPI implementation
#     include "picl.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C a unique integer (1<= MAPID <= 6) for each map array passed
C to transpose routine (used to avoid regeneration of existing groups)
      INTEGER MAPID
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in MAP array
      INTEGER MYINDEX
C message type offset to use in interprocessor communication
      INTEGER BASE
C context in which transpose occurs, and hence required data organization
      INTEGER DIR
C number of reals in datatype (1: REAL, 2: COMPLEX)
      INTEGER W
C dimensions of input and output arrays on processors in MAP array
      INTEGER M, N, H1, H2, ML, NL, MX
      INTEGER LM(0:MAPSIZE-1)
      INTEGER LN(0:MAPSIZE-1)
C Local component of the array that is to be transposed.
C (organized as REAL (W,LM(MYINDEX),H1,H2,N))
      REAL A(W*ML*H1*H2,N)
C
C     Work Space
C
C message buffers
C (organized as REAL (W,LN(MYINDEX),H1,H2,M))
      REAL WS(W*NL*H1*H2,M)
C
C     Output
C
C Local component of the transposed array.
C (organized as REAL (W,MX,NL,H1,H2), (W,MX,M,H1,H2), (W,MX,H1,M,H2),
C  (W,MX,H1,NL,H2), (W,MX,NL,M,H2), or (W,MX,M,NL,H2)) 
      REAL B(1)
C
C---- Local Variables --------------------------------------------------
C
C error return variable
      INTEGER IERR
C "world" group
      INTEGER GROUP
C "summation" group
      INTEGER TRS_GROUP
C number of bytes in a column segment being sent and in a row segment
C being received. 
      INTEGER COLSIZE, ROWSIZE
C arrays indicating beginning indices and sizes of
C message being sent during MPI_ALLTOALVL
      INTEGER SENDDEX(0:NPROCSX-1), SENDOFF(0:NPROCSX-1)
      INTEGER SENDSIZE(0:NPROCSX-1)
C arrays indicating beginning indices and sizes of messages
C message being received during MPI_ALLTOALLV
      INTEGER RECVDEX(0:NPROCSX-1), RECVOFF(0:NPROCSX-1)
      INTEGER RECVSIZE(0:NPROCSX-1)
C loop index
      INTEGER I, I1, I2, I3, I4, I5
C
C---- Executable Statements --------------------------------------------
C
C     Calculate group for transpose. Assuming that this is fixed
C     from call to call for same id.
      IF ((MAPID .GE. 1) .AND. (MAPID .LE. 6)) THEN
C
        IF (TRS_COMM(MAPID) .EQ. MPI_COMM_NULL) THEN
          CALL MPI_COMM_GROUP(PSTSWM_COMM, GROUP, IERR)
          CALL MPI_GROUP_INCL(GROUP, MAPSIZE, MAP, TRS_GROUP, IERR)
          CALL MPI_COMM_CREATE(PSTSWM_COMM, TRS_GROUP, 
     &                         TRS_COMM(MAPID), IERR)
        ENDIF
C
      ELSE
C
        WRITE (0,901) MAPID
  901   FORMAT(/,' PSTSWM: FATAL ERROR IN SUBROUTINE NATTR1:',/,
     &    ' BAD GROUP MAPID SPECIFIED.',/,' MAPID= ',I4)
        STOP                                                   
C        
      ENDIF
C
C     Calculate length of a single column segment (for sending) and a
C     single row segment (for receiving), in bytes.
      COLSIZE = W*LM(MYINDEX)*H1*H2
      ROWSIZE = W*LN(MYINDEX)*H1*H2
C
C     Calculate message sizes and offsets
      DO I=0,MAPSIZE-1
        SENDSIZE(I) = COLSIZE*LN(I)
        RECVSIZE(I) = ROWSIZE*LM(I)
      ENDDO
C
      SENDDEX(0) = 1
      RECVDEX(0) = 1
      SENDOFF(0) = 0
      RECVOFF(0) = 0
      DO I=1,MAPSIZE-1
        SENDDEX(I)  = SENDDEX(I-1) + LN(I-1)
        RECVDEX(I)  = RECVDEX(I-1) + LM(I-1)
        SENDOFF(I)  = SENDOFF(I-1) + SENDSIZE(I-1)
        RECVOFF(I)  = RECVOFF(I-1) + RECVSIZE(I-1)
      ENDDO
C
C     Transpose matrix.
      CALL MPI_ALLTOALLV(A, SENDSIZE, SENDOFF, DATATYPE, WS, RECVSIZE,
     &                   RECVOFF, DATATYPE, TRS_COMM(MAPID), IERR)
C
      DO I=0,MAPSIZE-1
C       Transpose received component into B.
        CALL TRANS(DIR, W, M, H1, H2, LM(I), LN(MYINDEX), MX,
     &             RECVDEX(I), WS(1,RECVDEX(I)), B) 
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
