C#######################################################################
C PSTSWM Version 4.2 (9/1/95)                                          #
C  A message-passing benchmark code and parallel algorithm testbed     #
C  that solves the nonlinear shallow water equations using the spectral#
C  transform method.                                                   #
C Written by:                                                          #
C  Patrick Worley of Oak Ridge National Laboratory                     #
C  Ian Foster of Argonne National Laboratory                           #
C Based on the sequential code STSWM 2.0 by James Hack and Ruediger    #
C  Jakob of the National Center for Atmospheric Research.              #
C Research and development funded by the Computer Hardware, Advanced   #
C  Mathematics, and Model Physics (CHAMMP) program of the U.S.         #
C  Department of Energy.                                               # 
C                                                                      #
C Questions and comments should be directed to worley@msr.epm.ornl.gov #
C Please notify and acknowledge the authors in any research or         #
C publications utilizing PSTSWM or any part of the code.               #
C                                                                      #
C NOTICE: Neither the institutions nor the authors make any            #
C representations about the suitability of this software for any       #
C purpose. This software is provided "as is", without express or       #
C implied warranty.                                                    #
C#######################################################################
C include precision declaration definitions                            #
#include "params.h"
C#######################################################################
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C rft235.f                                                             C
C                                                                      C
C The following routines compute the forward and inverse distributed   C
C real Fourier transform routines, using CFFTF and CFFTB to compute    C
C the sequential forward and inverse complex Fourier transforms.       C
C RFT_235 was written by B. D. Semeraro at Oak Ridge National          C
C Laboratory, and was subsequently modified for inclusion in PSTSWM    C
C by P. H. Worley.                                                     C
C                                                                      C
C External entry points:                                               C
C RFT_235  - driver routine for both forward and inverse real Fourier  C
C            transform.                                                C
C RFT_235I - routine that precomputes the twiddle factors used in the  C
C            real Fourier transform.                                   C
C MDEX     - calculates the global (ordered) index of a Fourier        C
C            given its local (unordered) index.                        C
C                                                                      C
C First level internal entry points:                                   C
C RFT_235F - calculates forward real Fourier transforms of a block of  C
C            distributed vectors.                                      C
C RFT_235F2- calculates forward real Fourier transforms of two blocks  C
C            of distributed vectors, overlapping the butterfly         C
C            calculation phase for one block with the communication    C
C            phase for the other block.                                C 
C RFT_235B - calculates inverse real Fourier transforms of a block of  C
C            distributed vectors.                                      C
C RFT_235B2- calculates inverse real Fourier transforms of two blocks  C
C            of distributed vectors, overlapping the butterfly         C
C            calculation phase for one block with the communication    C
C            phase for the other block.                                C 
C CFFTI, SETTWID2, SETTWID3, SETTWID4, SETTWID5                        C
C          - calculates twiddle factors for 2, 3, 4, 5 complex Fourier C
C            transforms.                                               C
C FAX235   - calculates the 4-2-3-5 factorization of a number.         C
C                                                                      C
C Second level internal entry points:                                  C
C PINIT_235- calculates swap partners and other information needed     C
C            by the distributed Fourier transform algorithms.          C
C BUTUPD   - calculates a "forward" power-of-two butterfly update for  C
C            each vector in a block, representing a factor-of-two stageC
C            of a forward Fourier transform.                           C
C IBTUPD   - calculates an "inverse" power-of-two butterfly update for C
C            each vector in a block, representing a factor-of-two stageC
C            of an inverse Fourier transform.                          C
C EXTRACT  - calculates the forward real Fourier transform from the    C
C            output of a forward complex Fourier transform.            C
C INJECT   - calculates an "inverse" power-of-two butterfly update for C
C            each vector in a block, representing a factor-of-two stageC
C            of an inverse Fourier transform.                          C
C                                                                      C
C Buffer copying routines:                                             C
C BUFCPYA, BUFCPYB, BUFCPYC, BUFCPYD, BUFCPYE                          C
C          - routines for copying between data arrays and work arrays  C
C            in various stages of the distributed real Fourier tranformC
C            algorithms.                                               C
C                                                                      C
C Index manipulation routines:                                         C
C BR, CL, CR, IR, RI                                                   C
C          - routines for the bitwise manipulation of indices:         C
C            bit reversal, left cyclic shift, right cyclic shift.      C
C            and two routines to calculate index reversal.             C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE RFT_235(COMMOPT, PROTOPT, MAPSIZE, MAP, MYINDEX,
     &                   BASE1, BASE2, ISIGN, NLOCAL, MVECS,
     &                   RJUMP, MXTRIGS, TRIG, WS, X)
C
C This subroutine calculates either the forward Fourier transform of 
C MVECS real vectors, each of length N, or the inverse fourier transform
C of MVECS complex vectors of length N/2+1, over a subset of MAPSIZE 
C processors. Here N=NLOCAL*MAPSIZE, and N has factors of 2, 3, or 5.
C It uses a distributed algorithm, i.e., an algorithm that moves the 
C data and partial results only as they are needed. The MAP array 
C defines the subset and the processor ordering to use. The routine 
C RFT_235I must be called before calling RFT_235. 
C
C NOTE: The complex vector of length N/2+1 is packed into a vector of 
C length N/2. The first and last Fourier coefficients are real, and the 
C real part of the last coefficient is packed into the imaginary part
C of the first coefficient. This element should be "unpacked" on the
C processor containing wavenumber 0 following a call to RFT_235 to 
C calculate the forward FFT using code like:
C
C     Unpack (N/2+1)st "complex" coefficient from imaginary
C     part of first coefficient. IM0 is the local index for wavenumber
C     0 and IMNFC is the local index for wavenumber NFC-1
C     IF (IM0 .NE. -1) THEN
C       RDEXNFC = 2*IMNFC-1
C       IDEXNFC = 2*IMNFC
C       IDEX0   = 2*IM0
C       DO I=1,MVECS
C         X(RDEXNFC,I) = X(IDEX0,I)
C         X(IDEXNFC,I) = 0.0
C         X(IDEX0,I)   = 0.0
C       ENDDO
C     ENDIF
C
C Similarly, this element should be "packed" on the processor containing
C wavenumber 0 preceding a call to RFT_235 to calculate the inverse FFT:
C 
C     Pack (N/2+1)st "complex" coefficient into imaginary part of 
C     first coefficient. IM0 is the local index for wavenumber
C     0 and IMNFC is the local index for wavenumber NFC-1
C     IF (IM0 .NE. -1) THEN
C       RDEXNFC = 2*IMNFC-1
C       IDEX0   = 2*IM0
C       DO I=1,MVECS
C         X(IDEX0,I) = X(RDEXNFC,I)
C       ENDDO
C     ENDIF
C
C Communication options (COMMOPT) for RFT_235 include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/sum
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/sum
C  IF (COMMOPT .EQ. 2) simple swap with recv-ahead
C  IF (COMMOPT .EQ. 3) ordered swap with recv-ahead
C Communication protocol options (PROTOPT) for RFT_235 include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5)     nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  nonblocking receive
C  IF (PROTOPT .EQ. 4 .OR. 5)         forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C called by: RFTLON
C calls: RFT_235B, RFT_235B2, RFT_235F, RFT_235F2
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
C define NLONX, the maximum vector length supported by this routine
#     include "params.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in MAP array
      INTEGER MYINDEX
C message type offsets to use in interprocessor communication
      INTEGER BASE1, BASE2
C direction of FFT (-1 is forward, +1 is inverse)
      INTEGER ISIGN
C the local length of each vector to be transformed
      INTEGER NLOCAL
C the number of vectors per processor to be transformed
      INTEGER MVECS
C the offset between the start of successive vectors
      INTEGER RJUMP
C the factorization of N (= NLOCAL*MAPSIZE)
      INTEGER NFAX(13)
C size of array containing trigonometric function values used by RFT_235 
      INTEGER MXTRIGS
C trigonometric function values used by RFT_235 
      COMPLEX TRIG(MXTRIGS/2,2)
C
C     Input/Output
C
C For a forward transform, (ISIGN .EQ. -1), X must contain the real data
C to be transformed. The Mth vector to be transformed extends from
C X((M-1)*JUMP+1) to X(M*JUMP). For an inverse transform, 
C (ISIGN .EQ. +1), X must contain the half-length complex data. Thus X 
C contains MVECS complex vectors each of length N/2+1. The real and 
C imaginary parts of one of these vectors are stored as consecutive 
C elements of X (with the exception of the first and last elements, as 
C described above).  Hence the real and imaginary parts of the Ith 
C component of the Jth complex vector are stored at
C X((J-1)*JUMP+2*I-1) and X((J-1)*JUMP+2*I). 
      REAL X(RJUMP,MVECS)
C
C     Work Space
C
C communication buffers
      REAL WS(RJUMP,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C base 2 logarithm of the number of processors (also, approximate number
C of steps in the distributed phase of the Fourier transform algorithm)
      INTEGER LGP
C number of complex vectors in each block in the two block algorithms
      INTEGER MVEC1, MVEC2
C
C---- External Functions -----------------------------------------------
C
C integer base 2 logarithm
      EXTERNAL LOG2
      INTEGER LOG2
C
C---- Executable Statements --------------------------------------------
C
C     CHeck correctness of input values.
      LGP = LOG2(MAPSIZE)
      IF (2**LGP .NE. MAPSIZE) THEN
        WRITE(0,100) MAPSIZE
  100   FORMAT (/,' PSTSWM: fatal error in subroutine RFT_235 ',/,
     &          ' MAPSIZE must be a power of two to use the ',/,
     &          ' distributed Fourier transform algorithm ',/,
     &          ' MAPSIZE = ',I4)
        STOP
      ENDIF
C
      IF (NLOCAL .GT. NLONX) THEN
        WRITE(6,101) NLOCAL, NLONX
  101   FORMAT (/,' PSTSWM: fatal error in subroutine RFT_235 ',/,
     &          ' NLOCAL out of range ',/,
     &          ' NLOCAL = ',I4, ' MAXIMUM = ',I4)
        STOP
      ENDIF
C
C     Call appropriate RFT routine.
      IF (ISIGN .EQ. -1) THEN
C       forward RFT
C
        IF ( MVECS .GE. 1 ) THEN
C         something to do
C
          IF ((COMMOPT .LE. 1) .OR. (MAPSIZE .EQ. 1)) THEN
C           one block (no overlap) algorithm
            CALL RFT_235F(COMMOPT, PROTOPT, MAPSIZE, MAP, MYINDEX, 
     &                    BASE1, NLOCAL, MVECS, RJUMP, TRIG(1,1), WS, X)
          ELSE
C           two blocks (overlap) algorithm
            MVEC1 = (MVECS+1)/2
            MVEC2 = MVECS - MVEC1
            CALL RFT_235F2(COMMOPT-2, PROTOPT, MAPSIZE, 
     &                  MAP, MYINDEX, BASE1, BASE2,
     &                  NLOCAL, MVEC1, MVEC2, RJUMP,
     &                  TRIG(1,1), WS, WS(1,MVEC1+1), X,
     &                  X(1,MVEC1+1)) 
          ENDIF
C
        ENDIF
C
      ELSEIF (ISIGN .EQ. 1) THEN
C       inverse RFT
C
        IF ( MVECS .GE. 1 ) THEN
C         something to do
C
          IF ((COMMOPT .LE. 1) .OR. (MAPSIZE .EQ. 1)) THEN
C           one block (no overlap) algorithm
            CALL RFT_235B(COMMOPT, PROTOPT, MAPSIZE, MAP, MYINDEX, 
     &                    BASE1, NLOCAL, MVECS, RJUMP, TRIG(1,2), WS, X)
          ELSE
C           two blocks (overlap) algorithm
            MVEC1 = (MVECS+1)/2
            MVEC2 = MVECS - MVEC1
            CALL RFT_235B2(COMMOPT-2, PROTOPT, MAPSIZE, 
     &                  MAP, MYINDEX, BASE1, BASE2,
     &                  NLOCAL, MVEC1, MVEC2, RJUMP,
     &                  TRIG(1,2), WS,
     &                  WS(1,MVEC1+1), X, X(1,MVEC1+1)) 
C
          ENDIF
C
        ENDIF
C
      ENDIF     
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE RFT_235I(NR, MAPSIZE, MYINDEX, MXTRIGS, TRIG)
C
C This routine calculates and returns the COMPLEX multipliers used in 
C the forward and inverse  real Fourier transform routines. It must be
C called before RFT_235 for a given decomposition and vector length.
C
C called by: INPUT
C calls: BR, CFFTI, CL, CR, FAX235, LOG2, IR, RI
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C The number of points to be transformed: N must be a power of 2, 3, or
C 5, and there must be at least 4 data points per processor.
      INTEGER NR
C number of processors over which the Fourier transform is to be
C distributed
      INTEGER MAPSIZE
C index of "me" in the set of processors
      INTEGER MYINDEX
C size of array containing trigonometric function values to be generated
C It should be at least as large as 
C 2*((LOG2(MAPSIZE)+4)*NR/(4*MAPSIZE) -1).
      INTEGER MXTRIGS
C
C     Output
C
C array containing twiddle factors used in RFT_235
      COMPLEX TRIG(0:(MXTRIGS/2)-1,2)
C
C---- Local Variables --------------------------------------------------
C
C real zero and one, to allow easy precision modifications
      REAL ZERO, ONE
C square root of -1
      COMPLEX EYE
C pi and -2*pi/nr
      REAL PI, TWOPIN
C number of complex values in a real array of size NR
      INTEGER NC
C number of complex values local to each processor
C respectively
      INTEGER NCLOC
C array containing factorization of NCLOC
      INTEGER IFAX(13)
C log2 of MAPSIZE
      INTEGER D
C number of complex values in each local vector segment swapped during
C a parallel stage 
      INTEGER NCOMM
C number of distributed steps in complex FFT
      INTEGER SSTOP
C temporaries used in calculating twiddle factors
      INTEGER I0, M, ODD, PTR
C loop indices
      INTEGER S, Q
C other array indices
      INTEGER ICT
C
C---- External Functions -----------------------------------------------
C
C integer base 2 logarithm
      EXTERNAL LOG2
      INTEGER LOG2
C
C index manipulation routines
      EXTERNAL BR, CL, CR, IR, RI
      INTEGER BR, CL, CR, IR, RI
C                                                                             
C---- Executable Statements --------------------------------------------
C
C     =========================================
C     Initialize "constants"
C     =========================================
C
      ZERO = 0.0
      ONE  = 1.0
      EYE  = CMPLX(ZERO,ONE)
      PI   = 4.0D0*ATAN(1.0D0)
      TWOPIN = (-2.0*PI)/NR
C
C     ===========================================================
C     Determine the number of elements of x which are local
C     (NCLOC), and LOG2 of number of processors
C     ===========================================================
C
      NC    = NR/2
      NCLOC = NC/MAPSIZE
      D     = LOG2(MAPSIZE)
C
C     ====================================================================
C     The next section of the code figures out the "parallel"
C     twiddle factors. Set sstop to d if d ne 0 and to -1 otherwise. 
C     If d = 0 then there are no parallel stages and the factors do not 
C     need to be calculated. Otherwise d "stages" of factors are computed.
C     ====================================================================
C
      IF (D .NE. 0) THEN
        SSTOP = D
      ELSE
        SSTOP = -1
      ENDIF
      ICT = 0
      NCOMM = NCLOC/2
C
      M = NC
      IF (MYINDEX .LT. MAPSIZE/2) THEN
        ODD = 0
      ELSE
        ODD = 1
      ENDIF
C
C     ====================================================================
C     Set up the parallel stage COMPLEX multipliers for the half length
C     COMPLEX data.
C     ====================================================================
C
      DO S=0,SSTOP
C
        IF (S .LT. D) THEN
          PTR = NCLOC*MOD(MYINDEX,MAPSIZE/2**(S+1)) + ODD*NCOMM
        ELSE
          PTR = 0
        ENDIF
C
        IF (S .LT. D) THEN
C
          DO Q=0,NCOMM-1
            TRIG(Q+ICT,1) = 
     &        EXP(CMPLX(ZERO,-((Q+PTR)*(2.0*PI))/M))
            TRIG(Q+ICT,2) = 
     &        EXP(CMPLX(ZERO,+((Q+PTR)*(2.0*PI))/M))
          ENDDO
C
          M=M/2
          ICT = ICT + NCOMM
C
        ENDIF
C
      ENDDO
C
C     ====================================================================
C        local multiplier calculation.
C     ====================================================================
C
      CALL CFFTI(NCLOC, TRIG(ICT,1), TRIG(ICT,2))
C
C     ====================================================================
C     calculate the information used in the extraction of the coefficients 
C     of the REAL data. 
C     ====================================================================
C
      ICT = ICT + (NCLOC-1)
C
      M = BR(CR(MYINDEX,1,D),D)
      CALL FAX235(NCLOC, IFAX)
      IF (MYINDEX .EQ. 0) THEN
        DO Q=1,NCLOC-1
          I0 = 2**D*RI(Q,NCLOC,IFAX)
          TRIG(Q+ICT,1) = EYE*EXP(CMPLX(ZERO,TWOPIN*I0))
          TRIG(Q+ICT,2) = EYE*CONJG(EXP(CMPLX(ZERO,TWOPIN*I0)))
        ENDDO
      ELSE
        DO Q=0,NCLOC/2-1
          I0 = 2**D*RI(Q,NCLOC,IFAX)+M
          TRIG(Q+ICT,1) = EYE*EXP(CMPLX(ZERO,+TWOPIN*I0))
          TRIG(NCLOC-Q-1+ICT,1) = EYE*EXP(CMPLX(ZERO,+TWOPIN*(NC-I0)))
          TRIG(Q+ICT,2) = EYE*EXP(CMPLX(ZERO,-TWOPIN*I0))
          TRIG(NCLOC-Q-1+ICT,2) = EYE*EXP(CMPLX(ZERO,-TWOPIN*(NC-I0)))
        ENDDO
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION MDEX(I, MYINDEX, MAPSIZE, N)
C
C MDEX returns the true (global, ordered) index of a Fourier coefficient
C given its local (unordered) index I on processor PID in a subset of
C NPROCS processors. If I is not a legal local Fourier coefficient
C index, then MDEX returns -1. This routine should be used with
C RFT_235, written by D. B. Semeraro (and modified by P. H. Worley).
C
C called by: ALGINP
C calls: BR, CR, LOG2, RI
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C local Fourier index
      INTEGER I
C (relative) processor index
      INTEGER MYINDEX
C number of processors collaborating in distributed real Fourier
C transform
      INTEGER MAPSIZE
C number of local Fourier coefficients
      INTEGER N
C
C---- Local Variables --------------------------------------------------
C
C log2(MAPSIZE)
      INTEGER D
C temporary used to hold index manipulations
      INTEGER M
C factorization of N
      INTEGER IFAX(13)
      SAVE IFAX
C remember whether this is the first time routine called with this
C vector length. If yes, then compute the factorization.
      INTEGER NSAVE
      SAVE NSAVE
C
C---- External Functions -----------------------------------------------
C
C integer base 2 logarithm
      EXTERNAL LOG2
      INTEGER LOG2
C
C index manipulation routines
      EXTERNAL BR, CR, RI
      INTEGER BR, CR, RI
C
C---- Data Statements --------------------------------------------------
C
      DATA NSAVE /0/
C
C---- Executable Statements --------------------------------------------
C
C     If necessary, compute factorization
      IF (N .NE. NSAVE) THEN
        CALL FAX235(N, IFAX)
        NSAVE = N
      ENDIF
C
      D = LOG2(MAPSIZE)
      M = BR(CR(MYINDEX,1,D),D)
      IF (I .EQ. N) THEN
C       Unless on processor 0, N is not a legal local index. On
C       processor 0, it represents the largest Fourier wavenumber.
        IF (MYINDEX .EQ. 0) THEN
          MDEX = N*MAPSIZE
        ELSE 
          MDEX = -1
        ENDIF
      ELSE
        IF(I .LT. N/2 .OR. MYINDEX .EQ. 0) THEN
           MDEX = MAPSIZE*RI(I,N,IFAX)+M
        ELSE
           MDEX = MAPSIZE*(N - RI(N-I-1,N,IFAX)) - M
        ENDIF
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE RFT_235F(COMMOPT, PROTOPT, MAPSIZE, MAP, MYINDEX, BASE,
     &                    NLOC, MVECS, RJUMP, TRIG, WS, X)
C
C This subroutine calculates the forward Fourier transform of MVECS
C real vectors, each of length N, over a subset of MAPSIZE processors. 
C Here N=NLOC*MAPSIZE. It uses a distributed algorithm, i.e., an
C algorithm that moves the data and partial results only as they are
C needed, and transforms all M vectors as a single block. The MAP array
C defines the subset and the processor ordering to use. 
C
C Communication options (COMMOPT) for RFT_235F include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/sum
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/sum
C Communication protocol options (PROTOPT) for RFT_235F include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5)     nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  nonblocking receive
C  IF (PROTOPT .EQ. 4 .OR. 5)         forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C called by: RFT_235
C calls: BUFCPYA, BUFCPYB, BUFCPYE, BUTUPD, EXTRACT, PINIT_235, CFFTF,
C        SWAP
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in map array
      INTEGER MYINDEX
C message type offset to use in interprocessor communication
      INTEGER BASE
C the local length of each vector to be transformed
      INTEGER NLOC
C the number of vectors per processor to be transformed in Y
      INTEGER MVECS
C the offset between the start of successive vectors.
      INTEGER RJUMP
C trigonometric function values used in forward real Fourier transform
      COMPLEX TRIG(*)
C
C     Work Space
C
C work array
      REAL WS((RJUMP*MVECS)/2,2)
C
C     Input/Output
C
C X contains MVECS complex vectors each of length NLOC/2. The real and
C imaginary parts of these vectors are stored as consecutive elements of
C X. Hence the real and imaginary parts of the Ith component of the Jth
C complex vector in X are stored at X(2*I-1,J) and X(2*I,J).
      REAL X(RJUMP,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C variables used to determine whether MAPSIZE and MYINDEX for this call
C are the same as in the previous call, allowing old values of SWAPNODE, 
C etc. to be reused.
      INTEGER PSAVE, MESAVE
      SAVE PSAVE, MESAVE
C number of "computational" swaps in distributed phase of transform
C algorithm. Total number of swaps includes one before sequential
C phase, and possibly one before real fix-up phase.
      INTEGER MAXSTEP
      SAVE MAXSTEP
C arrays indicating the destination of the message sent during a
C given swap, and whether this processor sends or receives first
      INTEGER SWAPNODE(0:LGPROCSX)
      INTEGER ORDER(0:LGPROCSX)
      SAVE SWAPNODE, ORDER
C which half (0: low, 1: high) of the data buffer that is sent during a 
C given swap
      INTEGER DIRN(0:LGPROCSX)
      SAVE DIRN
C index of the processor whose data is needed in order to compute the 
C real transform from the complex transform, and order in which swap 
C is to be performed
      INTEGER TWINDEX, TWORDER
      SAVE TWINDEX, TWORDER
C switch indicating whether thie is process 0 (0) and needs to permute
C the indices when doing the real fix-up, or whether this process needs 
C to do a final swap in order to extract the real transform (2). 
C Otherwise the value is 1.
      INTEGER TYPE
      SAVE TYPE
C number of complex values in real arrays of size NLOC
      INTEGER NCLOC
C number of complex values in each local vector segment swapped during
C a parallel stage 
      INTEGER NCOMM
C length of message swapped during each parallel stage (in bytes) 
      INTEGER BUFSIZ
C index offset into trigonometric function values array
      INTEGER ICT
C index offsets
      INTEGER OFFSET, OFFSET2
C loop index
      INTEGER S
C
C---- Data Statements --------------------------------------------------
C
      DATA PSAVE /0/
      DATA MESAVE /-1/
C                                                                             
C---- Executable Statements --------------------------------------------
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     initialization
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      IF ((MAPSIZE .NE. PSAVE) .OR. (MYINDEX .NE. MESAVE)) THEN
C       Calculate swap partners (and swap order for (COMMOPT .EQ. 1)
C       protocol).  
        CALL PINIT_235(MAPSIZE, MYINDEX, LGPROCSX, MAXSTEP,
     &                 SWAPNODE, ORDER, DIRN, TWINDEX, TWORDER)
C
C       Specify whether real fix-up stage requires a swap (2) or 
C       an index permutation (0).
        IF (MYINDEX .EQ. 0) THEN
          TYPE = 0
        ELSEIF (TWINDEX .EQ. MYINDEX) THEN
          TYPE = 1
        ELSE
          TYPE = 2
        ENDIF
C
        PSAVE  = MAPSIZE
        MESAVE = MYINDEX
      ENDIF
C
C     length of complex vector
      NCLOC = NLOC/2
C
C     number of complex values in each local vector segment swapped
C     during a parallel stage 
      NCOMM = NCLOC/2
C
C     length of message swapped during each parallel stage (in bytes) 
      BUFSIZ = RBYTES*NCLOC*MVECS
C
C     Initialize the pointer to the array of twiddle factors.
      ICT = 1
C
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     Calculate forward real Fourier transform.
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     ==============================================
C     parallel COMPLEX transform on half length data
C     ==============================================
C
C     Do the parallel computations first
      IF (MAXSTEP .NE. 0) THEN
C
C       Copy X into work arrays.
        call traceeventf('entry', 21, 0, 0)
        CALL BUFCPYA(NCLOC, RJUMP, MVECS, X, WS(1,1), WS(1,2))
        call traceeventf('exit', 21, 0, 0)
C
        DO S=0,MAXSTEP-1
C
C         Swap one half of data with current partner.
          call traceeventf('entry', 22, 0, 0)
          OFFSET = 2 - DIRN(S)
          CALL SWAP(COMMOPT, PROTOPT, ORDER(S), MAP(MYINDEX),
     &              BASE, MAP(SWAPNODE(S)), BUFSIZ, WS(1,OFFSET),
     &              BUFSIZ, X)
          call traceeventf('exit', 22, 0, 0)
C
C         Calculate update using current data.
          call traceeventf('entry', 23, 0, 0)
          CALL BUTUPD(DIRN(S), NCOMM, MVECS, TRIG(ICT), 
     &                WS(1,1), WS(1,2), X)
          call traceeventf('exit', 23, 0, 0)
C
          ICT = ICT + NCOMM
        ENDDO
C
C       Swap one half of data with current partner and refill X array.
        OFFSET  = 2 - DIRN(MAXSTEP)
        OFFSET2 = 1 + DIRN(MAXSTEP)
C
        call traceeventf('entry', 24, 0, 0)
        CALL BUFCPYB(NCLOC, RJUMP, NCLOC*DIRN(MAXSTEP), MVECS, 
     &               WS(1,OFFSET2), X)
        call traceeventf('exit', 24, 0, 0)
C
        call traceeventf('entry', 22, 0, 0)
        CALL SWAP(COMMOPT, PROTOPT, ORDER(MAXSTEP), MAP(MYINDEX),
     &            BASE, MAP(SWAPNODE(MAXSTEP)), BUFSIZ, WS(1,OFFSET),
     &            BUFSIZ, WS(1,OFFSET2))
        call traceeventf('exit', 22, 0, 0)
C
        call traceeventf('entry', 24, 0, 0)
        CALL BUFCPYB(NCLOC, RJUMP, NCLOC*(1-DIRN(MAXSTEP)), MVECS, 
     &               WS(1,OFFSET2), X)
        call traceeventf('exit', 24, 0, 0)
C
      ENDIF
C
C     This part is the local transform and requires no communication.
      call traceeventf('entry', 25, 0, 0)
      IF (MOD(RJUMP,2) .EQ. 0) THEN
        CALL CFFTF(NCLOC, RJUMP/2, MVECS, TRIG(ICT), X(1,1))
      ELSE
        DO S=1,MVECS
          CALL CFFTF(NCLOC, NCLOC, 1, TRIG(ICT), X(1,S))
        ENDDO
      ENDIF
      call traceeventf('exit', 25, 0, 0)
      ICT = ICT + (NCLOC-1)
C
C     ================================================
C     Swap some data in preparation for the extraction
C     of the coefficients of the REAL data.
C     ================================================
C
      IF (TYPE .EQ. 2) THEN
C
        call traceeventf('entry', 26, 0, 0)
        CALL BUFCPYE(NCLOC, RJUMP, NCLOC, MVECS, X, WS(1,1))
        call traceeventf('exit', 26, 0, 0)
C
        call traceeventf('entry', 22, 0, 0)
        CALL SWAP(COMMOPT, PROTOPT, TWORDER, MAP(MYINDEX),
     &            BASE, MAP(TWINDEX), BUFSIZ, WS(1,1),
     &            BUFSIZ, WS(1,2))
        call traceeventf('exit', 22, 0, 0)
C
      ENDIF
C
C     ===============================================
C     Extract the REAL transform from the half length
C     COMPLEX. Only half the frequency spectrum is 
C     needed because the transform of REAL data is
C     conjugate symmetric.
C     ===============================================
C
      call traceeventf('entry', 27, 0, 0)
      IF (MOD(RJUMP,2) .EQ. 0) THEN
        CALL EXTRACT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVECS, 
     &               TRIG(ICT), WS(1,2), X)
      ELSE
        DO S=1,MVECS
          CALL EXTRACT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                 TRIG(ICT), WS((S-1)*NCLOC+1,2), X(1,S))
        ENDDO
      ENDIF
      call traceeventf('exit', 27, 0, 0)
c
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE RFT_235F2(COMMOPT, PROTOPT, MAPSIZE, MAP, 
     &                     MYINDEX, BASE1, BASE2, NLOC, MVEC1, MVEC2,
     &                     RJUMP, TRIG, WS1, WS2, X1, X2)  
C
C This subroutine calculates the forward Fourier transform of 2 sets of
C real vectors, each of length N, over a subset of MAPSIZE processors. 
C Here N=NLOC*MAPSIZE. It uses a distributed algorithm, i.e., an
C algorithm that moves the data and partial results only as they are
C needed, and overlaps the butterfly calculation phase for one block
C with the communication phase for the other block. The MAP array
C defines the subset and the processor ordering to use. 
C
C NOTE: This routine assumes that (MAPSIZE .GE. 2.). If not, there
C is no reason not to use the single block routine RFT_235F.
C
C Communication options (COMMOPT) for RFT_235F2 include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/sum
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/sum
C Communication protocol options (PROTOPT) for RFT_235F2 include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5)     nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  nonblocking receive
C  IF (PROTOPT .EQ. 4 .OR. 5)         forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C Since this is a "send-ahead" algorithm, only the simple exchange 
C (send/recv) communication option is invoked in SWAP_SEND and 
C SWAP_RECV. (Using odd/even swaps would prevent overlap.)
C The odd/even option is between SWAP_SEND and SWAP_RECV for different
C swaps, and is handled explicitly in this routine.
C
C called by: RFT_235
C calls: BUFCPYA, BUFCPYB, BUFCPYE, BUTUPD, EXTRACT, PINIT_235, CFFTF,
C        SWAP_SEND, SWAP_RECV
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in map array
      INTEGER MYINDEX
C message type offsets to use in interprocessor communication
      INTEGER BASE1, BASE2
C the local length of each vector to be transformed
      INTEGER NLOC
C the number of vectors per processor to be transformed in X1
      INTEGER MVEC1
C the number of vectors per processor to be transformed in X2
      INTEGER MVEC2
C the offset between the start of successive vectors in X1 and X2
      INTEGER RJUMP
C trigonometric function values used by in forward real Fourier
C transform 
      COMPLEX TRIG(*)
C
C     Work Space
C
C work array 1
      REAL WS1((MVEC1*RJUMP)/2,2)
C work array 2
      REAL WS2((MVEC2*RJUMP)/2,2)
C
C     Input/Output
C
C X1 and X2 each contain complex vectors of length NLOC/2. X1 contains
C MVEC1 vectors and X2 contains MVEC2 vectors. The real and imaginary
C parts of these vectors are stored as consecutive elements of X1 and
C X2.  Hence the real and imaginary parts of the Ith component of the
C Jth complex vector in X1 are stored at X1(2*I-1,J) and X1(2*I,J).
      REAL X1(RJUMP,MVEC1)
      REAL X2(RJUMP,MVEC2)
C
C---- Local Variables --------------------------------------------------
C
C variables used to determine whether MAPSIZE and MYINDEX for this call
C are the same as in the previous call, allowing old values of SWAPNODE, 
C etc. to be reused.
      INTEGER PSAVE, MESAVE
      SAVE PSAVE, MESAVE
C number of "computational" swaps in distributed phase of transform
C algorithm. Total number of swaps includes one before sequential
C phase, and possibly one before real fix-up phase.
      INTEGER MAXSTEP
      SAVE MAXSTEP
C arrays indicating the destination of the message sent during a
C given swap, and whether this processor sends or receives first
      INTEGER SWAPNODE(0:LGPROCSX)
      INTEGER ORDER(0:LGPROCSX)
      SAVE SWAPNODE, ORDER
C which half (0: low, 1: high) of the data buffer that is sent during a 
C given swap
      INTEGER DIRN(0:LGPROCSX)
      SAVE DIRN
C index of the processor whose data is needed in order to compute the 
C real transform from the complex transform, and order in which swap 
C is to be performed
      INTEGER TWINDEX, TWORDER
      SAVE TWINDEX, TWORDER
C switch indicating whether thie is process 0 (0) and needs to permute
C the indices when doing the real fix-up, or whether this process needs 
C to do a final swap in order to extract the real transform (2). 
C Otherwise the value is 1.
      INTEGER TYPE
      SAVE TYPE
C number of complex values in real arrays of size NLOC
      INTEGER NCLOC
C number of complex values in each local vector segment swapped during
C a parallel stage 
      INTEGER NCOMM
C lengths of message swapped during each parallel stage (in bytes) 
      INTEGER BUFSZ1, BUFSZ2
C index offsets into trigonometric function values array
      INTEGER ICT1, ICT2
C index offsets
      INTEGER OFFSET, OFFSET2
C loop index
      INTEGER S
C
C---- Data Statements --------------------------------------------------
C
      DATA PSAVE /0/
      DATA MESAVE /-1/
C                                                                             
C---- Executable Statements --------------------------------------------
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     initialization
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      IF ((MAPSIZE .NE. PSAVE) .OR. (MYINDEX .NE. MESAVE)) THEN
C       Calculate swap partners (and swap order for (COMMOPT .EQ. 1)
C       protocol).  
        CALL PINIT_235(MAPSIZE, MYINDEX, LGPROCSX, MAXSTEP,
     &                 SWAPNODE, ORDER, DIRN, TWINDEX, TWORDER)
C
C       Specify whether real fix-up stage requires a swap (2) or 
C       an index permutation (0).
        IF (MYINDEX .EQ. 0) THEN
          TYPE = 0
        ELSEIF (TWINDEX .EQ. MYINDEX) THEN
          TYPE = 1
        ELSE
          TYPE = 2
        ENDIF
C
        PSAVE  = MAPSIZE
        MESAVE = MYINDEX
      ENDIF
C
C     length of complex vector
      NCLOC = NLOC/2
C
C     number of complex values in each local vector segment swapped
C     during a parallel stage 
      NCOMM = NCLOC/2
C
C     lengths of messages swapped during each parallel stage (in bytes) 
      BUFSZ1 = RBYTES*NCLOC*MVEC1
      BUFSZ2 = RBYTES*NCLOC*MVEC2
C
C     Initialize the pointer to the array of twiddle factors.
      ICT1 = 1
      ICT2 = 1
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     Calculate forward real Fourier transform. See RFT_235F  for a more 
C     readable outline of the algorithm.
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     Prime the pipeline: send X1, send X2, recv X1, update X1.
      CALL BUFCPYA(NCLOC, RJUMP, MVEC1, X1, WS1(1,1), WS1(1,2)) 
      OFFSET = 2-DIRN(0)
      CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1, 
     &               MAP(SWAPNODE(0)), 
     &               BUFSZ1, WS1(1,OFFSET), BUFSZ1, X1) 
C
      CALL BUFCPYA(NCLOC, RJUMP, MVEC2, X2, WS2(1,1), WS2(1,2)) 
      IF ((ORDER(0) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(0)), BUFSZ2, WS2(1,OFFSET), 
     &                 BUFSZ2, X2) 
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(0)), BUFSZ1, X1) 
      ELSE
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(0)), BUFSZ1, X1) 
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(0)), BUFSZ2, WS2(1,OFFSET), 
     &                 BUFSZ2, X2) 
      ENDIF
C
      CALL BUTUPD(DIRN(0), NCOMM, MVEC1, TRIG(ICT1), 
     &            WS1(1,1), WS1(1,2), X1) 
      ICT1 = ICT1 + NCOMM
C
C     Begin overlapped algorithm.
      DO S=1,MAXSTEP-1
C
C       Finish swap with old partner and start swap with new
C       partner: send X1, recv X2, update X2.
C
        OFFSET = 2-DIRN(S)
        IF ((ORDER(S) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, WS1(1,OFFSET), 
     &                   BUFSZ1, X1) 
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S-1)), BUFSZ2, X2) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S-1)), BUFSZ2, X2) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, WS1(1,OFFSET), 
     &                   BUFSZ1, X1) 
        ENDIF
C
        CALL BUTUPD(DIRN(S-1), NCOMM, MVEC2,
     &              TRIG(ICT2), WS2(1,1), WS2(1,2), X2) 
        ICT2 = ICT2 + NCOMM
C
C       Continue swap with new partner: send X2, recv X1, update X1.
        IF ((ORDER(S) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, X1) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, X1) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
        ENDIF
C
        CALL BUTUPD(DIRN(S), NCOMM, MVEC1,
     &              TRIG(ICT1), WS1(1,1), WS1(1,2), X1) 
        ICT1 = ICT1 + NCOMM
C
      ENDDO
C
C     Finish swap with old partner and start swap with new
C     partner: send X1, recv X2, update X2.
C
      OFFSET  = 2 - DIRN(MAXSTEP)
      OFFSET2 = 1 + DIRN(MAXSTEP)
      CALL BUFCPYB(NCLOC, RJUMP, NCLOC*DIRN(MAXSTEP), MVEC1, 
     &             WS1(1,OFFSET2), X1)
      IF ((ORDER(MAXSTEP) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ1, WS1(1,OFFSET), 
     &                 BUFSZ1, WS1(1,OFFSET2))
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(MAXSTEP-1)), BUFSZ2, X2) 
      ELSE
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(MAXSTEP-1)), BUFSZ2, X2) 
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ1, WS1(1,OFFSET), 
     &                 BUFSZ1, WS1(1,OFFSET2))
      ENDIF
C
      CALL BUTUPD(DIRN(MAXSTEP-1), NCOMM, MVEC2,
     &            TRIG(ICT2), WS2(1,1), WS2(1,2), X2) 
      ICT2 = ICT2 + NCOMM
C
C     Continue swap with new partner: send X2, recv X1, update X1.
      CALL BUFCPYB(NCLOC, RJUMP, NCLOC*DIRN(MAXSTEP), MVEC2, 
     &             WS2(1,OFFSET2), X2)
      IF ((ORDER(MAXSTEP) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET),
     &                 BUFSZ2, WS2(1,OFFSET2))
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ1, WS1(1,OFFSET2))
      ELSE
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ1, WS1(1,OFFSET2))
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET), 
     &                 BUFSZ2, WS2(1,OFFSET2))
      ENDIF
C
      CALL BUFCPYB(NCLOC, RJUMP, NCLOC*(1-DIRN(MAXSTEP)), MVEC1, 
     &             WS1(1,OFFSET2), X1)
C
C     Finish forward real Fourier transform.
      IF (TYPE .NE. 2) THEN
C
C       Compute sequential real RFT for X1.
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTF(NCLOC, RJUMP/2, MVEC1, TRIG(ICT1), X1(1,1))
        ELSE
          DO S=1,MVEC1
            CALL CFFTF(NCLOC, NCLOC, 1, TRIG(ICT1), X1(1,S))
          ENDDO
        ENDIF
        ICT1 = ICT1 + (NCLOC-1)
C
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL EXTRACT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC1, 
     &                 TRIG(ICT1), WS1(1,2), X1)
        ELSE
          DO S=1,MVEC1
            CALL EXTRACT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                   TRIG(ICT1), WS1((S-1)*NCLOC+1,2), X1(1,S))
          ENDDO
        ENDIF
C
C       Finish last dimensional swap: recv X2, update X2.
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET2)) 
        CALL BUFCPYB(NCLOC, RJUMP, NCLOC*(1-DIRN(MAXSTEP)), MVEC2, 
     &               WS2(1,OFFSET2), X2)
C
C       Compute sequential real RFT for X2.
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTF(NCLOC, RJUMP/2, MVEC2, TRIG(ICT2), X2(1,1))
        ELSE
          DO S=1,MVEC2
            CALL CFFTF(NCLOC, NCLOC, 1, TRIG(ICT2), X2(1,S))
          ENDDO
        ENDIF
        ICT2 = ICT2 + (NCLOC-1)
C
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL EXTRACT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC2, 
     &                 TRIG(ICT2), WS2(1,2), X2)
        ELSE
          DO S=1,MVEC2
            CALL EXTRACT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                   TRIG(ICT2), WS2((S-1)*NCLOC+1,2), X2(1,S))
          ENDDO
        ENDIF
C
      ELSE
C
C       Compute sequential real RFT for X1.
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTF(NCLOC, RJUMP/2, MVEC1, TRIG(ICT1), X1(1,1))
        ELSE
          DO S=1,MVEC1
            CALL CFFTF(NCLOC, NCLOC, 1, TRIG(ICT1), X1(1,S))
          ENDDO
        ENDIF
        ICT1 = ICT1 + (NCLOC-1)
C
C       Begin swap with TWIN and finish last dimensional swap:
C       send X1, recv X2, update X2.
        CALL BUFCPYE(NCLOC, RJUMP, NCLOC, MVEC1, X1, WS1(1,1))
        IF ((TWORDER .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(TWINDEX), BUFSZ1, WS1(1,1), BUFSZ1, WS1(1,2))
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET2))
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET2))
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(TWINDEX), BUFSZ1, WS1(1,1), BUFSZ1, WS1(1,2))
        ENDIF
C
        CALL BUFCPYB(NCLOC, RJUMP, NCLOC*(1-DIRN(MAXSTEP)), MVEC2, 
     &               WS2(1,OFFSET2), X2)
C
C       Compute sequential real RFT for X2.
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTF(NCLOC, RJUMP/2, MVEC2, TRIG(ICT2), X2(1,1))
        ELSE
          DO S=1,MVEC2
            CALL CFFTF(NCLOC, NCLOC, 1, TRIG(ICT2), X2(1,S))
          ENDDO
        ENDIF
        ICT2 = ICT2 + (NCLOC-1)
C
C       Continue swap with TWIN: send X2, recv X1, update X1.
        CALL BUFCPYE(NCLOC, RJUMP, NCLOC, MVEC2, X2, WS2(1,1))
        IF ((TWORDER .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(TWINDEX), BUFSZ2, WS2(1,1), BUFSZ2, WS2(1,2))
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(TWINDEX), BUFSZ1, WS1(1,2)) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(TWINDEX), BUFSZ1, WS1(1,2)) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(TWINDEX), BUFSZ2, WS2(1,1), BUFSZ2, WS2(1,2))
        ENDIF
C
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL EXTRACT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC1, 
     &                 TRIG(ICT1), WS1(1,2), X1)
        ELSE
          DO S=1,MVEC1
            CALL EXTRACT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                   TRIG(ICT1), WS1((S-1)*NCLOC+1,2), X1(1,S))
          ENDDO
        ENDIF
C
C       Finish swap with TWIN: recv X2, update X2.
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(TWINDEX), BUFSZ2, WS2(1,2)) 
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL EXTRACT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC2, 
     &                 TRIG(ICT2), WS2(1,2), X2)
        ELSE
          DO S=1,MVEC2
            CALL EXTRACT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                   TRIG(ICT2), WS2((S-1)*NCLOC+1,2), X2(1,S))
          ENDDO
        ENDIF
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE RFT_235B(COMMOPT, PROTOPT, MAPSIZE, MAP, MYINDEX, BASE,
     &                    NLOC, MVECS, RJUMP, TRIG, WS, X)
C
C This subroutine calculates the inverse (real) Fourier transform of 
C MVECS complex vectors over a subset of MAPSIZE processors, producing 
C MVECS real vectors of length N=NLOC*MAPSIZE. It uses a distributed 
C algorithm, i.e., an algorithm that moves the data and partial results
C only as they are needed, and transforms all MVECS vectors as a single
C block. The MAP array defines the subset and the processor ordering to
C use. 
C
C Communication options (COMMOPT) for RFT_235B include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/sum
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/sum
C Communication protocol options (PROTOPT) for RFT_235B include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5)     nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  nonblocking receive
C  IF (PROTOPT .EQ. 4 .OR. 5)         forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C called by: RFT_235
C calls: BUFCPYB, BUFCPYC, BUFCPYD, IBTUPD, INJECT, PINIT_235, CFFTB,
C        SWAP
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in map array
      INTEGER MYINDEX
C message type offset to use in interprocessor communication
      INTEGER BASE
C the local length of each vector to be transformed
      INTEGER NLOC
C the number of vectors per processor to be transformed in Y
      INTEGER MVECS
C the offset between the start of successive vectors.
      INTEGER RJUMP
C trigonometric function values used in forward real Fourier transform
      COMPLEX TRIG(*)
C
C     Work Space
C
C work array
      REAL WS((RJUMP*MVECS)/2,2)
C
C     Input/Output
C
C X contains MVECS complex vectors each of length NLOC/2. The real and
C imaginary parts of these vectors are stored as consecutive elements of
C X. Hence the real and imaginary parts of the Ith component of the Jth
C complex vector in X are stored at X(2*I-1,J) and X(2*I,J).
      REAL X(RJUMP,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C variables used to determine whether MAPSIZE and MYINDEX for this call
C are the same as in the previous call, allowing old values of SWAPNODE, 
C etc. to be reused.
      INTEGER PSAVE, MESAVE
      SAVE PSAVE, MESAVE
C number of "computational" swaps in distributed phase of transform
C algorithm. Total number of swaps includes one before sequential
C phase, and possibly one before real fix-up phase.
      INTEGER MAXSTEP
      SAVE MAXSTEP
C arrays indicating the destination of the message sent during a
C given swap, and whether this processor sends or receives first
      INTEGER SWAPNODE(0:LGPROCSX)
      INTEGER ORDER(0:LGPROCSX)
      SAVE SWAPNODE, ORDER
C which half (0: low, 1: high) of the data buffer that is sent during a 
C given swap
      INTEGER DIRN(0:LGPROCSX)
      SAVE DIRN
C index of the processor whose data is needed in order to compute the 
C real transform from the complex transform, and order in which swap 
C is to be performed
      INTEGER TWINDEX, TWORDER
      SAVE TWINDEX, TWORDER
C switch indicating whether thie is process 0 (0) and needs to permute
C the indices when doing the real fix-up, or whether this process needs 
C to do a final swap in order to extract the real transform (2). 
C Otherwise the value is 1.
      INTEGER TYPE
      SAVE TYPE
C number of complex values in real arrays of size NLOC
      INTEGER NCLOC
C number of complex values in each local vector segment swapped during
C a parallel stage 
      INTEGER NCOMM
C length of message swapped during each parallel stage (in bytes) 
      INTEGER BUFSIZ
C index offset into trigonometric function values array
      INTEGER ICT
C index offsets
      INTEGER OFFSET, OFFSET2
C loop index
      INTEGER S
C
C---- Data Statements --------------------------------------------------
C
      DATA PSAVE /0/
      DATA MESAVE /-1/
C                                                                             
C---- Executable Statements --------------------------------------------
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     initialization
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      IF ((MAPSIZE .NE. PSAVE) .OR. (MYINDEX .NE. MESAVE)) THEN
C       Calculate swap partners (and swap order for (COMMOPT .EQ. 1)
C       protocol).  
        CALL PINIT_235(MAPSIZE, MYINDEX, LGPROCSX, MAXSTEP,
     &                 SWAPNODE, ORDER, DIRN, TWINDEX, TWORDER)
C
C       Specify whether real fix-up stage requires a swap (2) or 
C       an index permutation (0).
        IF (MYINDEX .EQ. 0) THEN
          TYPE = 0
        ELSEIF (TWINDEX .EQ. MYINDEX) THEN
          TYPE = 1
        ELSE
          TYPE = 2
        ENDIF
C
        PSAVE  = MAPSIZE
        MESAVE = MYINDEX
      ENDIF
C
C     length of complex vector
      NCLOC = NLOC/2
C
C     number of complex values in each local vector segment swapped
C     during a parallel stage 
      NCOMM = NCLOC/2
C
C     length of message swapped during each parallel stage (in bytes) 
      BUFSIZ = RBYTES*NCLOC*MVECS
C
C     Initialize the pointer to the array of twiddle factors.
      ICT = MAXSTEP*NCOMM + NCLOC
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     Calculate inverse real Fourier transform.
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     ===================================================
C     Inject the REAL transform data into the half length
C     COMPLEX data. 
C     ===================================================
C
      call traceeventf('entry', 71, 0, 0)
      IF (MOD(RJUMP,2) .EQ. 0) THEN
        CALL INJECT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVECS, 
     &              TRIG(ICT), WS(1,1), X)
      ELSE
        DO S=1,MVECS
          CALL INJECT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                TRIG(ICT), WS((S-1)*NCLOC+1,1), X(1,S))
        ENDDO
      ENDIF
      call traceeventf('exit', 71, 0, 0)
C
C     =================================================
C     Swap some data to complete the injection process.
C     =================================================
C
      IF (TYPE .EQ. 2) THEN
        call traceeventf('entry', 73, 0, 0)
        CALL SWAP(COMMOPT, PROTOPT, TWORDER, MAP(MYINDEX),
     &            BASE, MAP(TWINDEX), BUFSIZ, WS(1,1),
     &            BUFSIZ, WS(1,2))
        call traceeventf('exit', 73, 0, 0)
C
        call traceeventf('entry', 72, 0, 0)
        CALL BUFCPYB(NCLOC, RJUMP, NCLOC, MVECS, WS(1,2), X)
        call traceeventf('exit', 72, 0, 0)
C
      ENDIF
C
C     ======================================================
C     parallel inverse COMPLEX transform on half length data
C     ======================================================
C
C     Update pointer to the array of twiddle factors.
      ICT = ICT - (NCLOC-1)
C
C     Compute the local inverse FFT first.
      call traceeventf('entry', 74, 0, 0)
      IF (MOD(RJUMP,2) .EQ. 0) THEN
        CALL CFFTB(NCLOC, RJUMP/2, MVECS, TRIG(ICT), X(1,1))
      ELSE
        DO S=1,MVECS
          CALL CFFTB(NCLOC, NCLOC, 1, TRIG(ICT), X(1,S))
        ENDDO
      ENDIF
      call traceeventf('exit', 74, 0, 0)
C
C     Do the parallel stages of the transform.
      IF (MAXSTEP .NE. 0) THEN
C
C       Copy X into work arrays.
        call traceeventf('entry', 75, 0, 0)
        CALL BUFCPYC(NCLOC, RJUMP, MVECS, X, WS(1,1), WS(1,2))
        call traceeventf('exit', 75, 0, 0)
C
        DO S=MAXSTEP,1,-1
C
C         Swap one half of data with current partner.
          call traceeventf('entry', 73, 0, 0)
          OFFSET = 2 - DIRN(S)
          CALL SWAP(COMMOPT, PROTOPT, ORDER(S), MAP(MYINDEX),
     &              BASE, MAP(SWAPNODE(S)), BUFSIZ, WS(1,OFFSET),
     &              BUFSIZ, X)
          call traceeventf('exit', 73, 0, 0)
C
C         Calculate update using current data.
          call traceeventf('entry', 76, 0, 0)
          ICT = ICT - NCOMM
          CALL IBTUPD(DIRN(S), NCOMM, MVECS, TRIG(ICT), 
     &                WS(1,1), WS(1,2), X)
          call traceeventf('exit', 76, 0, 0)
C
        ENDDO
C
C       Swap one half of data with current partner and refill X array.
        OFFSET  = 2 - DIRN(0)
        OFFSET2 = 1 + DIRN(0)
C
        call traceeventf('entry', 75, 0, 0)
        CALL BUFCPYD(NCLOC, RJUMP, NCLOC*DIRN(0), MVECS, 
     &               WS(1,OFFSET2), X)
        call traceeventf('exit', 75, 0, 0)
C
        call traceeventf('entry', 73, 0, 0)
        CALL SWAP(COMMOPT, PROTOPT, ORDER(0), MAP(MYINDEX),
     &            BASE, MAP(SWAPNODE(0)), BUFSIZ, WS(1,OFFSET),
     &            BUFSIZ, WS(1,OFFSET2))
        call traceeventf('exit', 73, 0, 0)
C
        call traceeventf('entry', 75, 0, 0)
        CALL BUFCPYD(NCLOC, RJUMP, NCLOC*(1-DIRN(0)), MVECS, 
     &               WS(1,OFFSET2), X)
        call traceeventf('exit', 75, 0, 0)
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE RFT_235B2(COMMOPT, PROTOPT, MAPSIZE, MAP, 
     &                     MYINDEX, BASE1, BASE2, NLOC, MVEC1, MVEC2,
     &                     RJUMP, TRIG, WS1, WS2, X1, X2)  
C
C This subroutine calculates the inverse (real) Fourier transform of 2
C sets of complex vectors over a subset of MAPSIZE processors, producing
C real vectors of length N=NLOC*MAPSIZE. It uses a distributed
C algorithm, i.e., an algorithm that moves the data and partial results
C only as they are needed, and overlaps the butterfly calculation phase
C for one block with the communication phase for the other block. The
C MAP array defines the subset and the processor ordering to use. 
C
C NOTE: This routine assumes that (MAPSIZE .GE. 2.). If not, there
C is no reason not to use single block routine RFT_235B.
C
C Communication options (COMMOPT) for RFT_235B2 include:
C  IF (COMMOPT .EQ. 0) simple swap: send/recv/sum
C  IF (COMMOPT .EQ. 1) ordered swap: [send/recv]|[recv/send]/sum
C Communication protocol options (PROTOPT) for RFT_235B2 include:
C  IF (PROTOPT .EQ. 1, 3, .OR. 5)     nonblocking send   
C  IF (PROTOPT .EQ. 2, 3, 4, .OR. 5)  nonblocking receive
C  IF (PROTOPT .EQ. 4 .OR. 5)         forcetype          
C  IF (PROTOPT .EQ. 6 .AND. COMMOPT .EQ. 1) synchronous  
C
C Since this is a "send-ahead" algorithm, only the simple exchange 
C (send/recv) communication option is invoked in SWAP_SEND and 
C SWAP_RECV. (Using odd/even swaps would prevent overlap.)
C The odd/even option is between SWAP_SEND and SWAP_RECV for different
C swaps, and is handled explicitly in this routine.
C
C called by: RFT_235
C calls: BUFCPYB, BUFCPYC, BUFCPYD, IBTUPD, INJECT, PINIT_235, CFFTB,
C        SWAP_SEND, SWAP_RECV
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C communication algorithm option
      INTEGER COMMOPT
C communication protocol option 
      INTEGER PROTOPT
C number of processors in subset
      INTEGER MAPSIZE
C processor subset (and processor ordering)
      INTEGER MAP(0:MAPSIZE-1)
C index of "me" in map array
      INTEGER MYINDEX
C message type offsets to use in interprocessor communication
      INTEGER BASE1, BASE2
C the local length of each vector to be transformed
      INTEGER NLOC
C the number of vectors per processor to be transformed in y1
      INTEGER MVEC1
C the number of vectors per processor to be transformed in X2
      INTEGER MVEC2
C the offset between the start of successive vectors.
      INTEGER RJUMP
C trigonometric function values used by in inverse real Fourier
C transform 
      COMPLEX TRIG(*)
C
C     Work Space
C
C work array 1
      REAL WS1((MVEC1*RJUMP)/2,2)
C work array 2
      REAL WS2((MVEC2*RJUMP)/2,2)
C
C     Input/Output
C
C X1 and X2 each contain complex vectors of length NLOC/2. 
C X1 contains MVEC1 vectors and X2 contains MVEC2 vectors. 
C The real and imaginary parts of these vectors are stored as 
C consecutive elements of X1 and X2. Hence the real and imaginary 
C parts of the Ith component of the Jth complex vector in X1 are stored
C at X1(2*I-1,J) and X1(2*I,J). 
      REAL X1(RJUMP,MVEC1)
      REAL X2(RJUMP,MVEC2)
C
C---- Local Variables --------------------------------------------------
C
C variables used to determine whether MAPSIZE and MYINDEX for this call
C are the same as in the previous call, allowing old values of SWAPNODE, 
C etc. to be reused.
      INTEGER PSAVE, MESAVE
      SAVE PSAVE, MESAVE
C number of "computational" swaps in distributed phase of transform
C algorithm. Total number of swaps includes one before sequential
C phase, and possibly one before real fix-up phase.
      INTEGER MAXSTEP
      SAVE MAXSTEP
C arrays indicating the destination of the message sent during a
C given swap, and whether this processor sends or receives first
      INTEGER SWAPNODE(0:LGPROCSX)
      INTEGER ORDER(0:LGPROCSX)
      SAVE SWAPNODE, ORDER
C which half (0: low, 1: high) of the data buffer that is sent during a 
C given swap
      INTEGER DIRN(0:LGPROCSX)
      SAVE DIRN
C index of the processor whose data is needed in order to compute the 
C real transform from the complex transform, and order in which swap 
C is to be performed
      INTEGER TWINDEX, TWORDER
      SAVE TWINDEX, TWORDER
C switch indicating whether thie is process 0 (0) and needs to permute
C the indices when doing the real fix-up, or whether this process needs 
C to do a final swap in order to extract the real transform (2). 
C Otherwise the value is 1.
      INTEGER TYPE
      SAVE TYPE
C number of complex values in real arrays of size NLOC
      INTEGER NCLOC
C number of complex values in each local vector segment swapped during
C a parallel stage 
      INTEGER NCOMM
C lengths of message swapped during each parallel stage (in bytes) 
      INTEGER BUFSZ1, BUFSZ2
C index offsets into trigonometric function values array
      INTEGER ICT1, ICT2
C index offsets
      INTEGER OFFSET, OFFSET2
C loop index
      INTEGER S
C
C---- Data Statements --------------------------------------------------
C
      DATA PSAVE /0/
      DATA MESAVE /-1/
C                                                                             
C---- Executable Statements --------------------------------------------
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     initialization
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      IF ((MAPSIZE .NE. PSAVE) .OR. (MYINDEX .NE. MESAVE)) THEN
C       Calculate swap partners (and swap order for (COMMOPT .EQ. 1)
C       protocol).  
        CALL PINIT_235(MAPSIZE, MYINDEX, LGPROCSX, MAXSTEP,
     &                 SWAPNODE, ORDER, DIRN, TWINDEX, TWORDER)
C
C       Specify whether real fix-up stage requires a swap (2) or 
C       an index permutation (0).
        IF (MYINDEX .EQ. 0) THEN
          TYPE = 0
        ELSEIF (TWINDEX .EQ. MYINDEX) THEN
          TYPE = 1
        ELSE
          TYPE = 2
        ENDIF
C
        PSAVE  = MAPSIZE
        MESAVE = MYINDEX
      ENDIF
C
C     length of complex vector
      NCLOC = NLOC/2
C
C     number of complex values in each local vector segment swapped
C     during a parallel stage 
      NCOMM = NCLOC/2
C
C     lengths of messages swapped during each parallel stage (in bytes) 
      BUFSZ1 = RBYTES*NCLOC*MVEC1
      BUFSZ2 = RBYTES*NCLOC*MVEC2
C
C     Initialize the pointer to the array of twiddle factors.
      ICT1 = MAXSTEP*NCOMM + NCLOC
      ICT2 = MAXSTEP*NCOMM + NCLOC
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C     Calculate inverse real Fourier transform. See RFT_235B for a more
C     readable outline of the algorithm.
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      IF (TYPE .NE. 2) THEN
C
C       Calculate sequential real RFT for X1: Inject the REAL transform
C       data into the half length COMPLEX data.
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL INJECT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC1, 
     &                TRIG(ICT1), WS1(1,1), X1)
        ELSE
          DO S=1,MVEC1
            CALL INJECT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                  TRIG(ICT1), WS1((S-1)*NCLOC+1,1), X1(1,S))
          ENDDO
        ENDIF
C
C       Compute the local inverse FFT.
        ICT1 = ICT1 - (NCLOC-1)
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTB(NCLOC, RJUMP/2, MVEC1, TRIG(ICT1), X1(1,1))
        ELSE
          DO S=1,MVEC1
            CALL CFFTB(NCLOC, NCLOC, 1, TRIG(ICT1), X1(1,S))
          ENDDO
        ENDIF
C
C       Prime the pipeline (step 1): send X1.
        CALL BUFCPYC(NCLOC, RJUMP, MVEC1, X1, WS1(1,1), WS1(1,2))
        OFFSET = 2 - DIRN(MAXSTEP)
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(MAXSTEP)), BUFSZ1, WS1(1,OFFSET), 
     &                 BUFSZ1, X1)
C
C       Calculate sequential real RFT for X2: Inject the REAL transform
C       data into the half length COMPLEX data.
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL INJECT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC2, 
     &                TRIG(ICT2), WS2(1,1), X2)
        ELSE
          DO S=1,MVEC2
            CALL INJECT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                  TRIG(ICT2), WS2((S-1)*NCLOC+1,1), X2(1,S))
          ENDDO
        ENDIF
C
C       Compute the local inverse FFT.
        ICT2 = ICT2 - (NCLOC-1)
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTB(NCLOC, RJUMP/2, MVEC2, TRIG(ICT2), X2(1,1))
        ELSE
          DO S=1,MVEC2
            CALL CFFTB(NCLOC, NCLOC, 1, TRIG(ICT2), X2(1,S))
          ENDDO
        ENDIF
C
C       Prime the pipeline (step 2): send X2, recv X1, update X1.
        CALL BUFCPYC(NCLOC, RJUMP, MVEC2, X2, WS2(1,1), WS2(1,2)) 
        IF ((ORDER(MAXSTEP) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ1, X1) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ1, X1) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
        ENDIF
C
        ICT1 = ICT1 - NCOMM
        CALL IBTUPD(DIRN(MAXSTEP), NCOMM, MVEC1, TRIG(ICT1),
     &              WS1(1,1), WS1(1,2), X1) 
C
      ELSE
C
C       Prime the pipeline (step 1): fix X1, send X1.
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL INJECT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC1, 
     &                TRIG(ICT1), WS1(1,1), X1)
        ELSE
          DO S=1,MVEC1
            CALL INJECT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                  TRIG(ICT1), WS1((S-1)*NCLOC+1,1), X1(1,S))
          ENDDO
        ENDIF
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(TWINDEX), BUFSZ1, WS1(1,1), BUFSZ1, WS1(1,2))
C
C       Prime the pipeline (step 2): fix X2, send X2, recv X1, RFT X1. 
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL INJECT(TYPE, NCLOC, RJUMP/2, NCLOC*MAPSIZE, MVEC2, 
     &                TRIG(ICT2), WS2(1,1), X2)
        ELSE
          DO S=1,MVEC2
            CALL INJECT(TYPE, NCLOC, NCLOC, NCLOC*MAPSIZE, 1, 
     &                  TRIG(ICT2), WS2((S-1)*NCLOC+1,1), X2(1,S))
          ENDDO
        ENDIF
        IF ((TWORDER .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(TWINDEX), BUFSZ2, WS2(1,1), BUFSZ2, WS2(1,2))
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(TWINDEX), BUFSZ1, WS1(1,2)) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(TWINDEX), BUFSZ1, WS1(1,2)) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(TWINDEX), BUFSZ2, WS2(1,1), BUFSZ2, WS2(1,2))
        ENDIF
C
        CALL BUFCPYB(NCLOC, RJUMP, NCLOC, MVEC1, WS1(1,2), X1)
        ICT1 = ICT1 - (NCLOC-1)
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTB(NCLOC, RJUMP/2, MVEC1, TRIG(ICT1), X1(1,1))
        ELSE
          DO S=1,MVEC1
            CALL CFFTB(NCLOC, NCLOC, 1, TRIG(ICT1), X1(1,S))
          ENDDO
        ENDIF
C
C       Prime the pipeline (step 3): send X1, recv X2, RFT X2.
        CALL BUFCPYC(NCLOC, RJUMP, MVEC1, X1, WS1(1,1), WS1(1,2)) 
        OFFSET = 2 - DIRN(MAXSTEP)
        IF ((ORDER(MAXSTEP) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ1, WS1(1,OFFSET), 
     &                   BUFSZ1, X1) 
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(TWINDEX), BUFSZ2, WS2(1,2)) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(TWINDEX), BUFSZ2, WS2(1,2)) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ1, WS1(1,OFFSET), 
     &                   BUFSZ1, X1) 
        ENDIF
C
        CALL BUFCPYB(NCLOC, RJUMP, NCLOC, MVEC2, WS2(1,2), X2)
        ICT2 = ICT2 - (NCLOC-1)
        IF (MOD(RJUMP,2) .EQ. 0) THEN
          CALL CFFTB(NCLOC, RJUMP/2, MVEC2, TRIG(ICT2), X2(1,1))
        ELSE
          DO S=1,MVEC2
            CALL CFFTB(NCLOC, NCLOC, 1, TRIG(ICT2), X2(1,S))
          ENDDO
        ENDIF
C
C       Prime the pipeline (step 4): send X2, recv X1, update X1.
        CALL BUFCPYC(NCLOC, RJUMP, MVEC2, X2, WS2(1,1), WS2(1,2)) 
        IF ((ORDER(MAXSTEP) .EQ. 1) .OR. (COMMOPT. NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ1, X1) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ1, X1) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(MAXSTEP)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
        ENDIF
C
        ICT1 = ICT1 - NCOMM
        CALL IBTUPD(DIRN(MAXSTEP), NCOMM, MVEC1, TRIG(ICT1),
     &              WS1(1,1), WS1(1,2), X1) 
C
      ENDIF
C
C     Begin overlapped algorithm.
      DO S=MAXSTEP-1,1,-1
C
C       Finish swap with old partner and start swap with new
C       partner: send X1, recv X2, update X2.
        OFFSET = 2 - DIRN(S)
        IF ((ORDER(S) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, WS1(1,OFFSET), 
     &                   BUFSZ1, X1)  
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S+1)), BUFSZ2, X2) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S+1)), BUFSZ2, X2) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, WS1(1,OFFSET), 
     &                   BUFSZ1, X1)  
        ENDIF
C
        ICT2 = ICT2 - NCOMM
        CALL IBTUPD(DIRN(S+1), NCOMM, MVEC2, TRIG(ICT2), 
     &              WS2(1,1), WS2(1,2), X2) 
C
C       Continue swap with new partner: send X2, recv X1, update X1.
        IF ((ORDER(S) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, X1) 
        ELSE
          CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                   MAP(SWAPNODE(S)), BUFSZ1, X1) 
          CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                   MAP(SWAPNODE(S)), BUFSZ2, WS2(1,OFFSET), 
     &                   BUFSZ2, X2) 
        ENDIF
C
        ICT1 = ICT1 - NCOMM
        CALL IBTUPD(DIRN(S), NCOMM, MVEC2, TRIG(ICT1), 
     &              WS1(1,1), WS1(1,2), X1) 
C
      ENDDO
C
C     step .eq. maxstep
C
C     Save completed local half of X1.
      OFFSET  = 2 - DIRN(0)
      OFFSET2 = 1 + DIRN(0)
      CALL BUFCPYD(NCLOC, RJUMP, NCLOC*DIRN(0), MVEC1, 
     &             WS1(1,OFFSET2), X1)
C
C     Finish swap with old partner and start swap with new
C     partner: send X1, recv X2, update X2.
      IF ((ORDER(0) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(0)), BUFSZ1, WS1(1,OFFSET), 
     &                 BUFSZ1, WS1(1,OFFSET2))
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(1)), BUFSZ2, X2) 
      ELSE
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(1)), BUFSZ2, X2) 
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(0)), BUFSZ1, WS1(1,OFFSET), 
     &                 BUFSZ1, WS1(1,OFFSET2))
      ENDIF
C
      ICT2 = ICT2 - NCOMM
      CALL IBTUPD(DIRN(1), NCOMM, MVEC2, TRIG(ICT2), 
     &            WS2(1,1), WS2(1,2), X2) 
C
C     Save completed local half of X2.
      CALL BUFCPYD(NCLOC, RJUMP, NCLOC*DIRN(0), MVEC2, 
     &             WS2(1,OFFSET2), X2)
C
C     Continue swap with new partner: send X2, recv X1, save
C     other half of X1.
      IF ((ORDER(0) .EQ. 1) .OR. (COMMOPT .NE. 1)) THEN
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(0)), BUFSZ2, WS2(1,OFFSET), 
     &                 BUFSZ2, WS2(1,OFFSET2))
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(0)), BUFSZ1, WS1(1,OFFSET2))
      ELSE
        CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE1,
     &                 MAP(SWAPNODE(0)), BUFSZ1, WS1(1,OFFSET2))
        CALL SWAP_SEND(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &                 MAP(SWAPNODE(0)), BUFSZ2, WS2(1,OFFSET), 
     &                 BUFSZ2, WS2(1,OFFSET2))
      ENDIF
C
      CALL BUFCPYD(NCLOC, RJUMP, NCLOC*(1-DIRN(0)), MVEC1, 
     &             WS1(1,OFFSET2), X1)
C
C     Finish last swap: recv X2 and save it.
      CALL SWAP_RECV(0, PROTOPT, 1, MAP(MYINDEX), BASE2,
     &               MAP(SWAPNODE(0)), BUFSZ2, WS2(1,OFFSET2)) 
      CALL BUFCPYD(NCLOC, RJUMP, NCLOC*(1-DIRN(0)), MVEC2, 
     &             WS2(1,OFFSET2), X2)
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE CFFTI(N, TRIGF, TRIGB)
C
C This routine calculates and returns the COMPLEX multipliers used in 
C the forward and inverse complex Fourier transform routines.
C
C called by: RFT_235I
C calls: FAX235, SETTWID2, SETTWID3, SETTWID4, SETTWID5
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C The number of points to be transformed: N must be a power of 2, 3, or
C 5.
      INTEGER N
C
C     Output
C
C arrays containing twiddle factors for forward and inverse transforms,
C respectively
      COMPLEX TRIGF(*), TRIGB(*)
C
C---- Local Variables --------------------------------------------------
C
C pi and -2*pi/len
      REAL PI, TWOPIN
C array containing factorization of N
      INTEGER IFAX(0:12)
C loop index
      INTEGER K
C temporaries for computing array indices and twiddle factors
      INTEGER CPTR, LEN
C                                                                             
C---- Executable Statements --------------------------------------------
C
C     define "constants"
      PI = 4.0D0*ATAN(1.0D0)
C
C     determine factors of N
      CALL FAX235(N, IFAX)
C
C     determine twiddle factors
      LEN = N
      CPTR = 1
      DO K=1,IFAX(0)
C
        TWOPIN = (2.0*PI)/LEN
        LEN = LEN/IFAX(K)
C
        IF (IFAX(K) .EQ. 2) THEN
          CALL SETTWID2(LEN, TRIGF(CPTR), TRIGB(CPTR), TWOPIN)
        ELSEIF (IFAX(K) .EQ. 3) THEN
          CALL SETTWID3(LEN, TRIGF(CPTR), TRIGB(CPTR), TWOPIN)
        ELSEIF (IFAX(K) .EQ. 4) THEN
          CALL SETTWID4(LEN, TRIGF(CPTR), TRIGB(CPTR), TWOPIN)
        ELSEIF (IFAX(K). EQ. 5) THEN
          CALL SETTWID5(LEN, TRIGF(CPTR), TRIGB(CPTR), TWOPIN)
        ELSE
          WRITE(6,100) 
  100     FORMAT (/,' PSTSWM: fatal error in subroutine CFFTI ',/,
     &            ' N contains factors other than 2,3,4,5')
          STOP
        ENDIF
C
        CPTR = CPTR + (IFAX(K)-1)*LEN
C
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE SETTWID2(LEN, TRIGF, TRIGB, TWOPIN)
C
C This routine calculates and returns COMPLEX multipliers for
C factors of two.
C
      IMPLICIT NONE
C
      INTEGER LEN
      COMPLEX TRIGF(0:*), TRIGB(0:*)
      REAL TWOPIN
C
      INTEGER I
      REAL ZERO
c
c     initialize "constants"
      ZERO = 0.0
c
      DO I=0,LEN-1
        TRIGF(I) = EXP(CMPLX(ZERO,-I*TWOPIN))
        TRIGB(I) = EXP(CMPLX(ZERO,+I*TWOPIN))
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE SETTWID3(LEN, TRIGF, TRIGB, TWOPIN)
C
C This routine calculates and returns COMPLEX multipliers for
C factors of three.
C
      IMPLICIT NONE
C
      INTEGER LEN
      COMPLEX TRIGF(0:*), TRIGB(0:*)
      REAL TWOPIN
C
      INTEGER I
      REAL ZERO
c
c     initialize "constants"
      ZERO = 0.0
c
      DO I=0,LEN-1
        TRIGF(I)     = EXP(CMPLX(ZERO,-I*TWOPIN))
        TRIGF(I+LEN) = TRIGF(I)*TRIGF(I)
        TRIGB(I)     = EXP(CMPLX(ZERO,+I*TWOPIN))
        TRIGB(I+LEN) = TRIGB(I)*TRIGB(I)
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE SETTWID4(LEN, TRIGF, TRIGB, TWOPIN)
C
C This routine calculates and returns COMPLEX multipliers for
C factors of four.
C
      IMPLICIT NONE
C
      INTEGER LEN
      COMPLEX TRIGF(0:*), TRIGB(0:*)
      REAL TWOPIN
C
      INTEGER I
      REAL ZERO
c
c     initialize "constants"
      ZERO = 0.0
c
      DO I=0,LEN-1
        TRIGF(I)       = EXP(CMPLX(ZERO,-I*TWOPIN))
        TRIGF(I+LEN)   = TRIGF(I)*TRIGF(I)
        TRIGF(I+2*LEN) = TRIGF(I)*TRIGF(I+LEN)
        TRIGB(I)       = EXP(CMPLX(ZERO,+I*TWOPIN))
        TRIGB(I+LEN)   = TRIGB(I)*TRIGB(I)
        TRIGB(I+2*LEN) = TRIGB(I)*TRIGB(I+LEN)
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE SETTWID5(LEN, TRIGF, TRIGB, TWOPIN)
C
C This routine calculates and returns COMPLEX multipliers for
C factors of five.
C
      IMPLICIT NONE
C
      INTEGER LEN
      COMPLEX TRIGF(0:*), TRIGB(0:*)
      REAL TWOPIN
C
      INTEGER I
      REAL ZERO
c
c     initialize "constants"
      ZERO = 0.0
c
      DO I=0,LEN-1
        TRIGF(I)       = EXP(CMPLX(ZERO,-I*TWOPIN))
        TRIGF(I+LEN)   = TRIGF(I)*TRIGF(I)
        TRIGF(I+2*LEN) = TRIGF(I)*TRIGF(I+LEN)
        TRIGF(I+3*LEN) = TRIGF(I)*TRIGF(I+2*LEN)
        TRIGB(I)       = EXP(CMPLX(ZERO,+I*TWOPIN))
        TRIGB(I+LEN)   = TRIGB(I)*TRIGB(I)
        TRIGB(I+2*LEN) = TRIGB(I)*TRIGB(I+LEN)
        TRIGB(I+3*LEN) = TRIGB(I)*TRIGB(I+2*LEN)
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE FAX235(N, IFAX)
C
C This routine factors an INTEGER into powers of 2,3,4 and 5.
C Most of the logic was lifted from a subroutine of the same name
C in FFT991. See also Swartztraubers FFTPACK routines.
C
C called by: CFFTI, EXTRACT, INJECT, RFT_235I
C calls:
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C integer to be factored
      INTEGER N
C
C     Output
C
C array to hold factorization of N. IFAX(1) contains the number
C of factors of N. If N = 16 this routine would factor N as 4x4. IFAX(1)
C would be 2, IFAX(2) = 4, and IFAX(3) = 4. 
      INTEGER IFAX(13)
C
C---- Local Variables --------------------------------------------------
C
C temporaries used in search for factors
      INTEGER I, J
C array index
      INTEGER K
C factors searched for
      INTEGER FACTOR(5)
C
C---- Data Statements --------------------------------------------------
C
      DATA FACTOR /4,2,3,5,0/
C
C---- Executable Statements -------------------------------------------
C
      I = N
      J = 2
      K = 1
      IFAX(1) = 0
      DO WHILE (FACTOR(K) .NE. 0)
C
C     find factors of FACTOR(K)
C
        DO WHILE(MOD(I,FACTOR(K)) .EQ. 0 .AND. I .NE. 1) 
          IFAX(J) = FACTOR(K)
          J = J+1
          I = I/FACTOR(K)
          IFAX(1) = IFAX(1) + 1
        END DO
C
        K = K+1
C
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE PINIT_235(MAPSIZE, MYINDEX, MXSTEPS, MAXSWAP,
     &                     SWAPNODE, ORDER, DIRN, TWINDEX, TWORDER) 
C
C This routine calculates swap partners and other information needed
C by the distributed Fourier transform algorithms.
C
C called by: RFT_235B, RFT_235F
C calls: BR, CL, CR
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C number of processors in subset
      INTEGER MAPSIZE
C index of "me" in MAP array
      INTEGER MYINDEX
C declared length of the output arrays, at least as long as 
C LOG2(MAPSIZE)
      INTEGER MXSTEPS
C
C     Output
C
C number of "computational" swaps in distributed phase of transform
C algorithm. Total number of swaps includes one before sequential
C phase, and possibly one before real fix-up phase.
      INTEGER MAXSWAP
C arrays indicating the destination of the message sent during a
C given swap, and whether this processor sends or receives first
      INTEGER SWAPNODE(0:MXSTEPS)
      INTEGER ORDER(0:MXSTEPS)
C which half (0: low, 1: high) of the data buffer that is sent during a 
C given swap
      INTEGER DIRN(0:MXSTEPS)
C index of the processor "twin" whose data is needed in order to
C compute the real transform from the complex transform, and order
C in which swap with TWIN should be performed 
      INTEGER TWINDEX, TWORDER
C
C---- Local Variables --------------------------------------------------
C
C distance between MYINDEX and swap partner index
      INTEGER DISTANCE
C index of swap partner
      INTEGER DEST
C log2(MAPSIZE)
      INTEGER D
C
C---- External Functions -----------------------------------------------
C
C bit manipulation routines for calculating twin processor index
      EXTERNAL BR, CL, CR
      INTEGER BR, CL, CR
C
C---- Executable Statements -------------------------------------------
C
      MAXSWAP = 0
      TWINDEX = MYINDEX
C
      IF (MAPSIZE .GT. 1) THEN
C       a distributed Fourier transform
C
        DISTANCE = MAPSIZE/2
C           
        DO WHILE (DISTANCE .GT. 0)
C
C         Calculate swap partners, swap order, and buffer pointers. The
C         order is chosen in such a way as to minimize collisions on a
C         bidirectional grid. 
          IF (MOD(MYINDEX,2*DISTANCE) .LT. DISTANCE) THEN
            DEST = MYINDEX + DISTANCE
            DIRN(MAXSWAP) = 0
            IF (MOD(MYINDEX, 2) .EQ. 0) THEN
              ORDER(MAXSWAP) = 1
            ELSE
              ORDER(MAXSWAP) = -1
            ENDIF
          ELSE
            DEST = MYINDEX - DISTANCE
            DIRN(MAXSWAP) = 1
            IF (MOD(DEST, 2) .EQ. 0) THEN
              ORDER(MAXSWAP) = -1
            ELSE
              ORDER(MAXSWAP) = 1
            ENDIF
          ENDIF
          SWAPNODE(MAXSWAP) = DEST
C
C         Update distance.
          DISTANCE = DISTANCE/2
C
C         Increment step index.
          MAXSWAP = MAXSWAP + 1
C
        ENDDO
C
C       First and last (excepting swap with TWIN) swap partners and
C       order are the same.
        SWAPNODE(MAXSWAP) = SWAPNODE(0)
        DIRN(MAXSWAP) = DIRN(0)
        ORDER(MAXSWAP) = ORDER(0)
C
C       Twin index and order of swap with twin
        IF ((MYINDEX .NE. 0) .AND. (MYINDEX .NE. MAPSIZE/2)) THEN
C
          D = MAXSWAP
          TWINDEX = CL(BR(MAPSIZE-BR(CR(MYINDEX,1,D),D),D),1,D)
C
          IF (MYINDEX .LT. TWINDEX) THEN
            IF (MOD(MYINDEX, 2) .EQ. 0) THEN
              TWORDER = 1
            ELSE
              TWORDER = -1
            ENDIF
          ELSE
            IF (MOD(TWINDEX, 2) .EQ. 0) THEN
              TWORDER = -1
            ELSE
              TWORDER = 1
            ENDIF
          ENDIF
C
        ENDIF
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE BUTUPD(DIRN, VECLTH, MVECS, TRIG, XA, XB, COMBUF)
C
C BUTUPD calculates a "forward" power-of-two butterfly update for each
C vector in X, using VECLTH complex data in XA, XB, and/or and COMMBUF
C to produce a new complex vector of length 2*VECLTH, VECLTH in XA
C and VECLTH in XB, This update represents a factor-of-two stage of a
C forward fast Fourier transform. 
C
C called by: CUBCTF
C calls:
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C switch indicating whether the data for vector J start in COMBUF(1,J)
C and XA(1,J) (DIRIN .EQ. 0) or in XB(1,J) and COMBUF(1,J) 
C (DIRIN .NE. 0) 
      INTEGER DIRN
C length of each complex output vector and (half the length of the
C "true" complex data vectors
      INTEGER VECLTH
C number of vectors being updated
      INTEGER MVECS
C trigonometric function values used in the update
      COMPLEX TRIG(VECLTH)
C half of the data used in the update
      COMPLEX COMBUF(VECLTH,MVECS)
C
C     Input/Output
C
C On input, half of the data for the update is contained in XA or XB. On
C output, XA and XB contain the results of the update. 
      COMPLEX XA(VECLTH,MVECS), XB(VECLTH,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J
C temporary used in computing the inverse power-of-two butterfly update 
      COMPLEX XT
C                                                                             
C---- Executable Statements --------------------------------------------
C
      IF ( DIRN .EQ. 0 ) THEN
C
        DO J=1,MVECS
C         Calculate power-of-two butterfly update.
          DO I=1,VECLTH
            XB(I,J) = (XA(I,J) - COMBUF(I,J))*TRIG(I)
            XA(I,J) = XA(I,J) + COMBUF(I,J)
          ENDDO
        ENDDO
C
      ELSE
C
        DO J=1,MVECS
C         Calculate power-of-two butterfly update.
          DO I=1,VECLTH
            XA(I,J) = COMBUF(I,J) + XB(I,J)
            XB(I,J) = (COMBUF(I,J) - XB(I,J))*TRIG(I)
          ENDDO
        ENDDO
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE IBTUPD(DIRN, VECLTH, MVECS, TRIG, XA, XB, COMBUF)
C
C IBTUPD calculates an "inverse" power-of-two butterfly update for each
C vector in X, using VECLTH complex data in XA, XB, and/or and COMMBUF
C to produce a new complex vector of length 2*VECLTH, VECLTH in XA
C and VECLTH in XB, This update represents a factor-of-two stage of an
C inverse fast Fourier transform. 
C
C called by: RFT_235B
C calls:
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C switch indicating whether the data for vector J start in COMBUF(1,J)
C and XA(1,J) (DIRIN .EQ. 0) or in XB(1,J) and COMBUF(1,J) 
C (DIRIN .NE. 0) 
      INTEGER DIRN
C length of each complex output vector and (half the length of the
C "true" complex data vectors
      INTEGER VECLTH
C number of vectors being updated
      INTEGER MVECS
C trigonometric function values used in the update
      COMPLEX TRIG(VECLTH)
C half of the data used in the update
      COMPLEX COMBUF(MVECS,VECLTH)
C
C     Input/Output
C
C On input, half of the data for the update is contained in XA or XB. On
C output, XA and XB contain the results of the update. 
      COMPLEX XA(MVECS,VECLTH), XB(MVECS,VECLTH)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J
C                                                                             
C---- Executable Statements --------------------------------------------
C
      IF ( DIRN .EQ. 0 ) THEN
C
        DO I=1,VECLTH
          DO J=1,MVECS
C           Calculate power-of-two butterfly update.
            XB(J,I) = XA(J,I) - (COMBUF(J,I)*TRIG(I))
            XA(J,I) = XA(J,I) + (COMBUF(J,I)*TRIG(I))
          ENDDO
        ENDDO
C
      ELSE
C
        DO I=1,VECLTH
          DO J=1,MVECS
C           Calculate power-of-two butterfly update.
            XA(J,I) = COMBUF(J,I) + (XB(J,I)*TRIG(I))
            XB(J,I) = COMBUF(J,I) - (XB(J,I)*TRIG(I))
          ENDDO
        ENDDO
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE EXTRACT(TYPE, NCLOC, CJUMP, NC, MVECS, TRIG, WS, X)
C
C EXTRACT extracts the REAL transform from the half length COMPLEX.
C Only half the frequency spectrum is needed because the transform of 
C REAL data is conjugate symmetric.
C
C called by: RFT_235F
C calls: FAX235
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C switch indicating whether processor 0 (0), the other processor without
C a twin (1), or a processor with a twin (2), thus requiring an 
C additional swap after the "injection."
      INTEGER TYPE
C length of local complex data
      INTEGER NCLOC
C length of first dimension in X
      INTEGER CJUMP
C length of global vector
      INTEGER NC
C number of vectors
      INTEGER MVECS
C trigonometric function values used in the extraction
      COMPLEX TRIG(NCLOC)
C work array that contains half of the data when FFT is distributed.
      COMPLEX WS(NCLOC/2,MVECS)
C
C     Input/Output
C
C on entry, contains complex data; on exit contains extracted real 
C transform
      COMPLEX X(CJUMP,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J, NCOMM, II, III
C normalization factor
      REAL SFAC
C temporaries
      COMPLEX C0PCC1, C2
C factorization of NCLOC
      INTEGER IFAX(13)
C permutation index array for processor 0.
      INTEGER IP(0:NLONX-1)
      SAVE IP
C remember whether this is the first time routine called with this
C vector length. If yes, then compute index permutation.
      INTEGER NSAVE
      SAVE NSAVE
C
C---- External Functions -----------------------------------------------
C
C bit manipulation routine for calculating index permutation
      EXTERNAL IR
      INTEGER IR
C
C---- Data Statements --------------------------------------------------
C
      DATA NSAVE /0/
C                                                                             
C---- Executable Statements --------------------------------------------
C
C     Calculate index permutation to use on processor 0.
      IF (NSAVE .NE. NCLOC) THEN
C
        CALL FAX235(NCLOC, IFAX)
C
        DO I=0,NLONX-1
          IP(I) = -1
        ENDDO
C
        DO I=1,NCLOC/2
          IP(IR(I,NCLOC,IFAX)) = IR(NCLOC-I,NCLOC,IFAX) 
        ENDDO
C
        NSAVE = NCLOC
C
      ENDIF
C
C     Precompute and combine normalization and extraction 
C     factors.
      SFAC = 0.25D0/NC
C
C     Extract.
      IF (TYPE .EQ. 2) THEN
C       Not processor 0, and half of results come from a work array 
C       filled in a previous swap.
C
        NCOMM = NCLOC/2
        DO J=1,MVECS
C
           DO I=1,NCOMM
C
             II  = NCLOC-(I-1)
             III = NCOMM-(I-1)
             C0PCC1  = X(I,J) + CONJG(WS(III,J))
             C2      = (X(I,J) - CONJG(WS(III,J)))*TRIG(I)
             X(I,J)  = SFAC*(C0PCC1 - C2)
             X(II,J) = SFAC*CONJG(C0PCC1 + C2)
C
           ENDDO
C
        ENDDO
C
      ELSEIF (TYPE .EQ. 1) THEN
C       Not processor 0, and no swap preceded this computation.
C
        NCOMM = NCLOC/2
        DO J=1,MVECS
C
           DO I=1,NCOMM
C
             II = NCLOC+1-I
             C0PCC1  = X(I,J) + CONJG(X(II,J))
             C2      = (X(I,J) - CONJG(X(II,J)))*TRIG(I)
             X(I,J)  = SFAC*(C0PCC1 - C2)
             X(II,J) = SFAC*CONJG(C0PCC1 + C2)
C
           ENDDO
C
        ENDDO
C
      ELSE
C       Processor 0.
C
        DO J=1,MVECS
C
          X(1,J) = 2.0*SFAC
     &           * CMPLX(AREAL(X(1,J))+AIMAG(X(1,J)), 
     &                   AREAL(X(1,J))-AIMAG(X(1,J)))
          DO I=2,NCLOC
C
            II = IP(I-1) + 1
            IF (II .NE. 0) THEN
              C0PCC1  = X(I,J)+CONJG(X(II,J))
              C2      = (X(I,J)-CONJG(X(II,J)))*TRIG(I)
              X(I,J)  = SFAC*(C0PCC1 - C2)
              X(II,J) = SFAC*(CONJG(C0PCC1 + C2))
            ENDIF
C
          ENDDO
C
        ENDDO
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE INJECT(TYPE, NCLOC, CJUMP, NC, MVECS, TRIG, WS, X)
C
C INJECT generates half length COMPLEX data from the REAL transform,
C as the first step in the real inverse transform.
C
C called by: RFT_235B
C calls: FAX235
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Parameters -------------------------------------------------------
C
#     include "params.i"
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C switch indicating whether processor 0 (0), the other processor without
C a twin (1), or a processor with a twin (2), thus requiring an 
C additional swap after the "injection."
      INTEGER TYPE
C length of local complex data
      INTEGER NCLOC
C length of first dimension in X
      INTEGER CJUMP
C length of global vector
      INTEGER NC
C number of vectors
      INTEGER MVECS
C trigonometric function values used in the extraction
      COMPLEX TRIG(NCLOC)
C
C     Input/Output
C
C on entry, contains complex data; on exit contains extracted real 
C transform
      COMPLEX X(CJUMP,MVECS)
C
C     Output
C
C for distributed FFT, half of output is put here
      COMPLEX WS(NCLOC/2,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J, NCOMM, II, III
C temporaries
      COMPLEX C0PCC1, C2
C factorization of NCLOC
      INTEGER IFAX(13)
C permutation index array for processor 0.
      INTEGER IP(0:NLONX-1)
      SAVE IP
C remember whether this is the first time routine called with this
C vector length. If yes, then compute index permutation.
      INTEGER NSAVE
      SAVE NSAVE
C
C---- External Functions -----------------------------------------------
C
C bit manipulation routine for calculating index permutation
      EXTERNAL IR
      INTEGER IR
C
C---- Data Statements --------------------------------------------------
C
      DATA NSAVE /0/
C                                                                             
C---- Executable Statements --------------------------------------------
C
C     Calculate index permutation to use on processor 0.
      IF (NSAVE .NE. NCLOC) THEN
C
        CALL FAX235(NCLOC, IFAX)
C
        DO I=0,NLONX-1
          IP(I) = -1
        ENDDO
C
        DO I=1,NCLOC/2
          IP(IR(I,NCLOC,IFAX)) = IR(NCLOC-I,NCLOC,IFAX) 
        ENDDO
C
        NSAVE = NCLOC
C
      ENDIF
C
C     Inject.
      IF (TYPE .EQ. 2) THEN
C       Not processor 0, and half of results go into work array in 
C       preparation for a swap.
C
        NCOMM = NCLOC/2
        DO J=1,MVECS
C
          DO I=1,NCOMM
C
            II  = NCLOC-(I-1)
            III = NCOMM-(I-1)
            C0PCC1    = X(I,J) + CONJG(X(II,J))
            C2        = (X(I,J) - CONJG(X(II,J)))*TRIG(I)
            X(I,J)    = C0PCC1 + C2
            WS(III,J) = CONJG(C0PCC1 - C2)
C
          ENDDO
C
        ENDDO
C
      ELSEIF (TYPE .EQ. 1) THEN
C       Not processor 0, and no swap follows this computation.
C
        NCOMM = NCLOC/2
        DO J=1,MVECS
C
          DO I=1,NCOMM
C
            II = NCLOC-(I-1)
            C0PCC1  = X(I,J) + CONJG(X(II,J))
            C2      = (X(I,J) - CONJG(X(II,J)))*TRIG(I)
            X(I,J)  = C0PCC1 + C2
            X(II,J) = CONJG(C0PCC1 - C2)
C
          ENDDO
C
        ENDDO
C
      ELSE
C       Processor 0.
C
        DO J=1,MVECS
C
          X(1,J) = CMPLX(AREAL(X(1,J))+AIMAG(X(1,J)), 
     &                   AREAL(X(1,J))-AIMAG(X(1,J)))
          DO I=2,NCLOC
C
            II = IP(I-1) + 1
            IF (II .NE. 0) THEN
              C0PCC1  = X(I,J)+CONJG(X(II,J))
              C2      = (X(I,J)-CONJG(X(II,J)))*TRIG(I)
              X(I,J)  = C0PCC1 + C2
              X(II,J) = CONJG(C0PCC1 - C2)
            ENDIF
C
          ENDDO
C
        ENDDO
C
      ENDIF
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE BUFCPYA(CPYLTH, RJUMP, MVECS, IN, OUT1, OUT2)  
C
C BUFCPYA copies the first CPYLTH elements of each real vector in IN 
C into the corresponding complex vector in OUT1, and the rest (also 
C CPYLTH) into OUT2. 
C
C called by: RFT_235F
C calls: 
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C lengths of the real vectors in IN
      INTEGER CPYLTH
C length of first dimension in IN
      INTEGER RJUMP
C number of vectors to be copied
      INTEGER MVECS
C source array for the copy
      REAL IN(RJUMP,MVECS)
C
C     Output
C
C destination arrays for the copy
      REAL OUT1(CPYLTH,MVECS), OUT2(CPYLTH,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J
C                                                                             
C---- Executable Statements --------------------------------------------
C
      DO J=1,MVECS
        DO I=1,CPYLTH
          OUT1(I,J) = IN(I,J)
        ENDDO
        DO I=1,CPYLTH
          OUT2(I,J) = IN(CPYLTH+I,J)
        ENDDO
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE BUFCPYB(CPYLTH, RJUMP, OFFSET, MVECS, IN, OUT)  
C
C BUFCPYB copies the CPYLTH elements of each vector in IN into the first
C or second half of the corresponding vector in OUT.
C
C called by: RFT_235F, RFT_235B
C calls: 
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C lengths of the real vectors in OUT
      INTEGER CPYLTH
C length of first dimension in OUT
      INTEGER RJUMP
C offset used in copy to OUT
      INTEGER OFFSET
C number of vectors to be copied
      INTEGER MVECS
C source array for the copy
      REAL IN(CPYLTH,MVECS)
C
C     Output
C
C destination array for the copy
      REAL OUT(RJUMP,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J
C                                                                             
C---- Executable Statements --------------------------------------------
C
      DO J=1,MVECS
        DO I=1,CPYLTH
          OUT(OFFSET+I,J) = IN(I,J)
        ENDDO
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE BUFCPYC(CPYLTH, RJUMP, MVECS, IN, OUT1, OUT2)  
C
C BUFCPYC copies the first CPYLTH elements of each real vector in IN 
C into the corresponding complex vector in OUT1, and the rest (also 
C CPYLTH) into OUT2. BUFCPYC differs from BUFCPYA in that the 
C dimensions of OUT1 and OUT2 are reversed and the arrays are complex.
C
C called by: RFT_235B
C calls: 
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C lengths of the real vectors in IN
      INTEGER CPYLTH
C length of first dimension in IN
      INTEGER RJUMP
C number of vectors to be copied
      INTEGER MVECS
C source array for the copy
      REAL IN(RJUMP,MVECS)
C
C     Output
C
C destination arrays for the copy
      COMPLEX OUT1(MVECS,CPYLTH/2), OUT2(MVECS,CPYLTH/2)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J
C complex vector length
      INTEGER CVECLTH
C                                                                             
C---- Executable Statements --------------------------------------------
C
      CVECLTH = CPYLTH/2
C
      DO J=1,MVECS
        DO I=1,CVECLTH
          OUT1(J,I) = CMPLX(IN(2*I-1,J),IN(2*I,J))
        ENDDO
        DO I=1,CVECLTH
          OUT2(J,I) = CMPLX(IN(CPYLTH+2*I-1,J),IN(CPYLTH+2*I,J))
        ENDDO
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE BUFCPYD(CPYLTH, RJUMP, OFFSET, MVECS, IN, OUT)  
C
C BUFCPYD copies the CPYLTH/2 elements of each complex vector in IN 
C into the first or second half of the corresponding real vector in OUT.
C BUFCPYD differs from BUFCPYB in that the dimensions of IN
C are reversed and the array is complex.
C
C called by: RFT_235F
C calls: 
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C lengths of the real vectors in OUT
      INTEGER CPYLTH
C length of first dimension in OUT
      INTEGER RJUMP
C offset used in copy to OUT
      INTEGER OFFSET
C number of vectors to be copied
      INTEGER MVECS
C source array for the copy
      COMPLEX IN(MVECS,CPYLTH/2)
C
C     Output
C
C destination array for the copy
      REAL OUT(RJUMP,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J
C complex vector length
      INTEGER CVECLTH
C                                                                             
C---- Executable Statements --------------------------------------------
C
      CVECLTH = CPYLTH/2
C
      DO I=1,CVECLTH
        DO J=1,MVECS
          OUT(OFFSET+2*I-1,J) = AREAL(IN(J,I))
          OUT(OFFSET+2*I,J)   = AIMAG(IN(J,I))
        ENDDO
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      SUBROUTINE BUFCPYE(CPYLTH, RJUMP, OFFSET, MVECS, IN, OUT)  
C
C BUFCPYE copies CPYLTH elements starting at OFFSET of each vector in IN
C into the corresponding vector in OUT.
C
C called by: RFT_235F
C calls: 
C
C---- Implicit None ----------------------------------------------------
C
      IMPLICIT NONE
C
C---- Arguments --------------------------------------------------------
C
C     Input
C
C lengths of the real vectors in OUT
      INTEGER CPYLTH
C length of first dimension in IN
      INTEGER RJUMP
C offset used in copy from IN
      INTEGER OFFSET
C number of vectors to be copied
      INTEGER MVECS
C source array for the copy
      REAL IN(RJUMP,MVECS)
C
C     Output
C
C destination array for the copy
      REAL OUT(CPYLTH,MVECS)
C
C---- Local Variables --------------------------------------------------
C
C loop indices
      INTEGER I, J
C                                                                             
C---- Executable Statements --------------------------------------------
C
      DO J=1,MVECS
        DO I=1,CPYLTH
          OUT(I,J) = IN(OFFSET+I,J)
        ENDDO
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION BR(I, N)
C
C bit reverse representation of an n bit integer
C
      IMPLICIT NONE
C
      INTEGER I, N
C
      INTEGER II, J
C
      II = I
      BR = 0
      DO J=0,N-1
        BR = BR+MOD(II,2)*2**(N-1-J)
        II = II/2
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION CR(I, J, N)
C
C right cyclic shift of an n bit integer
C
      IMPLICIT NONE
C
      INTEGER I, J, N
C
      CR = I*2**J
      IF (I .GE. 2**(N-1)) CR = CR+1
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION CL(I, J, N)
C
C left cyclic shift of an n bit integer
C
      IMPLICIT NONE
C
      INTEGER I, J, N
C
      CL = I/2**J
      IF (MOD(I,2) .EQ. 1) CL = CL+2**(N-1)
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION IR(K, N, IFAX)
C
C routine to calculate the index reversed value of an index.
C
      IMPLICIT NONE
C
      INTEGER K, N, IFAX(0:12)
C
      INTEGER KK, M, J(0:12), I
c
c     Find the index
c
      M = N
      KK = K
      DO I=IFAX(0),1,-1
        M = M/IFAX(I)
        J(I-1) = KK/M
        KK = MOD(KK,M)
      ENDDO
c
c     Reverse the index
c
      IR = J(0)
      DO I=1,IFAX(0)-1
        IR = J(I)+ IR*IFAX(I+1)
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
      INTEGER FUNCTION RI(K, N, IFAX)
C
C routine to calculate the index reversed value of an index.
C
      IMPLICIT NONE
C
      INTEGER K, N, IFAX(0:12)
C
      INTEGER KK, M, J(0:12), I
C
C     Find the index
C
      M = N
      KK = K
      DO I=1,IFAX(0)
        M = M/IFAX(I)
        J(IFAX(0)-I) = KK/M
        KK = MOD(KK,M)
      ENDDO
c
c     Reverse the index
c
      RI = J(0)
      DO I=IFAX(0)-1,1,-1
        RI = J(IFAX(0)-I)+ RI*IFAX(I)
      ENDDO
C
      RETURN
      END
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
